-- $Id: gods.lua 2872 2017-05-10 00:29:08Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Birth descriptors corresponding to T2's gods.

local DamageType = require('engine.DamageType')

newBirthDescriptor {
  type = 'god',
  name = 'Nobody',
  desc = [[Your character need not worship any gods at all.]],
  eru_faction = 0,
  yavanna_faction = 0,
  copy = {
    resolvers.birth_valar_faction(0, 0),
    resolvers.birth_piety(false),
  }
}

newBirthDescriptor {
  type = 'god',
  name = 'Eru Ilúvatar',
  desc = [[He is the supreme god, he created the world, and most of its inhabitants.]],
  eru_faction = 100,
  yavanna_faction = 0,
  prayer_talent = ActorTalents.T_PRAYER_ERU,
  talents = {
    [ActorTalents.T_PRAYER_ERU] = 1,
  },
  copy = {
    -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
    -- normal-speed player turn.
    piety_regen_step = 100,
    piety_regen_f = function(who, pray)
      if pray == nil then pray = who:isPraying() end
      return pray and 0 or math.max(1, who:getStat('WIS', {scale=10}))
    end,
    resolvers.birth_valar_faction(100, 0),
    resolvers.birth_piety(true),
  },
  apply = function(self, who)
    local p = who.piety
    who:adjustStatForEquip('WIS', util.bound(math.floor(p/10000), 0, 3))
  end,
  quest = {
    relic = 'PIECE_OF_THE_RELIC_OF_ERU',
    achievement = 'RELIC_ERU',
    zone = {
      init_npc = 14,
      respawn_in = 200,
      floor = { weighted = { DIRT={70,10}, GRASS={30,90} } },
      fill = { 'MOUNTAIN_CHAIN' },
      outer_wall = 'MOUNTAIN_CHAIN',
      inner_wall = 'WALL',
      fill_step = 2,
      obj_weights = { treasure=45, combat=5, magic=45, tools=5, junk=0 },
      flags = {
	BIG = true,
	NO_DOORS = true,
	CIRCULAR_ROOMS = true,
	EMPTY = true,
	TOWER = true,
	FLAT = true,
	NO_SHAFT = true,
      },
      actor_filters = {
	{ prob=50, mode='or', flags={ 'EVIL' } },
	{ prob=50, mode='or', flags={ 'CAN_FLY' } },
      },
      actor_level_range = function(lev)
	return math.floor(lev/2), lev*2
      end
    },
  },
  callbacks = {
    on_take_hit_pre = function(who, d, src, p, wpn)
      -- Eru will block melee attacks from evil monsters if you're praying,
      -- depending on your piety.
      if who:isPraying() and src.has and src:has('EVIL') and wpn then
	local chance = math.min(who.piety, 50000) - (src.level or 0)*300
	if rng.range(100000) < chance then
	  game.logPlayer(who, game.flash.GOOD, [[The hand of Eru Ilúvatar blocks %s's blow.]], src:getName{def_art=true})
	  src:learnFlags('flags', 'EVIL')
	  p.damage = 0
	end
      end
    end,
    on_die_pre = function(who, data, src, p)
      -- If something else has saved our proverbial, no need to do anything.
      if p.no_die then return end
      -- If our piety is high enough, Eru might save us.
      if who:isPraying() and who.piety > 100000 and rng.percent(70) then
	game.logPlayer(who, '#LIGHT_GREEN#The power of %s raises you back from the grave!', d.name)
	p.no_die = true
	who.piety = -200000
      end
    end,
    on_kill = function(who, d, target, rsrc)
      if rsrc or not who.incPiety then return end
      -- Eru doesn't like you killing good monsters.
      if target:has('GOOD') then who:incPiety(-7 * target.level) end
    end,
    priorities = {
      -- Handle this before Protection from Evil.
      on_take_hit_pre = -4,
    }
  },
}

newBirthDescriptor {
  type = 'god',
  name = 'Manwë Súlimo',
  desc = [[He is the king of the Valar, most powerful of them after Melkor.]],
  eru_faction = 100,
  yavanna_faction = 0,
  prayer_talent = ActorTalents.T_PRAYER_MANWE,
  talents = {
    [ActorTalents.T_PRAYER_MANWE] = 1,
  },
  copy = {
    -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
    -- normal-speed player turn.
    piety_regen_step = 300,
    piety_regen_f = function(who, pray)
      if pray == nil then pray = who:isPraying() end
      local d = 4 - who:getStat('WIS', {scale=3})
      if pray then d = d + 1 end
      if who.base_flags and who.base_flags.ELF then
	d = d - who:getStat('WIS', {scale=2})
      end
      return -math.max(d, 1)
    end,
    resolvers.birth_valar_faction(100, 0),
    resolvers.birth_piety(true),
  },
  apply = function(self, who)
    local p = who.piety
    if who:isPraying() then
      who.speed = who.speed + math.min(math.floor(p/5000), 7)
      if p >= 7000 then who.flags.FREE_ACT = true end
      if p >= 15000 then who.flags.FLY = true end
    end
    if p >= 2000 then who.flags.FEATHER = true end
  end,
  quest = {
    relic = 'PIECE_OF_THE_RELIC_OF_MANWE',
    achievement = 'RELIC_MANWE',
    zone = {
      init_npc = 18,
      respawn_in = 160,
      floor = { weighted = { CLOUD_LIKE_VAPOUR=70, CONDENSING_WATER=30 } },
      fill = { 'HAIL_STONE_WALL' },
      outer_wall = 'DENSE_MIST',
      inner_wall = 'HAIL_STONE_WALL',
      fill_step = 4,
      obj_weights = { treasure=15, combat=25, magic=55, tools=5, junk=0 },
      flags = {
	NO_DOORS = true,
	TOWER = true,
	CAVERN = true,
	NO_SHAFT = true,
      },
      actor_filters = {
	{ prob=20, mode='or', flags={ 'INVISIBLE' } },
	{ prob=20, mode='or', flags={ 'ORC', 'IM_POIS' } },
	{ prob=20, mode='or', flags={ 'BR_POIS', 'BR_GRAV' } },
	{ prob=20, mode='or', flags={ 'BA_POIS' } },
	{ prob=20, mode='or', flags={ 'CAN_FLY' } },
      },
      actor_level_range = function(lev)
	return lev*2, lev*2
      end
    },
  },
  callbacks = {
    on_kill = function(who, d, target, rsrc)
      if rsrc or not who.incPiety then return end
      -- Manwe doesn't like you killing good monsters.
      if target:has('GOOD') then who:incPiety(-10 * target.level) end
      -- But he does like you killing evil monsters while praying.
      local d2 = math.max(math.floor(who.level/2), 1)
      if target:has('EVIL') then who:incPietyFor('Manwe', d2, true) end
    end,
  },
}

newBirthDescriptor {
  type = 'god',
  name = 'Tulkas',
  desc = [[He is the last of the Valar that came to the world, and the fiercest fighter.]],
  eru_faction = 0,
  yavanna_faction = 0,
  prayer_talent = ActorTalents.T_PRAYER_TULKAS,
  talents = {
    [ActorTalents.T_PRAYER_TULKAS] = 1,
  },
  copy = {
    -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
    -- normal-speed player turn.
    piety_regen_step = 300,
    piety_regen_f = function(who, pray)
      if pray == nil then pray = who:isPraying() end
      return pray and -math.max(1, 4 - who:getStat('WIS', {scale=3})) or 0
    end,
    resolvers.birth_valar_faction(0, 0),
    resolvers.birth_piety(true),
  },
  apply = function(self, who)
    local p = who.piety
    who:adjustStatForEquip('CON', util.bound(math.floor(p/5000), 0, 3))
    who:adjustStatForEquip('STR', util.bound(math.floor((p-5000)/5000), 0, 3))
  end,
  quest = {
    relic = 'PIECE_OF_THE_RELIC_OF_TULKAS',
    achievement = 'RELIC_TULKAS',
    zone = {
      init_npc = 20,
      respawn_n = 120,
      floor = { 'FLOOR' },
      fill = { 'WALL' },
      outer_wall = 'WALL',
      inner_wall = 'WALL',
      fill_step = 0,
      obj_weights = { treasure=10, combat=70, magic=5, tools=15, junk=0 },
      flags = {
	NO_DESTROY,
      },
      actor_filters = {
	{ prob=50, mode='or', flags={ 'DEMON', 'EVIL' } },
      },
      actor_level_range = function(lev)
	return lev*2, lev*2
      end
    },
  },
  callbacks = {
    on_kill = function(who, d, target, rsrc)
      if rsrc or not who.incPiety then return end
      -- Tulks likes you killing evil monsters, particularly if they're
      -- demons and/or you're praying.
      if target:has('EVIL') then
	local d2 = math.max(math.floor(who.level/2), 1)
	local d4 = math.max(math.floor(who.level/4), 1)
	who:incPiety(d4)
	who:incPietyFor('Tulkas', d4, true)
	if target:has('DEMON') then who:incPietyFor('Tulkas', d2, true) end
      end
    end,
  },
}

newBirthDescriptor {
  type = 'god',
  name = 'Melkor Bauglir',
  desc = [[He is the most powerful of the Valar. He became corrupted and he's now the greatest threat of Arda, he is also known as Morgoth, the dark enemy.]],
  eru_faction = -100,
  yavanna_faction = -100,
  prayer_talent = ActorTalents.T_PRAYER_MELKOR,
  talents = {
    [ActorTalents.T_PRAYER_MELKOR] = 1,
  },
  copy = {
    -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
    -- normal-speed player turn.
    piety_regen_step = 300,
    piety_regen_f = function(who, pray)
      if pray == nil then pray = who:isPraying() end
      local d = 8 - who:getStat('WIS', {scale=6})
      if pray then d = d + 1 end
      if who.base_flags and who.base_flags.ELF then
	d = d + 5 - who:getStat('WIS', {scale=4})
      end
      return -math.max(d, 1)
    end,
    resolvers.birth_valar_faction(-100, -100),
    resolvers.birth_piety(true),
  },
  apply = function(self, who)
    local p = who.piety
    local ds = util.bound(math.floor(p/10000), 0, 3)
    who:adjustStatForEquip('STR', ds)
    who:adjustStatForEquip('CON', ds)
    who:adjustStatForEquip('CHA', ds)
    who:adjustStatForEquip('INT', -ds)
    who:adjustStatForEquip('WIS', -ds)
    if who:isPraying() then
      if p > 5000 then who.invisibility = who.invisibility + 30 end
      if p > 15000 then who.immune_dam[DamageType.FIRE] = true end
    end
    local sac_bonus = who:getStat('WIS', {scale=4})*(who.melkor_sacrifice or 0)
    who:adjustDis('to_d', sac_bonus, true)
  end,
  quest = {
    relic = 'PIECE_OF_THE_RELIC_OF_MELKOR',
    achievement = 'RELIC_MELKOR',
    zone = {
      init_npc = 24,
      respawn_in = 80,
      floor = { weighted = { DIRT={45,35}, MUD={45,35}, NETHER_MIST={10,30} } },
      fill = { 'GLASS_WALL' },
      outer_wall = 'GLASS_WALL',
      inner_wall = 'WALL',
      fill_step = 1,
      obj_weights = { treasure=25, combat=25, magic=25, tools=25, junk=0 },
      flags = {
	SMALL = true,
	LAVA_RIVERS = true,
      },
      actor_filters = {
	{ prob=80, mode='any' },
	{ prob=20, mode='or', flags={ 'GOOD' } },
      },
      actor_level_range = function(lev)
	return math.floor(lev/2), lev
      end
    },
  },
  callbacks = {
    on_take_hit_post = function(who, d, src, dam)
      if who.life < who.max_life/4 and who:isPraying() then
	-- Chance of summoned monsters to help you.
	-- [cf. src/spells1.c lines 1488-1517]
	local chance = math.floor(who.piety/500) - 10
	if chance > 0 and rng.percent(chance) then
	  local filters = {
	    -- SUMMON_DEMON
	    function(e) return e:has('DEMON') and not e.unique end,
	    -- SUMMON_UNDEAD
	    function(e) return e:has('UNDEAD') and not e.unique end,
	    -- SUMMON_HI_DEMON
	    function(e)
	      return e:has('DEMON') and e.display == 'U' and not e.unique
	    end,
	    -- SUMMON_HI_UNDEAD
	    function(e)
	      return e.display == 'L' or e.display == 'V' or e.display == 'W'
	    end,
	  }
	  local i = rng.range(1, 2) + (who.piety > 10000 and 2 or 0)
	  local n = math.max(1, math.floor(who.piety/5000))
	  local ret = spell.summon(who, who.x, who.y, {special=filters[i]}, n)
	  if #ret > 0 then
	    game.logPlayer(who, 'Melkor summons monsters to help you!')
	  end
	end
      end
    end,
    on_kill = function(who, d, target, rsrc)
      if rsrc or not who.incPiety then return end
      -- Melkor likes you killing pretty much anything, but especially good
      -- monsters.
      if target:has('GOOD') then
	who:incPiety(3 * target.level)
      else
	who:incPiety(1 + math.floor(target.level/2))
      end
    end,
  },
}

newBirthDescriptor {
  type = 'god',
  name = 'Yavanna Kementári',
  desc = [[She is the Vala of nature, protectress of the great forests of Middle-earth.]],
  eru_faction = 0,
  yavanna_faction = 100,
  prayer_talent = ActorTalents.T_PRAYER_YAVANNA,
  talents = {
    [ActorTalents.T_PRAYER_YAVANNA] = 1,
  },
  copy = {
    -- NOTE:  piety_regen_step is in engine turns, which are 1/10 of a
    -- normal-speed player turn.
    piety_regen_step = 400,
    piety_regen_f = function(who, pray)
      -- TODO Check for not wilderness map, in dungeon, and 'did_nothing'
      local d = 5 - who:getStat('WIS', {scale=3})
      if who.descriptor and who.descriptor.race == 'Ent' then
	d = d - who:getStat('WIS', {scale=2})
      end
      return -math.max(d, 1)
    end,
    resolvers.birth_valar_faction(0, 100),
    resolvers.birth_piety(true),
  },
  apply = function(self, who)
    local p = who.piety
    if p >= 9000 and who:isPraying() then who.flags.PASS_TREE = true end
  end,
  quest = {
    relic = 'PIECE_OF_THE_RELIC_OF_YAVANNA',
    achievement = 'RELIC_YAVANNA',
    zone = {
      init_npc = 22,
      respawn_in = 100,
      floor = { weighted = { GRASS=40, GRASS_WITH_FLOWERS=15, DIRT=45 } },
      fill = { 'TREE' },
      outer_wall = 'SMALL_TREE',
      inner_wall = 'TREE',
      fill_step = 1,
      obj_weights = { treasure=20, combat=10, magic=30, tools=40, junk=0 },
      flags = {
	NO_DOORS = true,
	WATER_RIVERS = true,
	NO_DESTROY = true,
	NO_RECALL = true,
	NO_SHAFT = true,
	NO_GENO = true,
      },
      actor_filters = {
	{ prob=100, mode='or', flags={ 'DEMON', 'UNDEAD', 'NONLIVING' } },
      },
      actor_level_range = function(lev)
	return math.floor(lev/2), lev
      end
    },
  },
  callbacks = {
    on_kill = function(who, d, target, rsrc)
      if rsrc or not who.incPiety then return end
      local d2 = math.max(math.floor(who.level/2), 1)
      -- Yavanna likes you killing corrupt stuff.
      if target:has('NONLIVING') or target:has('DEMON') or target:has('UNDEAD') then
	who:incPiety(d2)
      end
      -- On the other hand, Yavanna doesn't like you killing while praying
      -- to her, especially animals.
      who:incPietyFor('Yavanna', -d2, true)
      if target:has('ANIMAL') then who:incPietyFor('Yavanna', -3*d2, true) end
    end,
  },
}
