-- $Id: lost-sword.lua 2925 2017-06-02 03:06:25Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

add_hooks = function()
  class:bindHook('ActorGenerator:T2LevelFilteredActors:generate', function(self, data)
    local zone, level = data.gen.zone, data.gen.level

    -- Not in non-mainline dungeons.
    if not zone.principal then return end
    -- Not unless we actually have a lost sword quest on this level.
    local lev = zone:level_adjust_level(level, 'actor')
    local state = game.state.sword_quests[lev]
    if not state or state.done then return end

    -- Remember the modified quest ID, which we'll need in our on_die()
    -- callback.
    state.q_id = 'lost-sword-' .. zone.short_name .. '-' .. lev

    -- Generate and place the quest monsters.
    local all_ok, mons = true, {}
    for _ = 1, state.n do
      m = data.gen:generateOne{ name=state.mon, ret=true, no_grp=true }
      if not m then
	all_ok = false
	break
      end
      m.quest_mon = true
      m.quest_on_die = state.q_id
    end
    if not all_ok then
      -- Failed to replace all the quest monsters for some reason; back out
      -- the ones we placed.
      print(('[LOST SWORD] only placed %d of %d quest mons; punting'):format(#m, state.n))
      for _, m in ipairs(mons) do game.level:removeEntity(m) end
    else
      -- Let the Game:enterLevel hook know that we successfully placed our
      -- monsters and that it can do its thing.
      state.placed= true
    end
  end)

  class:bindHook('Game:enterLevel', function(self, data)
    -- Make sure we successfully place the quest monsters.
    local lev = data.zone:level_adjust_level(data.level, 'actor')
    local state = game.state.sword_quests[lev]
    if not state or not state.placed then return end

    -- Clear this for next time.
    state.placed = nil

    -- Invoke the intro chat, which will give us the quest.
    local Chat = require 'engine.Chat'
    local Entity = require 'engine.Entity'
    -- A dummy NPC to display.
    local npc = Entity.new {
      name = 'Swordless Adventurer',
      chat_display_entity = game.zone.npc_list.ADVENTURER,
    }
    local chat = Chat.new('quest-lost-sword', npc, game.player)
    chat:invoke()
  end)

  class:bindHook('CharacterSheet:dumpLine', function(self, data)
    local n = 0
    for _, state in pairs(game.state.sword_quests) do
      if state.done then n = n + 1 end
    end
    if n == 1 then
      table.insert(data.lines, 'You have completed one lost sword quest.')
    elseif n > 1 then
      table.insert(data.lines, ('You have completed %d lost sword quests.'):format(n))
    end
  end)
end

-- Remember for use by desc().  [Have to be careful here; the "static"
-- pre-loading done in load.lua happens before game is defined.]
q_zone = game and game.zone and game.zone.name or 'Nowhere'
q_lev = game and game.level and game.zone:level_adjust_level(game.level, 'actor') or 0
local state = game and game.state and game.state.sword_quests[q_lev] or {}
q_num = state.n
q_pl_name = state.pl_name
q_killed = 0

id = 'lost-sword-' .. (game and game.zone and game.zone.short_name or '???') .. '-' .. q_lev
name = 'Lost Sword - ' .. q_zone .. ' level ' .. q_lev

desc = function(self, who)
  if self:isStatus(engine.Quest.DONE) then
    local fmt = [[You retrieved an adventurer's sword from %d %s on level %d of %s.  ]]
    local text = fmt:format(self.q_num, self.q_pl_name, self.q_lev, self.q_zone)
    if self.q_reward_mod > 0 then
      fmt = [[In return, he granted you access to the %s skill.]]
    else
      fmt = [[In return, he increased your %s skill.]]
    end
    text = text .. fmt:format(self.q_reward_skill)
    return text
  else
    local fmt = 'An adventurer on level %d of %s has lost his sword to a group of %d %s; you must kill all of them to retrieve his sword for him.'
    local text = fmt:format(self.q_lev, self.q_zone, self.q_num, self.q_pl_name)
    if self.q_killed > 0 then
      fmt = '  You have killed %d of them so far.'
      text = text .. fmt:format(self.q_killed)
    end
    return text
  end
end

on_status_change = function(self, who, status, sub)
  if self:isEnded() then
    -- Mark this quest done.
    game.state.sword_quests[self.q_lev].done = true
  end
end

on_quest_mon_die = function(self, who, src)
  self.q_killed = self.q_killed + 1
  if self.q_killed >= self.q_num then
    local Entity = require 'engine.Entity'
    local Chat = require 'engine.Chat'
    -- A dummy NPC to display.
    local npc = Entity.new {
      name = 'Swordless Adventurer',
      chat_display_entity = game.zone.npc_list.ADVENTURER,
    }
    local chat = Chat.new('quest-lost-sword-finish', npc, game.player)
    chat:invoke()
  end
end

callbacks = {
  on_leave_level = function(src, q, zone, level, lev, params)
    -- Reset the kill count if we leave the level before completing the
    -- quest.
    q.q_killed = 0
  end,
}
