-- $Id: princess.lua 2925 2017-06-02 03:06:25Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

add_hooks = function()
  class:bindHook('MapGenerator:T2DungeonLevel:firstRoom', function(self, data)
    -- Not if someone else has already created the first room.
    if data.done then return end
    -- Not in non-mainline dungeons.
    if not data.gen.zone.principal then return end
    -- Not unless we actually have a princess quest on this level.
    local lev = data.gen.zone:level_adjust_level(data.gen.level, 'actor')
    local state = game.state.princess_quests[lev]
    if not state or state.done then return end

    -- Build the quest room.  This is loosely based on T4's 'lesser_vault'
    -- room.
    local Map = require 'mod.class.Map'
    local Static = require 'mod.class.generator.map.Static'
    local gen_data = table.clone(data.gen.data)
    gen_data.map = ('quests/princess-%d'):format(state.n)
    -- Sneaky Hack(TM):  The map generator will need our quest info.
    gen_data.quest_info = state
    -- It will also need the modified quest ID for the on_die() callback
    -- it's going to install.
    state.q_id = 'princess-' .. data.gen.zone.short_name .. '-' .. lev

    -- Questionable Hack(TM):  Create a new map to load the quest room
    -- into, and wedge it in everywhere the generator will expect the level
    -- map to be.
    local room_map = Map.new(50, 50)
    local orig_gen_level_map = data.gen.level.map
    data.gen.level.map = room_map
    local orig_game_level = game.level
    game.level = data.gen.level

    -- Create a new Static map generator and load our room map.
    local room_gen = Static.new(data.gen.zone, room_map, data.gen.level, gen_data)
    local _, _, _, _, room_spots = room_gen:generate(data.lev, data.old_lev)

    -- Questionable Hack(TM), redux:  Put back the actual level map.
    game.level = orig_game_level
    data.gen.level.map = orig_gen_level_map

    -- Now that we have the room size, try to allocate it on the map.
    local width, height = room_map.w, room_map.h
    local ctr = data.gen:roomAlloc(width+2, height+2, false, data.gx, data.gy)
    if not ctr then
      error('[QUEST] ??? Could not roomAlloc() princess quest room')
      return
    end

    local y1 = ctr.y - math.floor(height/2)
    local x1 = ctr.x - math.floor(width/2)
    local y2 = y1 + height - 1
    local x2 = x1 + width - 1

    -- First build an ordinary rectangular room.  [We mostly duplicate
    -- data/rooms/basic.lua here.]
    --   - Floor
    data.gen:build_floor(x1, y1, x2, y2, 'floor', true, {'room'})
    --   - Walls.
    data.gen:build_rectangle(x1-1, y1-1, x2+1, y2+1, 'outer_wall', true, {'room'})

    -- Then overlay the quest room map on top of the room we just built.
    data.gen.map:overlay(room_map, x1, y1)

    -- Pass the spots defined by the map back to the caller.
    for _, spot in ipairs(room_spots) do
      spot.x = spot.x + x1
      spot.y = spot.y + y1
    end
    table.append(data.spots, room_spots)

    -- Let the Game:enterLevel hook know that we successfully built our
    -- room and it can do its thing.
    state.placed = true
    -- Tell our hook caller that we succeeded...
    data.done = true
    -- ...and that it shouldn't place any other down stairs on the level.
    data.no_down = true
  end)

  class:bindHook('Game:enterLevel', function(self, data)
    -- Make sure we successfully placed the princess quest room.
    local lev = data.zone:level_adjust_level(data.level, 'actor')
    local state = game.state.princess_quests[lev]
    if not state or not state.placed then return end

    -- Clear this for next time.
    state.placed = nil

    -- Invoke the intro chat, which will give us the quest.
    local Entity = require 'engine.Entity'
    local Map = require 'mod.class.Map'
    local Chat = require 'engine.Chat'
    local spot = data.level:pickSpot{ type='princess-quest', subtype='princess' }
    local npc = data.level.map(spot.x, spot.y, Map.ACTOR)
    npc = npc or Entity.new {
      name = 'The Princess (proxy)',
      chat_display_entity = game.zone.npc_list.THE_PRINCESS,
    }
    local chat = Chat.new('quest-princess', npc, game.player)
    chat:invoke()
  end)

  class:bindHook('CharacterSheet:dumpLine', function(self, data)
    local n = 0
    for _, state in pairs(game.state.princess_quests) do
      if state.done then n = n + 1 end
    end
    if n == 1 then
      table.insert(data.lines, 'You have completed one princess quest.')
    elseif n > 1 then
      table.insert(data.lines, ('You have completed %d princess quests.'):format(n))
    end
  end)
end

-- Remember for use by desc().  [Have to be careful here; the "static"
-- pre-loading done in load.lua happens before game is defined.]
q_zone = game and game.zone and game.zone.name or 'Nowhere'
q_lev = game and game.level and game.zone:level_adjust_level(game.level, 'actor') or 0
local state = game and game.state and game.state.princess_quests[q_lev] or {}
q_num = state.n
q_name = state.name
q_pl_name = state.pl_name
q_pronoun = state.pronoun
q_killed = 0

id = 'princess-' .. (game and game.zone and game.zone.short_name or '???') .. '-' .. q_lev
name = 'Captive Princess - ' .. q_zone .. ' level ' .. q_lev

desc = function(self, who)
  local qname = self.q_num == 1 and self.q_name or self.q_num..' '..self.q_pl_name
  if self:isStatus(engine.Quest.DONE) then
    local fmt = [[You rescued a princess being held captive by %s on level %d of %s.  As a reward, she gave you %s.]]
    return fmt:format(qname, self.q_lev, self.q_zone, self.q_reward)
  else
    local them = self.q_num == 1 and self.q_pronoun or 'all of them'
    local fmt = [[A princess is being held captive by %s on level %d of %s.  You must slay %s to free the princess]]
    local ret = fmt:format(qname, self.q_lev, self.q_zone, them)
    if self.q_killed > 0 then
      -- If q_killed > 0 and the quest is still pending, there must be
      -- multiple quest monsters.
      fmt = [[; you have killed %d of them so far.]]
      ret = ret .. fmt:format(self.q_killed)
    end
    return ret .. '.'
  end
end

on_status_change = function(self, who, status, sub)
  if self:isEnded() then
    -- Mark this quest done.
    game.state.princess_quests[self.q_lev].done = true
  end
end

on_quest_mon_die = function(self, who, src)
  self.q_killed = self.q_killed + 1
  if self.q_killed >= self.q_num then
    -- Completed the quest.  First, drop the walls, for cosmetic appeal.
    local _, walls = game.level:pickSpot{ type='princess-quest', subtype='glass-wall' }
    game:onTickEnd(function()
      local Map = require 'mod.class.Map'
      for _, wall in ipairs(walls) do
	game.level.map(wall.x, wall.y, Map.TERRAIN, game.zone.grid_list.FLOOR)
      end
    end)

    -- Generate the quest reward choices.
    local Object = require 'mod.class.Object'
    local choices = {}
    for _ = 1, 3 do
      Object.resolve_flags.force_great = true
      local filter = game.zone:createWeightedObjectFilter({ treasure=100, combat=100, magic=100, tools=100, junk=0 })
      local o = game.zone:makeEntity(game.level, 'object', filter, nil, true)
      choices[#choices+1] = o
    end

    -- Then go to the quest finish chat, which will take care of the rest
    -- of the terraforming and placing the chosen reward.
    local Entity = require 'engine.Entity'
    local Map = require 'mod.class.Map'
    local Chat = require 'engine.Chat'
    local pr_spot = game.level:pickSpot{ type='princess-quest', subtype='princess' }
    local princess = game.level.map(pr_spot.x, pr_spot.y, Map.ACTOR) or Entity.new {
      name = 'The Princess (proxy)',
      chat_display_entity = game.zone.npc_list.THE_PRINCESS,
    }
    local chat = Chat.new('quest-princess-finish', princess, game.player, { choices=choices })
    chat:invoke()
  end
end

callbacks = {
  on_leave_level = function(src, q, zone, level, lev, params)
    -- Reset the kill count if we leave the level before completing the
    -- quest.
    q.q_killed = 0
  end,
}
