-- $Id: alchemy.lua 3041 2017-08-30 02:40:05Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Alchemy powers',
  type = 'skill/alchemy',
  description = 'Talents conferred by the Alchemy skill',
}

newTalent {
  name = [[Forge Item]],
  short_name = 'ALCH_FORGE',
  display_entity = Entity.new { image='talents/forge_item.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = 'Forge item',
    allow_from = { inven=true },
    filter = function(o) return game.player:hasForgeItemRecipes(o) end,
    close_before = true,
    action = function(self, t, o, inven, item, floor)
      return self:playerForgeItem(o, inven, item)
    end,
  },
  callbacks = {
    -- This could live on any of the Alchemy-related talents that we get at
    -- Alchemy skill level 1; we choose this one.
    on_identify = function(who, t, o, lvl)
      -- Annoying Hack(TM):  Stores *ID* their stock as they're loading it,
      -- and we don't want that to trigger recipe learning.
      local Store = require 'mod.class.Store'
      if lvl == 'full' and not Store.stocking then who:alchemyLearnObject(o) end
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item)
      self:talentDialogReturn(true, {o=o, item=item})
      return false
    end
    local ShowInventory = require 'mod.dialogs.ShowInventory'
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInventory.new('Select base item', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, false)
  end,
  info = [[Infuse a plain item like a Ring of Nothing with essences to create a more useful item.  You will need to know the appropriate recipe for the type of item you want to create; you can learn recipes by using a Scroll of *Identify* (or similar effect) on an item, or sometimes by disassembling it with the Extract Essences talent #{italic}#(q.v.)#{normal}#.  You will also learn some recipes automatically as you increase your Alchemy skill.

Note that creating particularly powerful items comes with a chance of failure, increasing with item level and decreasing with your Alchemy skill level.  You can spend gold to decrease the failure chance; the amount needed will depend on the object's value.

You can review your known recipes via the Show Alchemist Recipes dialog on the main game menu.]],
}

newTalent {
  name = [[Add Ego to Item]],
  short_name = 'ALCH_EGO',
  display_entity = Entity.new { image='talents/add_ego_to_item.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = 'Add ego',
    allow_from = { inven=true },
    filter = function(o) return game.player:hasForgeEgoRecipes(o) end,
    close_before = true,
    action = function(self, t, o, inven, item, floor)
      return self:playerForgeEgo(o, inven, item)
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item)
      self:talentDialogReturn(true, {o=o, item=item})
      return false
    end
    local ShowInventory = require 'mod.dialogs.ShowInventory'
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInventory.new('Select base item', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, false)
  end,
  info = [[Infuse an object with essences to add an ego to it.  You will need to know the appropriate recipe for the type of ego you want to add; you can learn recipes by using a Scroll of *Identify* (or similar effect) on an ego item, or sometimes by disassembling it with the Extract Essences talent #{italic}#(q.v.)#{normal}#.  You will also learn some ego recipes automatically as you increase your Alchemy skill.

Note that adding particularly powerful egos comes with a chance of failure, increasing with item level and decreasing withr your Alchemy skill level.  You can spend gold to decrease the failure chance; the amount needed will depend on the ego's value.

You can review your known recipes via the Show Alchemist Recipes dialog on the main game menu.]],
}

newTalent {
  name = [[Recharge Wand/Staff]],
  short_name = 'ALCH_RECHARGE',
  display_entity = Entity.new { image='talents/recharge_item.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = 'Recharge using essences',
    allow_from = { inven=true },
    filter = function(o)
      local def = game.player:getExtractRecipeFor(o, true, true)
      if not def or not def.charges then return false end
      return game.player:numRecipeEssences(def) > 0
    end,
    action = function(self, t, o, inven, item, floor)
      return self:playerRecharge(o, inven, item)
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item)
      self:talentDialogReturn(true, {o=o, item=item})
      return false
    end
    local ShowInventory = require 'mod.dialogs.ShowInventory'
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInventory.new('Select wand/staff to recharge', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, false)
  end,
  info = [[Use essences in your inventory to add charges to a wand or staff also in your inventory.  Note that unlike magical recharging via spell or scroll, this does not prevent charges from being extracted as essences later.]],
}

newTalent {
  name = [[Extract Essences]],
  short_name = 'ALCH_EXTRACT',
  display_entity = Entity.new { image='talents/extract_essences.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = 'Extract essences',
    allow_from = { inven=true, floor=true },
    filter = function(o)
      return game.player:getExtractRecipeFor(o)
    end,
    close_before = true,
    action = function(self, t, o, inven, item, floor)
      return self:playerExtract(o, inven, item, floor)
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item, floor)
      self:talentDialogReturn(true, {o=o, item=item, floor=floor})
      return true
    end
    local ShowInvenFloor = require('mod.dialogs.ShowInvenFloor')
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInvenFloor.new('Select object to extract', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, p.floor)
  end,
  info = [[Disassemble an item, extracting alchemic essences from it.  This can be done in several ways:
  
· You can extract charges from a wand or staff (unless the wand has been magically recharged, which disrupts the process).

· Certain egos (like Resist Fire on armor) can be extracted from various enchanted items, leaving behind an unenchanted item.

· Certain magical items like amulets or rings, or potent natural items like mushrooms, can be disassembled entirely.  Usually this will leave behind a "blank" item; rings, for instance, will leave a Ring of Nothing (mushrooms, notably, will be consumed entirely).

Typically the essences gained from this extraction will be of the same type and number needed to reconstruct the item as it was, although occasionally a few will be lost in the process (the chance of this will decrease with your Alchemy skill level).  There is also a 1 in 3 chance that you will learn how to reassemble the item as you are disassembling it.]],
}

newTalent {
  name = [[Detach Rod Tip]],
  short_name = 'ALCH_DETACH',
  display_entity = Entity.new { image='talents/detach_rod_tip.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = 'Detach rod tip',
    allow_from = { inven=true, floor=true },
    filter = function(o) return o.rod_tip end,
    close_before = true,
    action = function(self, t, o, inven, item, floor)
      if not o.rod_tip then return false end
      local tip = o.rod_tip
      o.rod_tip = nil
      game.logPlayer(self, 'You detach %s from your %s.', tip:getName(), o:getName{no_count=true})
      self:addOrDropObject(self.INVEN_INVEN, tip, nil, nil, {vocal_drop=true})
      self:sortInven()
      util.live_update_inventory_dialogs()
      self:recalcEverything()
      return true
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item, floor)
      self:talentDialogReturn(true, {o=o, item=item, floor=floor})
      return true
    end
    local ShowInvenFloor = require('mod.dialogs.ShowInvenFloor')
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInvenFloor.new('Select rod to detach tip', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, p.floor)
  end,
  info = [[Separate a rod base and its attached rod tip, allowing either to be reused with more suitable gear.]],
}

newTalent {
  name = [[Forge Artifact]],
  short_name = 'ALCH_ARTIFACT',
  display_entity = Entity.new { image='talents/forge_artifact.png' },
  type = 'skill/alchemy',
  mode = 'activated',
  points = 0,
  use_item = {
    name = function(self, t, o, _, _, _)
      if o.art_exp then return 'Update or finish artifact' end
      return 'Begin artifact conversion'
    end,
    allow_from = { inven=true },
    filter = function(o)
      -- Slight Hack(TM):  T2 will destroy all but one of a stack of things
      -- if you try to Becoming them; we just don't let you select them.
      return not o.cursed and (o:isArtifact() or o:isEgo()) and o:getNumber() == 1
    end,
    close_before = true,
    action = function(self, t, o, inven, item, floor)
      if o.art_exp then
	local ArtifactDialog = require 'mod.dialogs.AlchemyArtifactDialog'
	local ReviewDialog = require 'mod.dialogs.AlchemyReviewDialog'
	local state = self:talentDialog(ArtifactDialog.new(self, o))
	while true do
	  if not state then return nil end
	  local done = self:talentDialog(ReviewDialog.new(self, o, state))
	  if done == 'cancel' then return nil end
	  if done then return true end
	  state = self:talentDialog(ArtifactDialog.new(self, o, state))
	end
      else
	return self:startBecoming(o)
      end
    end,
  },
  action = function(self, t)
    local function sel_cb(o, item)
      self:talentDialogReturn(true, {o=o, item=item})
      return false
    end
    local ShowInventory = require 'mod.dialogs.ShowInventory'
    local inven = self:getInven('INVEN')
    local ret, p = self:talentDialog(ShowInventory.new('Select item', inven, t.use_item.filter, sel_cb, self))
    if not ret then return nil end
    return t.use_item.action(self, t, p.o, inven, p.item, false)
  end,
  on_learn = function(self, t)
    -- Pre-learn all alchemy recipes that don't have a flag to match;
    -- otherwise we'd learn them all with the first artifact we *ID*, which
    -- is aesthetically unsatisfying.
    self:alchemyLearnObject(true)
  end,
  info = function(self, t)
    local fmt = [[The pinnacle of alchemic skill -- create an artifact of your own custom design.  This is done in stages:
  
· To begin the process, use this talent on an existing artifact or a non-artifact enchanted item with at least one ego, converting it to a "Becoming" artifact.  This incomplete item will siphon off some of the experience you gain as you fight (if you're wearing it at the time), storing it for use in the next step.  You will need a number of Essences of Magic equal to your current Alchemy skill level (currently %d).  #{bold}#NOTE:  Creating a Becoming artifact will permanently reduce your maximum life by 10.#{normal}#

· Later, use this talent again on your Becoming artifact to add powers like resistances or slays to it, using artifact recipes that you have learned by using Scrolls of *Identify* (or similar effect) on other artifacts with the powers you want.  Most recipes will require extra ingredients and will use up some of the experience your artifact has been accumulating.

You can review your known recipes via the Show Alchemist Recipes dialog on the main game menu.]]
    return fmt:format(self:getSkillLevel('ALCHEMY'))
  end,
}
