-- $Id: Home.lua 3002 2017-07-26 01:36:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Store = require 'engine.Store'
local ShowStore = require 'mod.dialogs.ShowStore'
local GetQuantity = require "engine.dialogs.GetQuantity"

module(..., package.seeall, class.inherit(Store))

_M.homes_def = {}

function _M:loadHomes(f)
  self.homes_def = self:loadList(f)
end

function _M:init(t, no_default)
  Store.init(self, t, no_default)
  -- There's no obvious way to override the 't.body = {INVEN=10000}' in
  -- Store.init(), so we patch that up here.
  self:getInven('INVEN').max = 24
end

function _M:canRestock()
  -- Homes don't restock.
  return false
end

function _M:loadup(level, zone, force_nb)
  -- Nor should they be filled with an initial stock.
end

-- Mostly the same as Store.interact(), with a few tweaks.
function _M:interact(who, name)
  local store, inven = self:getInven("INVEN"), who:getInven("INVEN")
  local d
  local action_cb = function(what, o, item)
    if what == "buy" then
      if o:getNumber() > 1 then
	local q = GetQuantity.new(nil, nil, o:getNumber(), o:getNumber(), function(qty) self:doBuy(who, o, item, qty, d) end)
	game:registerDialog(q)
      else
	self:doBuy(who, o, item, 1, d)
      end
    else
      if o:getNumber() > 1 then
	local q
	q = GetQuantity.new(nil, nil, o:getNumber(), o:getNumber(), function(qty) self:doSell(who, o, item, qty, d) end)
	game:registerDialog(q)
      else
	self:doSell(who, o, item, 1, d)
      end
    end
  end
  local desc_cb = function(what, o) return self:descObject(who, what, o) end
  local descprice_cb = function(what, o) return '' end
  d = ShowStore.new((name or self.name), self, store, who, inven, nil, nil,
		    action_cb, desc_cb, descprice_cb,
		    self.allow_sell, self.allow_buy,
		    function(item) end, self, who)
  game:registerDialog(d)

  -- Remember the last home visited, for purposes of the Rod Tip of Home
  -- Summoning.
  local lev = self.zone:level_adjust_level(self.level, 'trap')
  game.state.last_visited_home = game.zone.short_name .. '-' .. lev .. '/home-' .. self.define_as
end

function _M:trySell(who, o, item, qty)
  -- Do we have room in the home?
  local inven = self:getInven('INVEN')
  local can_add = #inven < inven.max
  if not can_add and o and o:stackable() then
    -- Maybe we can stack it with something?
    local stack_limit = self:invenStackLimit('INVEN')
    for i, io in ipairs(inven) do
      if o:canStack(io) and io:getNumber() + qty <= stack_limit then
	can_add = true
	break
      end
    end
  end
  if not can_add then
    Dialog:simplePopup('Home full', 'There is no room for that in your home.')
    return nil, nil
  end

  return qty, 0
end

function _M:doBuy(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  self:transfer(self, who, item, qty, true)
  if store_dialog then store_dialog:updateStore() end
  game.log('%s gets from home:  %s', who.name, o:getName{ force_qty=qty })
end

function _M:doSell(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  qty, _ = self:trySell(who, o, item, qty)
  if not qty then return end
  local o_name = o:getName{ force_qty=qty }
  self:transfer(who, self, item, qty, false)
  if store_dialog then store_dialog:updateStore() end
  game.log('%s puts in home:  %s', who.name, o:getName{ force_qty=qty })
end

-- We replace the parent transfer() to support dropping the "bought" object
-- if it doesn't fit into the player's inventory.
function _M:transfer(src, dest, item, nb, buying)
  local src_inven, dest_inven = src:getInven("INVEN"), dest:getInven("INVEN")
  for i = 1, nb do
    local o = src:removeObject(src_inven, item)
    if buying then
      dest:addOrDropObject(dest_inven, o, nil, nil, {vocal_drop=true})
    else
      dest:addObject(dest_inven, o)
    end
  end
  src:sortInven(src_inven)
  dest:sortInven(dest_inven)
end

function _M:buyActionName()
  return 'Get'
end

function _M:sortInven(inven)
  if not inven then inven = self.inven[self.INVEN_INVEN] end
  inven = self:getInven(inven)
  if not inven then return end
  obj_util.sort_inven(inven)
  self.changed = true
end
