-- $Id: NPC.lua 2919 2017-06-01 02:15:31Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
local ActorAI = require "engine.interface.ActorAI"
local Faction = require "engine.Faction"
local DamageType = require 'engine.DamageType'
local AttackType = require 'mod.class.AttackType'
require "mod.class.Actor"
require 'mod.class.interface.NPCFields'
local Object = require 'mod.class.Object'

module(..., package.seeall, class.inherit(mod.class.Actor,
					  engine.interface.ActorAI,
					  mod.class.interface.NPCFields))

_M.resolve_flags = {}

function _M:init(t, no_default)
  mod.class.Actor.init(self, t, no_default)
  ActorAI.init(self, t)

  self.max_level = 150
  self.exp = self:getExpChart(self.level)

  -- For later reference if the monster gets leveled.
  self.dflt_depth = self.level

  -- Default regeneration rate for NPCs.
  self.life_regen_pct = self.flags and self.flags.REGENERATE and 0.02 or 0.01
end

function _M:resolve(t, last, on_entity, key_chain)
  -- Dodgy Hack(TM):  resolve() calls itself, and we only want to do our
  -- special stuff once at the top level.
  if self.nested then
    mod.class.Actor.resolve(self, t, last, on_entity, key_chain)
    return
  end
  self.nested = true

  local NPCEgo = require 'mod.class.NPCEgo'
  if not last then
    NPCEgo:tryAddEgo(self)
    NPCEgo:applyEgo(self)
  end

  mod.class.Actor.resolve(self, t, last, on_entity, key_chain)

  if last then
    -- Clean these up behind ourself.
    _M.resolve_flags = {}
    -- This isn't getting handled right, and I can't figure out how to
    -- disentangle it; ham-hand it manually here.
    self.life = self.max_life
  end
  self.nested = nil
end

function _M:act()
  while self:enoughEnergy() and not self.dead do
    -- Do basic actor stuff
    if not mod.class.Actor.act(self) then return end
    local old_energy = self.energy.value

    -- Compute FOV, if needed
    self:computeFOV(self.sight or 20)

    -- Let the AI think .... beware of Shub !
    -- If AI did nothing, use energy anyway
    if not self:hasEffect(self.EFF_PARALYZED) then
      self:doAI()
    end
    if not self.energy.used then self:useEnergy() end
    if old_energy == self.energy.value then break end -- Prevent infinite loops
  end
end

-- Overridden to regen by percentage.
function _M:regenLife()
  self.life = self.life + self.life * self.life_regen_pct
  self.life = util.bound(self.life, self.die_at, self.max_life)
  self.changed = true
  -- If we're a symbiote, tell our host to redraw our HP.
  if self.symbiosis then self.symbiosis.host.changed = true end
end

--- Called by ActorLife interface
-- We use it to pass aggression values to the AIs
function _M:onTakeHit(value, src)
  if not self.ai_target.actor and src.targetable then
    self.ai_target.actor = src
  end

  -- If we're a symbiote, tell our host to redraw our HP.
  if self.symbiosis then self.symbiosis.host.changed = true end

  return mod.class.Actor.onTakeHit(self, value, src)
end

function _M:move(x, y, force)
  local ox, oy = self.x, self.y
  local moved = mod.class.Actor.move(self, x, y, force)

  if moved and (self.x ~= ox or self.y ~= oy) then
    if self:has('TAKE_ITEM') or self:has('KILL_ITEM') then
      -- Try to pick up or destroy items on the grid we just entered.
      local i = 1
      local o = game.level.map:getObject(x, y, i)
      while o do
	local got = false
	if self:has('KILL_ITEM') and not o:isArtifact() then
	  game.level.map:removeObject(x, y, i)
	  got = true
	  self:learnFlags('flags', 'KILL_ITEM')
	  game.logSeen(self, "%s crushes %s.", self:getName():capitalize(), o:getName())
	  if game:hasEntity(o) then game:removeEntity(o) end
	end
	if not got and self:has('TAKE_ITEM') and self:reactionToward(game.player) < 0 then
	  if not self:canTouch(o) then
	    self:learnFlags('flags', 'TAKE_ITEM')
	    game.logSeen(self, "%s tries to pick up %s, but fails.", self:getName():capitalize(), o:getName())
	  elseif self:pickupFloor(i, true) then
	    got = true
	    self:learnFlags('flags', 'TAKE_ITEM')
	    game.logSeen(self, "%s picks up  %s.", self:getName():capitalize(), o:getName())
	  end
	end
	if not got then i = i + 1 end
	o = game.level.map:getObject(x, y, i)
      end
    end
  end

  return moved
end

function _M:canTouch(o)
  -- No NPC can pick up an artifact.
  if o:isArtifact() then return false end
  -- NPCs can't pick up a weapon with a "slay" flag that would affect them.
  for f, _ in pairs(o.slay_flags or {}) do
    local ff = f:gsub('^X_', '')
    if self:has(ff) then return false end
  end
  -- Touch away, then.
  return true
end

function _M:canPushPast(actor)
  -- We can push past if:
  --   - we have the MOVE_BODY flag
  --   - we are not confused
  --   - we are not hostile to the actor (and actor is not the player)
  --   - we are "stronger (ie. more experienced) than the actor
  --   - we and the actor can both move into each other's grid
  -- [cf. src/melee2.c lines 7103,7120-7122]
  return self:has('MOVE_BODY') and actor:reactionToward(self) >= 0 and
	 not self:has('CONFUSED') and
	 not actor.player and self.exp > actor.exp and
	 self:canMove(actor.x, actor.y, true) and
	 actor:canMove(self.x, self.y, true)
end

function _M:canSquish(actor)
  -- We can squish if:
  --   - we have the KILL_BODY flag
  --   - we are hostile to the actor
  --   - the actor is not the player or a unique or quest monster
  --   - we are "stronger (ie. more experienced) than the actor
  -- [cf. src/melee2.c lines 7076-7085]
  return self:has('KILL_BODY') and actor:reactionToward(self) < 0 and
	 not actor.player and not actor.unique and not actor.quest_mon and
	 self.exp > actor.exp
end

function _M:tryOpenDoor(g, x, y, dry_run)
  if dry_run then return self:has('OPEN_DOOR') or self:has('BASH_DOOR') end

  -- TODO Allow configuring whether pets can open doors.
  if self:has('OPEN_DOOR') then
    -- Only try to open non-jammed doors.
    if not g.door_jam_power and (not g.try_open_door or g:try_open_door(self)) then
      g:openDoor(x, y, self)
      self:learnFlags('flags', 'OPEN_DOOR')
      self:useEnergy()
      return true
    end
  end
  -- TODO Bashing should use the same can-pets-open-doors setting as above.
  if self:has('BASH_DOOR') then
    -- This is the same probability check as in try_open_door()
    -- [cf. src/melee2.c lines 6908-6922]
    local power = g.door_jam_power or g.door_lock_power or 0
    if rng.range(1, math.floor(self.life/10)) > power then
      game.log('You hear a door burst open!')
      -- TODO Possibly interrupt resting.
      g:openDoor(x, y, self, 'BROKEN_DOOR')
      self:learnFlags('flags', 'BASH_DOOR')
      self:useEnergy()
      return true
    end
  end
  return false
end

-- *Very* loosely based on roff_aux() in src/monster1.c
function _M:getDesc(t)
  t = t or {}
  local raw = self:rawData(true)
  local learned = self:learnedData()
  local use_learned = config.settings.tome2.npc_learn
  local src = use_learned and learned or self
  local rsrc = use_learned and raw or self

  local desc = ''
  if not t.no_name then
    desc = self:getName{ indef_art=true, no_your=true, cite_given_name=true }
    if not t.no_image then desc = self:getDisplayString() .. desc end
    if t.with_level then
      desc = desc .. (' (level %d)'):format(self.level)
    end
    if self.quest_mon then
      desc = desc .. ' #{italic}##LIGHT_BLUE#(quest monster)#LAST##{normal}#'
    end
    if config.settings.cheat then
      desc = desc .. (' [UID: %d]'):format(self.uid)
    end
    desc = desc .. '\n\n'
  end

  local p_kills, w_kills = self:killCount()
  local kill_p = self:playerKillCount()
  local p_kills_f = tostring(p_kills == 0 and 'none' or p_kills)
  local w_kills_f = tostring(w_kills == 0 and 'none' or w_kills)
  kill_p = (kill_p == 0 and 'never') or (kill_p == 1 and 'once') or ('%d times'):format(kill_p)
  desc = desc .. ('#{bold}#Kills:#{normal}# %s (you)/%s (all characters)'):format(p_kills_f, w_kills_f)
  desc = desc .. '\n#{bold}#Killed by:#{normal}# ' .. kill_p

  if self.faction and Faction.factions[self.faction] then
    local facname = Faction.factions[self.faction].name
    local react = Faction:factionReaction(self.faction, game.player.faction)
    local color = react < 0 and 'LIGHT_RED' or react > 0 and 'LIGHT_GREEN' or 'ANTIQUE_WHITE'
    local state = react < 0 and 'hostile' or react > 0 and 'friendly' or 'neutral'
    desc = desc .. ('\n#{bold}#Faction:#{normal}# #%s#%s(%s, %d)#LAST#'):format(color, facname, state, react)
  end

  if rsrc.level == 0 then
    desc = desc .. '\n#{bold}#Default depth:#{normal}# #LIGHT_GREEN#Town/wilderness#LAST#'
  elseif w_kills > 0 then
    local z_lev = game.zone:level_adjust_level(game.level, 'actor')
    local color = rsrc.level > z_lev and 'LIGHT_RED' or 'LIGHT_GREEN'
    desc = desc .. ('\n#{bold}#Default depth:#{normal}# #%s#%d#LAST#'):format(color, rsrc.level)
  end

  local spd = 'normal'
  local ai_state = src.ai_state or {}
  if self.global_speed ~= 1 then
    spd = ('%+d%% normal'):format(math.floor(self.global_speed*100) - 100)
  end
  color = self.global_speed > 3 and 'RED' or
	  self.global_speed > 2 and 'ORANGE' or
	  self.global_speed > 1 and 'LIGHT_RED' or
	  self.global_speed < 1/3 and 'LIGHT_GREEN' or
	  self.global_speed < 0.5 and 'BLUE' or
	  self.global_speed < 1 and 'LIGHT_BLUE' or 'WHITE'
  desc = desc .. ('\n#{bold}#Speed:#{normal}# #%s#%s#LAST#'):format(color, spd)
  if ai_state.random_move or src.flags.NEVER_MOVE then
    local tt = {}
    if ai_state.random_move then
      tt[#tt+1] = ('%d%% random'):format(ai_state.random_move)
    end
    if src.flags.NEVER_MOVE then tt[#tt+1] = 'never-moving' end
    tt = table.concat(tt, ', ')
    desc = desc .. ' (' .. tt .. ')'
  end

  -- cf. know_armour() in src/monster1.c
  local know_ac = w_kills > 304/(4*rsrc.level) or self.unique and w_kills > 304/(38 + (5*rsrc.level)/4)
  if not use_learned or know_ac then
    local hp = ('%dd%d'):format(rsrc.hd[1], rsrc.hd[2])
    if raw.flags.FORCE_MAXHP then
      hp = ('%d (%s)'):format(rsrc.hd[1]*rsrc.hd[2], hp)
    end
    desc = desc .. '\n#{bold}#HP:#{normal}#  #LIGHT_GREEN#' .. hp .. '#LAST#'
    desc = desc .. '\n#{bold}#AC:#{normal}#  #LIGHT_GREEN#' .. rsrc.ac .. '#LAST#'
  end

  desc = desc .. self:field_list(src, '#{bold}#Resists:#{normal}# ', 'resist_dam', 'damage')
  desc = desc .. self:field_list(src, '#{bold}#Vulnerable to:#{normal}# ', 'vuln_dam', 'damage')

  desc = desc .. self:field_list(src, '#{bold}#Is:#{normal}# ', 'flags', 'is')

  if w_kills > 0 then
    local plyr_lev = game.player and game.player.level or 1
    local exp = rsrc.exp_worth * rsrc.level/plyr_lev
    desc = desc .. ('\n#{bold}#Exp. value:#{normal}# #ORANGE#%.2f#LAST# (for level-%d character)'):format(exp, plyr_lev)
  end

  desc = desc .. self:field_list(src, '#{bold}#Can:#{normal}# ', 'flags', 'can')

  if src.flags.ESCORT or src.flags.ESCORTS then
    desc = desc .. '\nUsually appears with escorts'
  elseif src.flags.FRIEND or src.flags.FRIENDS then
    desc = desc .. '\nUsually appears in groups'
  end

  local seen_slp = learned.sleep or { ignore=0, wake=0 }
  local know_slp = seen_slp.wake*seen_slp.wake > rsrc.alertness or
		   seen_slp.ignore >= 255 or
		   (rsrc.alertness == 0 and w_kills >= 10)
  if not use_learned or know_slp then
    local slp_desc = rsrc.alertness > 200 and 'Usually asleep' or
		     rsrc.alertness > 95 and 'Very sleepy' or
		     rsrc.alertness > 75 and 'Sleepy' or
		     rsrc.alertness > 45 and 'Fairly sleepy' or
		     rsrc.alertness > 25 and 'Moderately sleepy' or
		     rsrc.alertness > 10 and 'Moderately vigilant' or
		     rsrc.alertness > 5 and 'Fairly vigilant' or
		     rsrc.alertness > 3 and 'Vigilant' or
		     rsrc.alertness > 1 and 'Very vigilant' or
		     rsrc.alertness > 0 and 'Extremely vigilant' or 'Always awake'
    desc = desc .. ('\n#{bold}#Alertness:#{normal}# %s ("sleepiness" %d), sense range %d'):format(slp_desc, rsrc.alertness, rsrc.vision)
  end

  if src.drop_desc then
    local dd = src.drop_desc
    local n_max = math.max(dd.items, dd.gold)
    if n_max > 0 then
      local s = n_max == 1 and '' or s
      local what = {}
      if dd.items > 0 then what[#what] = 'object'..s end
      if dd.gold > 0 then what[#what] = 'treasure'..s end
      what = table.concat(what, ' or ')
      local kind = dd.great and 'exceptional' or dd.good and 'good ' or ''
      desc = desc .. ('\n#{bold}#Drops:#{normal}# up to %d %s%s'):format(n_max, kind, what)
    end
  end

  if src.attacks then
    local atk_descs = {}
    for _, atk in ipairs(src.attacks) do
      local atk_desc = self:getAttackDesc(atk)
      if atk_desc then table.insert(atk_descs, atk_desc) end
    end
    if #atk_descs > 0 then
      desc = desc .. '\n#{bold}#Attacks#{normal}#:  ' .. table.concat(atk_descs, ', ')
    end
  end

  if src.melee_retal then
    local tt = {}
    local l = rsrc.level
    for rtype, method in pairs(src.melee_retal) do
      local rf = util.colorize_damtype(rtype)
      local ds, dd
      if method == 'mondam' then
	dd, ds = 1 + math.floor(rsrc.level/26), 1 + math.floor(rsrc.level/17)
      elseif method == 'plyrdam' then
	dd, ds = 2, 6
      end
      if dd and ds then
	tt[#tt+1] = ('%dd%d %s'):format(dd, ds, rf)
      end
    end
    if #tt > 0 then
      desc = desc .. '\n#{bold}#Melee retal.#{normal}#:  ' .. table.concat(tt, ', ')
    end
  end

  -- TODO reflection

  if src.talents then
    local t_seen = learned.seen and learned.seen.talent or 0
    local n = raw.ai_state.talent_in or 1
    local rate = ('(1/%d) '):format(n)
    if use_learned and t_seen <= 100 then
      -- "Guess at the frequency" [cf. src/monster1.c lines 831-838]
      n = math.floor(100/n)
      n = math.floor((n + 9)/10) * 10
      n = math.floor(100/n)
      rate = ('(est. 1/%d) '):format(n)
    end
    local tt = {}
    for t_id, _ in pairs(src.talents) do
      local t = self:getTalentFromId(t_id)
      tt[#tt+1] = '#LIGHT_GREEN#' .. t.name .. '#LAST#'
    end
    if #tt > 0 then
      tt = table.concat(tt, ', ')
      desc = desc .. '\n#{bold}#Talents:#{normal}# ' .. rate .. tt
    end
  end

  if self.desc then desc = desc .. '\n\n#CYAN#' .. self.desc .. '#LAST#' end

  return desc
end

function _M:field_list(src, pfx, field_group, desc_group)
  if not src[field_group] then return '' end

  local out = {}
  if desc_group == 'damage' then
    -- Slight Hack(TM):  DamageTypes index by integer; walk them in defined
    -- order and check against our fields.
    for i, _ in pairs(DamageType.dam_def) do
      if src[field_group][i] then
	table.insert(out, util.colorize_damtype(i))
      end
    end
  else
    local fg = self.field_desc[desc_group]
    for _, f in ipairs(fg) do
      if src[field_group][f] then table.insert(out, fg[f]) end
    end
  end
  
  if #out == 0 then return '' end
  local ret = table.concat(out, ', ')
  ret = pfx and pfx .. ' ' .. ret or ret:capitalize()
  return '\n' .. ret
end

function _M:getAttackDesc(atk)
  if not atk or not atk.method then return nil end
  if atk.hit_count and atk.hit_count == 0 then return nil end
  local mthd = AttackType:get(atk.method)
  local desc = mthd.short_desc or '???'
  if atk.type then
    desc = desc .. ' for '
    if atk.dam then
      local ok = true
      if atk.hit_count then
	-- Based on know_damage() in src/monster1.c.  Roughly, the number
	-- of hits you need to know damage dice decrease with increasing
	-- monster level and increases with increasing max damage.
	local d = atk.dam[1] * atk.dam[2]
	local lc = (4 + (self.level or 0)) * atk.hit_count
	if self.unique then lc = lc * 2 end
	ok = lc > 80 * d
      end
      if ok then desc = desc .. ('%dd%d '):format(atk.dam[1], atk.dam[2]) end
    end
    desc = desc .. util.colorize_damtype(atk.type)
  end
  return desc
end

--[=[
function _M:tooltip()
  local str = mod.class.Actor.tooltip(self)
  return str..([[

  Target: %s
  UID: %d]]):format(
  self.ai_target.actor and self.ai_target.actor.name or "none",
  self.uid)
end
]=]

function _M:canSeeInvis(target)
  -- TODO Can NPCs see other invisible NPCs?  Assume yes for now.
  if not target.player then return true end

  -- cf. player_invis() in src/melee2.c
  local inv = target.invisibility or 0
  local mlv = self.level

  if self.quest_mon or self:has('INVISIBLE') then return true end

  if self:resistDam(DamageType.OLD_SLEEP) then mlv = mlv + 10 end
  if self:has('DRAGON') then mlv = mlv + 20 end
  if self:has('UNDEAD') then mlv = mlv + 15 end
  if self:has('DEMON') then mlv = mlv + 15 end
  if self:has('ANIMAL') then mlv = mlv + 15 end
  if self:has('ORC') then mlv = mlv - 15 end
  if self:has('TROLL') then mlv = mlv - 10 end
  if self:has('STUPID') then mlv = math.floor(mlv/2) end
  if self:has('SMART') then mlv = math.floor((mlv * 5)/4) end
  mlv = math.max(mlv, 1)
  return rng.range(1, mlv*2) >= inv
end

function _M:attackTarget(target)
  local p = { no_attack=false }
  self:fireCallbacks('on_melee_attack_pre', target, p)
  if p.no_attack then return end

  local blows, hits = 0, 0
  for i, atk in ipairs(self.attacks or {}) do
    -- If the target died (or we died from retaliatory damage), stop
    -- attacking.
    if self.dead or target.dead then break end
    blows = blows + 1
    local hit, dam, learn = AttackType:monAttackTarget(self, target, atk)
    if hit then hits = hits + 1 end
    self:learnFlags('attacks', i, learn or dam > 0)
  end
  if not self.dead and not target.dead then
    self:fireCallbacks('on_melee_attack', target, hits, blows)
  end
  self:useEnergy()
end

-- Sort of Elegant Hack(TM):  Override this to set the appropriate
-- :learnFlags() if the player sees us use this talent.
function _M:useTalent(id, who, force_level, ignore_cd, force_target, silent, no_confirm)
  local success = engine.interface.ActorTalents.useTalent(self, id, who, force_level, ignore_cd, force_target, silent, no_confirm)

  if success then self:learnFlags('talents', id) end
  return success
end

function _M:getExpChart(level)
  return level*level*level * 6
end

function _M:levelup(down)
  local newlev = self:levelupMsg(down)

  if not newlev then
    if rng.range(1, 10) <= 8 then
      -- Gain hp
      self.max_life = self.max_life + self.hd[2]
      self.life = self.life + self.hd[2]
    end
    if rng.range(1, 10) <= 4 then
      -- Gain speed
      self.speed = self.speed + rng.range(1, 2)
    end
    if rng.range(1, 10) <= 5 then
      -- Increase AC
      self.ac = self.ac + math.max(math.floor(self.ac/10), 1)
    end
    if rng.range(1, 10) <= 3 then
      -- Increase attack power
      for _ = 1, 20 do
	local atk = rng.table(self.attacks or {})
	if atk.dam then
	  atk.dam[1] = atk.dam[1] + 1
	  break
	end
      end
    end
  end

  return newlev
end

function _M:immuneDam(damtype, p)
  p = p or {}
  local dt = DamageType:get(damtype)
  if dt.is_immune then
    local imm, msg = dt:is_immune(self, p.learn)
    if imm then return imm, msg end
  end
  return false
end

function _M:resistDam(damtype, p)
  p = p or {}
  if p.or_immune and self:immuneDam(damtype, p) then return true end
  local dt = DamageType:get(damtype)
  if dt.is_resist then
    local res, msg = dt:is_resist(self, p.learn)
    if res then return res, msg end
  end

  if self.resist_dam and self.resist_dam[damtype] then
    if p.learn then self:learnFlags('resist_dam', damtype) end
    return true, ' resists.'
  end
  return false
end

function _M:vulnDam(damtype, p)
  p = p or {}
  local dt = DamageType:get(damtype)
  if dt.is_vulnerable then
    local vuln, msg = dt:is_vulnerable(self, p.learn)
    if vuln then return vuln, msg end
  end
  if self.vuln_dam and self.vuln_dam[damtype] then
    if p.learn then self:learnFlags('vuln_dam', damtype) end
    return true
  end
  return false
end

local amts = {
  ['60%']	= function() return rng.percent(60) and 1 or 0 end,
  ['90%']	= function() return rng.percent(90) and 1 or 0 end,
  ['1d2']	= function() return rng.dice(1, 2) end,
  ['2d2']	= function() return rng.dice(2, 2) end,
  ['3d2']	= function() return rng.dice(3, 2) end,
  ['4d2']	= function() return rng.dice(4, 2) end,
}
function _M:populateInitialInventory(t)
  local n_drop = 0
  for _, flag in ipairs(t) do
    if amts[flag] then
      n_drop = n_drop + amts[flag]()
    else
      print(("[ACTOR_INVEN] unrecognized drop flag '%s'"):format(flag))
    end
  end

  -- Average dungeon level and monster level.
  local lev = game.zone:level_adjust_level(game.level, 'object')
  local obj_level = math.floor((lev + self.level)/2)

  if t.randart then
    -- Create a randart.
    Object.resolve_flags.force_level = obj_level
    local o = obj_util.create_randart()
    if o then
      self:addObject(self.INVEN_INVEN, o)
      o:setFoundHere {
	type = 'mon_drop',
	mon_name = self:getName{ indef_art=true, no_given_name=true },
      }
    else
      print('[NPC] failed to create randart for '..self.name)
    end
  end

  local o
  for _ = 1, n_drop do
    if not t.only_item and (t.only_gold or rng.chance(2)) then
      local coin_type = t.coin_type or obj_util.choose_coin_type(obj_level)
      o = game.zone:makeEntityByName(game.level, 'object', coin_type)
    else
      -- TODO "Try hard for mimics" (cf. src/monster2.c line 2562).
      local RF = Object.resolve_flags
      RF.force_level = obj_level
      RF.force_good = t.good
      RF.force_great = t.great
      local filter = game.zone:createWeightedObjectFilter(self.drop)
      o = game.zone:makeEntity(game.level, 'object', filter, obj_level, true)
    end
    if o then
      self:addObject(self.INVEN_INVEN, o)
      o:setFoundHere {
	type = 'mon_drop',
	mon_name = self:getName{ indef_art=true, no_given_name=true },
      }
    else
      print('[NPC] failed to generate inventory object for '..self.name)
    end
  end
end
