-- $Id: Trap.lua 2835 2017-04-11 00:13:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
require 'engine.Trap'
require 'engine.interface.ObjectIdentify'
require 'engine.interface.ActorProject'
require 'mod.class.interface.Combat'
require 'mod.class.interface.ActorCallbacks'

module(..., package.seeall, class.inherit(engine.Trap,
					  engine.interface.ObjectIdentify,
					  engine.interface.ActorProject,
					  mod.class.interface.Combat,
					  mod.class.interface.ActorCallbacks))

function _M:init(t, no_default)
  engine.Trap.init(self, t, no_default)
  engine.interface.ObjectIdentify.init(self, t)
  engine.interface.ActorProject.init(self, t)
  mod.class.interface.ActorCallbacks.init(self, t)
end

--- Gets the full name of the object
function _M:getName(t)
  t = t or {}
  if self.trapkit then
    local name = self.trapkit:getName{no_count=true}
    if t.show_setter and self.setter and not self.trapkit.unique then
      name = self.setter:getName(t) .. "'s " .. name
    elseif t.def_art and not self.trapkit.unique then
      name = 'the ' .. name
    end
    return name
  end
  local name = self:isIdentified() and self.name or 'unknown trap'
  if t.def_art then name = 'the ' .. name end
  return name
end

-- A utility method for use in Map:checkAllEntities()
function _M:known_by(_, _, who)
  return self:knownBy(who)
end

-- A utility method for use in Map:checkAllEntities()
-- cf. spell.player_forget_map()
function _M:forget(_, _, who)
  self:setKnown(who, false)
  return nil
end

function _M:block_move(x, y, e, act, couldpass)
  -- For purposes of Astar or PlayerExplore path finding, assume traps
  -- don't block.
  if couldpass then return false end
  -- Store blocking will be clobbered in the store resolver anyway (q.v.).
  if self.is_store then return true end
  -- Block the player on known traps if "Automatically disarm traps" is
  -- disabled, unless we just failed to disarm said trap.
  -- [cf. Player:playerDisarm()]
  if not config.settings.tome2.auto_disarm then
    if e and e.player and self:knownBy(e) and not e.disarm_failure then
      return true
    end
  end
  -- We're good.
  return false
end

--- Returns a tooltip for the trap
function _M:tooltip()
  if self:knownBy(game.player) then
    local res = tstring{{"uid", self.uid}, self:getName()}
    if self.is_store then res:add(true, {"font","italic"}, "<Store>", {"font","normal"}) end
    if self.setter then
      res:add(true, {'font','bold'}, 'Set by: ', {'font','normal'}, self.setter:getName())
    end
    return res
  end
end

function _M:resolveSource()
  if self.setter then
    return self.setter.resolveSource and self.setter:resolveSource() or self.setter
  end
  return self
end

-- Convenience method to call a trap kit's num_shots() method.
function _M:numShots(max)
  return self.trapkit and self.trapkit:num_shots(max) or 1
end

-- Checks the trap specialization flags on player traps.  Returns true for
-- normal traps; normal handling should take care of what we need for those.
-- [cf. mon_hit_trap() in src/traps.c]
-- Intuitively, this should go in :canTrigger(), but we attempt to
-- :disarm() first, so we need it in both :canTrigger() and :canDisarm().
function _M:canAffect(x, y, who)
  if not self.player_trap then return true end
  -- PASS_WALL actors are only triggered by traps with the right flag.
  if who:has('PASS_WALL') and not self.trapkit:hasFlag('flags', 'KILL_GHOST') then
    return false
  end
  -- If no specialization flags, trigger for everything.
  if not self.trapkit.trap_trigger then return true end
  -- If it matches any flags, trigger for it.
  for f, _ in pairs(self.trapkit.trap_trigger) do
    if who:has(f) then return true end
  end
  -- Oh well...
  return false
end

function _M:canTrigger(x, y, who)
  if self.faction then
    -- Player-set traps are triggered by anything that is hostile to the
    -- player and matches the trap specialization flags, if any.
    return self:canAffect(x, y, who) and who.reactionToward and who:reactionToward(self) < 0
  else
    -- As per original T2, only the player can trigger non-player-set
    -- traps.
    return who.player
  end
end

function _M:canDisarm(x, y, who)
  -- Just in case...
  if self.is_store or not self.disarmable then return false end
  if not engine.Trap.canDisarm(self, x, y, who) then return false end

  if not who.skill_disarm then
    -- Monsters can only disarm player traps.
    if not self.player_trap then return false end
    local tk = self.trapkit

    -- cf. mon_hit_trap() in src/traps.c for all the following

    -- If the trap's specialization flags don't apply to us, don't try to
    -- disarm or trigger.
    if not self:canAffect(x, y, who) then return false end

    -- Chance of noticing the trap.
    local difficulty = 25
    if tk:hasFlag('pval_flags', 'CONCEAL') then
      difficulty = difficulty + 10*tk.pval
    end
    local smartness = who.level
    if who:has('SMART') then smartness = smartness + 10 end
    -- TODO Actors that have already spotted a player trap should get a
    -- bonus spotting subsequent ones.
    if who:has('STUPID') or who:has('EMPTY_MIND') then smartness = -150 end
    if rng.range(1, 300) > difficulty - smartness + 150 then
      self:setKnown(who, true)
      -- TODO Remember that we've seen a player trap.
    end
    
    if self:knownBy(who) then
      -- If we know about the trap, try to disarm it.
      difficulty = tk.armor + (tk.to_a or 0)
      smartness = math.floor(who.level/5)
      if who:has('SMART') then smartness = smartness * 2 end
      if who:has('STUPID') then smartness = -150 end
      if who:has('ANIMAL') and not who:has('SMART') then smartness = -150 end
      return rng.range(1, 120) > difficulty - smartness + 80
    else
      -- No disarm for you.
      return false
    end
  end

  -- The player can always disarm player-set traps.
  if self.player_trap then return true end

  -- If we failed while disarming a trap, we move onto it and trigger it.
  -- [cf. Player:playerDisarm()]
  if who.resting and who.resting.disarm_failure then return false end

  -- cf. do_cmd_disarm_aux() in src/cmd2.c
  local chance = math.max(who:getDisarmPower() - self.difficulty, 2)
  return rng.percent(chance)
end

-- Duplicated from engine.Trap with some extra flags to control behavior.
function _M:disarm(x, y, who, p)
  local Map = require 'mod.class.Map'
  p = p or {}
  -- Don't even try to disarm store "traps".
  if self.is_store then return false end
  if not p.force and not self:canDisarm(x, y, who) then
    if not p.silent then
      game.logSeen(who, "%s fails to disarm a trap (%s).", who.name:capitalize(), self:getName())
    end
    return false
  end
  game.level.map:remove(x, y, Map.TRAP)
  if self.removed then
    self:removed(x, y, who)
  end
  if not p.silent then
    game.logSeen(who, "%s disarms a trap (%s).", who.name:capitalize(), self:getName())
  end
  if not who.no_exp then
    who:gainExp(self.difficulty)
  end
  self:onDisarm(x, y, who)
  return true
end

function _M:onDisarm(x, y, who)
  self:check('on_disarm', x, y, who)
end

-- Duplicated from engine.Trap with modifications to allow for identifying
-- trap types.  We now expect traps' :triggered() method to return three
-- boolean values:  whether the trap is now known, whether the trap should
-- be removed after triggering, and whether the trap should be identified.
function _M:trigger(x, y, who, force_no_disarm)
  local Map = require 'mod.class.Map'
  -- Try to disarm, unless specifically told not to.
  if not force_no_disarm and self:knownBy(who) then
    if self:disarm(x, y, who) then
      return
    end
  end

  if not self:canTrigger(x, y, who) then return end

  if self.message == nil then
    game.logSeen(who, "%s triggers a trap (%s)!", who.name:capitalize(), self:getName())
  elseif self.message == false then
    -- Nothing
  else
    local tname = who.name
    local str =self.message
    str = str:gsub("@target@", tname)
    str = str:gsub("@Target@", tname:capitalize())
    game.logSeen(who, "%s", str)
  end
  local known, del, ident = false, false, false
  if self.summoner then self.summoner.__project_source = self end -- intermediate projector source
  if self.triggered then known, del, ident = self:triggered(x, y, who) end
  if self.summoner then self.summoner.__project_source = nil end
  if known then
    self:setKnown(who, true, x, y)
    game.level.map:updateMap(x, y)
  end
  if ident and who.player then
    self:identify(true)
    game.logPlayer(who, 'You identify the trap as %s.', self:getName())
  end
  if del then
    game.level.map:remove(x, y, Map.TRAP)
    if self.removed then self:removed(x, y, who) end
  end
end

-- Override this from ObjectIdentify's version, which tries to walk a
-- non-existent stack of objects.
function _M:identify(id)
  self.identified = id
end
