-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Zone = require 'engine.Zone'
local Object = require 'mod.class.Object'

module(..., package.seeall, class.inherit(Zone))

_M.debug = {
  rarities = false,
}

-- Bizarre and Annoying Hack(TM):  Apparently engine.Zone.npc_class and
-- mod.class.Zone.npc_class are separate variables, and Zone.setup only
-- sets the one in whichever Zone you required()'d.  We manually do them
-- both.
function _M:setup(t)
  Zone.setup(self, t)
  Zone.setup(Zone, t)
end

function _M:onLoadZoneFile(basedir)
  if not self.town and not self.wilderness and not self.worldmap then
    -- Remember some info about this zone for reference in other places
    -- without having to re-load zone.lua
    game.state.zone_info[self.short_name] = {
      name = self.name,
      level_range = table.clone(self.level_range)
    }
  end
  -- Questionable Hack(TM):  Some vaults depend on inner walls that can be
  -- tunneled through or otherwise passed.  We populate a 'diggable_inner_wall'
  -- terrain type that can be used for that purpose.
  -- [cf. data/maps/vaults/common.lua]
  local inner_wall, ok = table.get(self, 'generator', 'map', 'inner_wall')
  if ok then
    local diggables = {
      -- Currently this is the only inner_wall used that is not TUNNELABLE
      -- and can't be otherwise passed (like DEEP_LAVA or DARK_PIT).
      MOUNTAIN_CHAIN = 'DIGGABLE_MOUNTAIN_CHAIN',
    }
    self.generator.map.diggable_inner_wall = diggables[inner_wall] or inner_wall
  end
end

-- Load the artifacts list along with everything else.
function _M:loadBaseLists()
  Zone.loadBaseLists(self)
  if #self.object_list > 0 then
    self.artifacts_list = self.object_class:loadList('/data/general/objects/artifacts.lua', nil, { import_source=self.object_list })
    self.artifacts_list.import_source = nil
    self.artifacts_list.__real_type = 'object'
  end
end

function _M:updateBaseLevel()
  self.base_level = self.wild_base_level or self.level_range[1]
end

function _M:adjustComputeRaritiesLevel(level, type, lev)
  if type == 'actor' then
    -- Slight chance of OoD monsters.  Check twice for extra evil. >;)
    -- (cf. src/monster2.c, get_mon_num(), line 1017)
    if lev > 0 then
      if rng.chance(50) then lev = lev + math.min(5, math.floor(lev/4 + 2)) end
      if rng.chance(50) then lev = lev + math.min(5, math.floor(lev/4 + 2)) end
    end
  elseif type == 'object' then
    -- Slight chance of OoD objects.
    if lev > 0 then
      -- This is described in the T2 source as "What a bizarre calculation".
      if rng.chance(20) then lev = 1 + math.floor(lev * 128 / rng.range(1, 128)) end
    end
  end
  return lev
end

-- Dodgy Hack(TM):  Zone:computeRarities() doesn't really work the way we
-- want, so we replace it with our own Frankensteinian abomination. ^^;
-- We split out different implementations for different entity types.
function _M:computeRarities(type, list, level, filter, add_level, rarity_field)
  -- Slight Hack(TM):  Traps have probabilities rather than rarities.
  rarity_field = rarity_field or (type == 'trap' and 'probability') or "rarity"
  local r = { total=0 }
  if self.debug.rarities then print("******************", level.level, type) end

  local base_lev = self:level_adjust_level(level, type) + (add_level or 0)
  base_lev = Object.resolve_flags.force_level or base_lev
  local lev = self:adjustComputeRaritiesLevel(level, type, base_lev)

  if type == 'object' then
    -- Spectacularly Dodgy Hack(TM):  small chance to produce a "special"
    -- artifact [cf. src/object2.c lines 4837-4844].
    local prob = 1000
    local RF = Object.resolve_flags
    if RF.force_good or RF.force_great then
      prob = 10 - game.player:getLuckScale(-9, 9)
    end
    if rng.chance(prob) then r = self:trySpecialArtifact(lev) or r end
    if r.total == 0 then
      -- Special artifact failed or wasn't tried; do it the normal way.
      -- For objects, ignore the rarity field and use the allocation{}
      -- data.
      for _, o in ipairs(list) do
	if self:okToGenerate(o, type, level, base_lev, lev, filter, 'allocation') then
	  for _, grp in ipairs(o.allocation) do
	    if grp.level <= lev then
	      local genprob = math.floor(100 / grp.rarity)
	      if self.debug.rarities then print(("Entity(%30s) got %3d (=100 / %3d) chance to generate, current %2d"):format(o.name, math.floor(genprob), grp.rarity, lev)) end
	      if genprob > 0 then
		r.total = r.total + genprob
		-- This can get the object placed on the list several times
		-- for different levels.
		r[#r+1] = { e=o, genprob=r.total }
	      end
	    end
	  end
	end
      end
    end
  else
    -- For everything else, use the specified rarity field.
    for _, e in ipairs(list) do
      if self:okToGenerate(e, type, level, base_lev, lev, filter, rarity_field) then
	-- Slight Hack(TM), redux
	local genprob = type == 'trap' and e[rarity_field] or math.floor(100 / e[rarity_field])
	if self.debug.rarities then
	  print(("Entity(%30s) got %3d (=100 / %3d) chance to generate, current %2d"):format(e.name, math.floor(genprob), e[rarity_field], lev))
	end

	if genprob > 0 then
	  r.total = r.total + genprob
	  r[#r+1] = { e=e, genprob=r.total }
	end
      end
    end
  end

  table.sort(r, function(a, b) return a.genprob < b.genprob end)

  local prev = 0
  local tperc = 0
  for i, ee in ipairs(r) do
    local perc = 100 * (ee.genprob - prev) / r.total
    tperc = tperc + perc
    if self.debug.rarities then print(("entity chance %2d : chance(%4d : %4.5f%%): %s"):format(i, ee.genprob, perc, ee.e.name)) end
    prev = ee.genprob
  end
  if self.debug.rarities then print("*DONE", r.total, tperc.."%") end

  -- TODO We want to try again twice to get "tougher" monsters (cf.
  -- src/monster2.c, get_mon_num(), line 1122) and objects (cf.
  -- src/object2.c, get_obj_num(), line 767); not sure where to put that.
  return r
end

-- Spectacularly Dodgy Hack(TM) to produce "special" artifacts, the ones
-- with non-standard base objects that don't appear in normal rotation.
-- [cf. make_artifact_special() in src/object2.c]
function _M:trySpecialArtifact(lev)
  for _, a in ipairs(self.artifacts_list) do
    -- *sigh* My kingdom for a 'continue' keyword...
    local ok = a.insta_art and a.keep_base and not a.special_gene

    -- Already-created check
    local key = 'mod.class.Object/'..a.unique
    ok = ok and not game.uniques[key]

    -- OoD check
    if ok and a.level > lev then ok = rng.chance((a.level - lev)*2) end

    -- Artifact rarity check
    local adj_rarity = math.floor(a.rarity/2)
    adj_rarity = game.player:getLuckScale(-adj_rarity, adj_rarity)
    ok = ok and rng.chance(a.rarity - adj_rarity)

    -- Base object OoD check
    if ok then
      local o = self.object_list[a.keep_base]
      if o and o.level > lev then ok = rng.chance((o.level - lev)*5) end
    end

    if ok then
      -- All clear!  Place the artifact in the list (it'll be the only one,
      -- so Zone:pickEntity() will be forced to pick it.
      return { [1] = { e=a, genprob=100 }, total=100 }
    end
  end
end

function _M:okToGenerate(e, type, level, base_lev, max_ood_lev, filter, rarity_field)
  if type == 'actor' then
    -- Roughly mimics get_mon_num() in src/monster2.c.
    local force_depth = e.flags.FORCE_DEPTH or e.flags.ONLY_DEPTH
    return e[rarity_field] and e.level <= max_ood_lev and
	   (not force_depth or e.level == base_lev) and
	   (not filter or filter(e))
  elseif type == 'object' then
    -- Roughly mimics get_obj_num() in src/object2.c.
    if e[rarity_field] and not e.level then print('[GENERATOR] ??? '..e.define_as..' has no level') end
    -- Vaguely Annoying Hack(TM):  Filter out things not allowed in stores.
    local RF = Object.resolve_flags
    if RF.store_pop then
      if e.norm_art or e.insta_art then return false end
      if RF.store_min_level and e.level < RF.store_min_level then return false end
    end
    return e[rarity_field] and e.level <= max_ood_lev and
	   not e.special_gene and (not filter or filter(e))
  elseif type == 'trap' then
    -- Roughly mimics place_trap() in src/traps.c
    -- TODO check for quest levels here.
    local no_trap_door = level.level == self.max_level or self.generator.map.flags.FLAT
    return e[rarity_field] and e.min_level <= base_lev and
	   not (e.trap_door and no_trap_door) and
	   (not filter or filter(e))
  else
    -- From the original Zone:computeRarities()
    return e[rarity_field] and e.level_range and (not filter or filter(e))
  end
end

-- Vaguely Annoying Hack(TM):  Override Zone:makeEntityByName() to be able
-- to generate artifacts by name.
function _M:makeEntityByName(level, type, name, force_unique)
  if type == 'artifact' or (type == 'object' and not self.object_list[name] and self.artifacts_list[name]) then
    return Zone.makeEntityByName(self, level, self.artifacts_list, name, force_unique)
  end
  return Zone.makeEntityByName(self, level, type, name, force_unique)
end

-- Vaguely Annoying Hack(TM):  Clear __forced_level on the way into
-- Zone:addEntity(); it's set by Zone:makeEntity() from the 'add_levels'
-- field, which we don't want.
function _M:addEntity(level, e, typ, x, y, no_added)
  e.__forced_level = nil
  Zone.addEntity(self, level, e, typ, x, y, no_added)
end

-- Mostly duplicates the first part of Zone:makeEntity(), enough to choose
-- an entity template according to the usual rules.
function _M:chooseEntity(level, type, filter, force_level, prob_filter)
  resolvers.current_level = self.base_level + level.level - 1
  if force_level then resolvers.current_level = force_level end

  if prob_filter == nil then
    prob_filter = util.getval(self.default_prob_filter, self, type)
  end
  if filter == nil then
    filter = util.getval(self.default_filter, self, level, type)
  end
  if filter and self.alter_filter then
    filter = util.getval(self.alter_filter, self, level, type, filter)
  end

  local e = nil
  if not prob_filter then
    -- No probability list, use the default one and apply filter
    local list = self:getEntities(level, type)
    local tries = filter and filter.nb_tries or 500
    -- CRUDE ! Brute force ! Make me smarter !
    while tries > 0 do
      e = self:pickEntity(list)
      if e and self:checkFilter(e, filter, type) then break end
      tries = tries - 1
    end
    if tries == 0 then return nil end
  elseif filter then
    -- Generate a specific probability list, slower to generate but no need
    -- to "try and be lucky"
    local base_list = nil
    if filter.base_list then
      if _G.type(filter.base_list) == "table" then base_list = filter.base_list
      else
	local _, _, class, file = filter.base_list:find("(.*):(.*)")
	if class and file then
	  base_list = require(class):loadList(file)
	end
      end
    elseif type == "actor" then
      base_list = self.npc_list
    elseif type == "object" then
      base_list = self.object_list
    elseif type == "trap" then
      base_list = self.trap_list
    else
      base_list = self:getEntities(level, type)
      if not base_list then return nil end
    end
    local function filter_cb(e)
      return self:checkFilter(e, filter, type)
    end
    local list = self:computeRarities(type, base_list, level, filter_cb, filter.add_levels, filter.special_rarity)
    e = self:pickEntity(list)
    print("[MAKE ENTITY] prob list generation", e and e.name, "from list size", #list)
    if not e then return nil end
  else
    -- No filter
    local list = self:getEntities(level, type)
    local tries = filter and filter.nb_tries or 50 -- A little crude here too but we only check 50 times, this is simply to prevent duplicate uniques
    while tries > 0 do
      e = self:pickEntity(list)
      if e and self:checkFilter(e, nil, type) then break end
      tries = tries - 1
    end
    if tries == 0 then return nil end
  end

  return e
end

function _M:checkFilter(e, filter, type)
  -- If no filter, punt to the parent method to get its unique check.
  if not filter then return Zone.checkFilter(self, e, filter, type) end

  if filter.display and filter.display ~= e.display then return false end
  if filter.displays and not string.find(filter.displays, e.display, 1, true) then return false end

  local recurs_all_flags
  recurs_all_flags = function(ee, ff)
    if not ee then return false end
    for f, v in pairs(ff) do
      local vv = ee[f]
      if _G.type(v) == 'table' then vv = recurs_all_flags(ee[f], v) end
      if not vv then return false end
    end
    return true
  end
  if filter.deep_flags and not recurs_all_flags(e, filter.deep_flags) then return false end

  -- Most of the time we're dealing with entity templates, so we can't just
  -- use :knowTalent() to check for talents.  Instead scrape the template
  -- looking for a resolvers.talents() and check it manually.
  local function resolv_all_talents(tt)
    for _, v in ipairs(e) do
      if _G.type(v) == 'table' and v.__resolver == 'talents' then
	-- Found a resolvers.talents().  If it has all our requested
	-- talents, return true; if it's missing any, return false.
	for _, t_id in ipairs(tt) do
	  if not v[1][t_id] then return false end
	end
	return true
      end
    end
    -- Didn't find a resolvers.talents(); return false.
    return false
  end
  if filter.talents then
    if e.talents and e.knowTalent then
      -- If we're a resolved entity, we can actually check :knowTalent().
      for _, t_id in ipairs(filter.talents) do
	if not e:knowTalent(t_id) then return false end
      end
    else
      if not resolv_all_talents(filter.talents) then return false end
    end
  end

  -- These only work on resolve()'d corpses.
  if filter.corpse_id and e.mon_id ~= filter.corpse_id then return false end
  if filter.corpse_flags then
    local a = e.mon_id and game.zone.npc_list[e.mon_id]
    if not a or not recurs_all_flags(a, filter.corpse_flags) then return false end
  end
  if filter.corpse_talents then
    local a = e.mon_id and game.zone.npc_list[e.mon_id]
    if not a or not resolv_all_talents(a, filter.corpse_talents) then return false end
  end
  if filter.corpse_display then
    local a = e.mon_id and game.zone.npc_list[e.mon_id]
    if not a or a.display ~= filter.corpse_display then return false end
  end
  if filter.corpse_displays then
    local a = e.mon_id and game.zone.npc_list[e.mon_id]
    if not a or not string.find(filter.corpse_displays, a.display, 1, true) then return false end
  end

  -- If we cleared all that, fall through to the parent method.
  return Zone.checkFilter(self, e, filter, type)
end

-- cf. set_mon_num_hook() in src/monster1.c and related functions.
local wild_actor_flags = {
  TOWN = 'WILD_TOWN',
  DEEP_WATER = 'WILD_OCEAN',
  SHALLOW_WATER = 'WILD_SHORE',
  DIRT = 'WILD_WASTE',
  GRASS = 'WILD_GRASS',
  TREES = 'WILD_TREES',
  SHALLOW_LAVA = 'WILD_VOLCANO',
  DEEP_LAVA = 'WILD_VOLCANO',
  MOUNTAIN = 'WILD_VOLCANO',
}

function _M:createLevelActorFilter(level, extra)
  -- For future use (cf. src/monster2.c line 978)
  if not level.data.generator.actor or not level.data.generator.actor.filters or level.summon_bypass_filter then
    if extra and type(extra) == 'function' then return { special=extra } end
    return nil
  end

  local cb = function(e)
    if extra and type(extra) == 'function' and not extra(e) then return false end

    -- cf. set_mon_num_hook() in src/monster1.c and related functions.
    if self.town then
      -- Only WILD_TOWN monsters can be placed in town.
      if not e.placement_flags.WILD_TOWN then return false end
    elseif self.wilderness then
      -- Some wilderness terrain types have required flags also.
      local mx, my = game.player.wild_x, game.player.wild_y
      local g = game.state:getWorldGrid(mx, my)
      local flag = wild_actor_flags[g.terrain_type]
      if flag and not e.placement_flags[flag] then return false end
    else
      -- Dungeons and any other wilderness terrain types allow anything
      -- that doesn't have the WILD_ONLY flag.
      if e.placement_flags.WILD_ONLY then return false end
    end

    -- Pick a weighted random filter from the filters list
    local filters = level.data.generator.actor.filters
    local f, tot_prob = nil, 0
    for _, ff in ipairs(filters) do tot_prob = tot_prob + ff.prob end
    local p = rng.range(1, tot_prob)
    for _, ff in ipairs(filters) do
      if p <= ff.prob then f = ff; break end
      p = p - ff.prob
    end

    local mode, negate = f.mode, false
    if mode == 'any' then return true end
    if mode == 'nand' then mode, negate = 'and', true end
    if mode == 'nor' then mode, negate = 'or', true end
    local all = mode == 'and'

    for _, sym in ipairs(f.syms or {}) do
      if e.display == sym and not all then return not negate end
      if e.display ~= sym and all then return negate end
    end
    for _, flag in ipairs(f.flags or {}) do
      if e.flags[flag] and not all then return not negate end
      if not e.flags[flag] and all then return negate end
    end
    for _, talent in ipairs(f.talents or {}) do
      if e:knowTalent(talent) and not all then return not negate end
      if not e:knowTalent(talent) and all then return negate end
    end
    for _, dt in ipairs(f.resist_dam or {}) do
      if e:resistDam(dt) and not all then return not negate end
      if not e:resistDam(dt) and all then return negate end
    end
    if f.random_move then
      if e.ai_state.random_move and not all then return not negate end
      if not e.ai_state.random_move and all then return negate end
    end
    -- Didn't match anything; return false unless we're negating.
    return negate
  end
  return { special=cb }
end

function _M:createWeightedObjectFilter(wgts, merge)
  local m_spec = merge and merge.special
  local cb = function(e)
    if wgts then
      if type(wgts) == 'boolean' then
	wgts = game.level.data.generator.object.weights
      end
      if not e.category then print('[FILTER] ??? '..e.name..' has no treasure category') end
      if wgts[e.category] <= 0 or not rng.percent(wgts[e.category]) then return false end
    end
    if e.special_gene then return false end
    if e.type == 'misc' and e.subtype == 'corpse' then
      -- The T2 source code (cf. object2.c line 4663) manually filters out
      -- all the svals under TV_CORPSE except raw meat.  Beats me why they
      -- went about it backward...
      return e.name == 'RAW_MEAT'
    end
    if e.type == 'misc' and e.subtype == 'totem' then return false end

    -- Dodgy Hack(TM):  Forced good/great objects need extra filtering (cf.
    -- kind_is_good() in src/object2.c).  This doesn't appear to be applied
    -- to objects that are just randomly good/great (cf. make_object() in
    -- src/object2.c).
    local RF = Object.resolve_flags
    if (RF.force_good or RF.force_great) and not obj_util.good_kind(e) then
      return false
    end

    -- Vaguely Clumsy Hack(TM):  Similarly, objects that are going to have
    -- a randart built on top of them must be suitable for such.
    if RF.force_randart_base and not obj_util.randartable_kind(e) then
      return false
    end

    -- Slight Hack(TM):  Don't let a chest contain chests.
    if RF.force_no_chest and e.type == 'misc' and e.subtype == 'chest' then
      return false
    end

    if m_spec and m_spec(e) then return true end
    return true
  end

  local ret = merge and table.clone(merge, true, { special=true }) or {}
  ret.special = cb
  return ret
end

function _M:default_filter(level, type)
  if type == 'actor' then
    return self:createLevelActorFilter(level)
  elseif type == 'object' then
    return self:createWeightedObjectFilter(level.data.generator.object.weights)
  elseif type == 'trap' then
    -- By default we're checking for floor traps.
    return trap_util.default_filter
  else
    return nil
  end
end
