-- $Id: PlayerAlchemy.lua 3029 2017-08-16 03:24:16Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Object = require 'mod.class.Object'
local Ego = require 'mod.class.Ego'

module(..., package.seeall, class.make)

_M.alchemy_ego_recipe_def = { by_ego={} }
_M.alchemy_obj_recipe_def = { by_obj={}, by_base_obj={} }
_M.alchemy_flag_grp_def = {}
_M.alchemy_flag_recipe_def = { by_grp={} }

function _M:loadDefinition(file, env)
  env = env or {
    Alchemy = self,
    Talents = require 'engine.interface.ActorTalents',
    DamageType = require 'engine.DamageType',
    newAlchemyEgoRecipe = function(t) self:newAlchemyEgoRecipe(t) end,
    newAlchemyObjectRecipe = function(t) self:newAlchemyObjectRecipe(t) end,
    newAlchemyArtifactFlagGroup = function(t) self:newAlchemyArtifactFlagGroup(t) end,
    newAlchemyArtifactFlagRecipe = function(t) self:newAlchemyArtifactFlagRecipe(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then
    error(err)
    os.exit()
  end
  local ok, err = pcall(f)
  if not ok and err then error(err) end
end

function _M:newAlchemyEgoRecipe(t)
  assert(t.define_as, 'no ego recipe define_as')
  assert(not self.alchemy_ego_recipe_def[t.define_as], 'ego recipe define_as collision')
  assert(t.name, 'no ego recipe name')
  assert(t.ego_def, 'no ego recipe ego_def')
  assert(Ego:get(t.ego_def), 'undefined ego_def for ego recipe')
  assert(not self.alchemy_ego_recipe_def.by_ego[t.ego_def], 'ego recipe ego_def collision')
  assert(t.essences, 'no ego recipe essences')
  table.insert(self.alchemy_ego_recipe_def, t)
  self.alchemy_ego_recipe_def[t.define_as] = t
  self.alchemy_ego_recipe_def.by_ego[t.ego_def] = t
end

function _M:newAlchemyObjectRecipe(t)
  assert(t.define_as, 'no object recipe define_as')
  assert(not self.alchemy_obj_recipe_def[t.define_as], 'object recipe define_as collision')
  assert(t.name, 'no object recipe name')
  assert(t.obj_def, 'no object recipe obj_def')
  -- TODO How can we test obj_def corresponds to an actual object?  We're
  -- being called from load.lua, so we don't yet have a zone with an
  -- object_list{} to check.
  assert(not self.alchemy_obj_recipe_def.by_obj[t.obj_def], 'object recipe obj_def collision')
  assert(t.essences, 'no object recipe essences')
  table.insert(self.alchemy_obj_recipe_def, t)
  self.alchemy_obj_recipe_def[t.define_as] = t
  self.alchemy_obj_recipe_def.by_obj[t.obj_def] = t
  if t.base_obj then
    local by_base_obj = self.alchemy_obj_recipe_def.by_base_obj
    by_base_obj[t.base_obj] = by_base_obj[t.base_obj] or {}
    table.insert(by_base_obj[t.base_obj], t)
  end
end

function _M:newAlchemyArtifactFlagGroup(t)
  assert(t.tag, 'no artifact flag group tag')
  assert(not self.alchemy_flag_grp_def[t.tag], 'artifact flag group tag collision')
  assert(t.name, 'no artifact flag group name')
  table.insert(self.alchemy_flag_grp_def, t)
  self.alchemy_flag_grp_def[t.tag] = t
  self.alchemy_flag_recipe_def.by_grp[t.tag] = {}
end

function _M:newAlchemyArtifactFlagRecipe(t)
  assert(t.define_as, 'no flag recipe define_as')
  assert(not self.alchemy_flag_recipe_def[t.define_as], 'flag recipe define_as collision on ' .. t.define_as)
  assert(t.name, 'no flag recipe name')
  assert(t.flag_group, 'no flag recipe group')
  assert(self.alchemy_flag_grp_def[t.flag_group], 'unrecognized flag recipe group')
  assert(t.exp, 'no flag recipe exp')
  assert(t.skill, 'no flag recipe skill')
  assert(t.flags or t.copy or t.keep_open, 'no flag recipe flags or copy')
  table.insert(self.alchemy_flag_recipe_def, t)
  self.alchemy_flag_recipe_def[t.define_as] = t
  table.insert(self.alchemy_flag_recipe_def.by_grp[t.flag_group], t)
end

function _M:init(t)
  self.known_alchemy_recipes = { ego={}, object={}, flag={} }
end

function _M:getAlchemyRecipeById(id)
  return self.alchemy_ego_recipe_def[id] or
	 self.alchemy_obj_recipe_def[id] or
	 self.alchemy_ego_recipe_def[id]
end

function _M:knowsAlchemyRecipe(def)
  if type(def) == 'string' then def = self:getAlchemyRecipeById(def) end
  if not def then return false end
  local knowns = self.known_alchemy_recipes
  return (def.ego_def and knowns.ego[def.define_as]) or
	 (def.obj_def and knowns.object[def.define_as]) or
	 ((def.flags or def.copy or def.keep_open) and knowns.flag[def.define_as])
end

function _M:learnAlchemyRecipe(def, verbose)
  if type(def) == 'string' then def = self:getAlchemyRecipeById(def) end
  if not def or self:knowsAlchemyRecipe(def) then return end
  local knowns = self.known_alchemy_recipes
  local grp = (def.ego_def and 'ego') or
	      (def.obj_def and 'object') or
	      ((def.flags or def.copy or def.keep_open) and 'flag')
  if not grp or not knowns[grp] then return end
  knowns[grp][def.define_as] = true
  if verbose then
    local grp_name = grp == 'flag' and (def.flag_group == 'activate' and 'artifact activation' or def.keep_open and 'artifact' or 'artifact flag') or grp
    game.logPlayer(self, [[#LIGHT_BLUE#You learn %s recipe '%s'!#LAST#]], grp_name, def.name)
  end
  self.changed = true
end

function _M:alchemyLearnObject(o)
  if o == true then
    -- Special case to learn all artifact flag recipes that don't have a
    -- flag to match; mainly this will be activations, I think.
    for _, def in ipairs(self.alchemy_flag_recipe_def) do
      if not def.flags then self:learnAlchemyRecipe(def, true) end
    end
    return
  end
  -- Learn the recipe for the object itself, if any.
  local def = self.alchemy_obj_recipe_def.by_obj[o.define_as]
  if def then self:learnAlchemyRecipe(def, true) end
  if o:isArtifact() then
    -- Learn artifact flags.
    for _, def in ipairs(self.alchemy_flag_recipe_def) do
      local all = true
      for k, v in pairs(def.flags or {}) do
	if not o[k] then
	  all = false
	elseif k == 'lite' then
	  -- We assume that light radius 4 is LITE3|LITE1 and radius 5 is
	  -- LITE3|LITE2
	  if v == 3 and o.lite < 3 or v < 3 and (o.lite % 3) ~= v then
	    all = false
	  end
	elseif type(v) == 'table' then
	  for kk, _ in pairs(v) do
	    if not o[k][kk] then all = false end
	  end
	end
      end
      if all then self:learnAlchemyRecipe(def, true) end
    end
    return
  end
  -- Learn egos.
  for _, tag in pairs(o.ego_names or {}) do
    def = self.alchemy_ego_recipe_def.by_ego[tag]
    if def then self:learnAlchemyRecipe(def, true) end
  end
end

function _M:hasForgeItemRecipes(o)
  local recipes = self.alchemy_obj_recipe_def.by_base_obj[o.define_as]
  if not recipes then return false end
  for _, def in ipairs(recipes) do
    if self:knowsAlchemyRecipe(def) then return true end
  end
  return false
end

function _M:getForgeItemRecipes(o)
  local ret = {}
  local recipes = self.alchemy_obj_recipe_def.by_base_obj[o.define_as]
  for _, def in ipairs(recipes or {}) do
    if self:knowsAlchemyRecipe(def) then ret[#ret+1] = def end
  end
  return ret
end

-- Static
function _M:canApplyEgoRecipeTo(def, o)
  local ego = def.ego_def and Ego:get(def.ego_def)
  return ego and ego:canApplyTo(o)
end

function _M:hasForgeEgoRecipes(o)
  for _, def in ipairs(self.alchemy_ego_recipe_def) do
    if self:knowsAlchemyRecipe(def) and self:canApplyEgoRecipeTo(def, o) then
      return true
    end
  end
  return false
end

function _M:getForgeEgoRecipes(o)
  local ret = {}
  for _, def in ipairs(self.alchemy_ego_recipe_def) do
    if self:knowsAlchemyRecipe(def) and self:canApplyEgoRecipeTo(def, o) then
      ret[#ret+1] = def
    end
  end
  return ret
end

-- Returns an ego or object alchemy recipe suitable for extracting essences
-- from this object.  Ego recipes are checked before object recipes, except
-- for non-empty wands and staffs.  If 'known' is true, the recipe must be
-- known by the player.
function _M:getExtractRecipeFor(o, known, force_recharge)
  -- Can't extract anything from artifacts.
  if o:isArtifact() then return nil end
  -- Or from cursed objects.
  if o.cursed then return nil end
  -- Detach rod tips before extracting ego from rod base.
  if o.rod_tip then return nil end

  local knowns = self.known_alchemy_recipes
  -- Extract stick charges before stick ego.  Don't allow extracting
  -- charges from recharged stick, to prevent infinite essence farming.
  if o.power and o.power > 0 and (force_recharge or not o.recharged) and
     o.type == 'utility' and
     ((o.subtype == 'wand' and o.define_as ~= 'WAND_NOTHING') or
      (o.subtype == 'staff' and o.define_as ~= 'STAFF_NOTHING'))
  then
    local def = self.alchemy_obj_recipe_def.by_obj[o.define_as]
    if def and (not known or knowns.object[def.define_as]) then return def end
  end
  if o.ego_names then
    local pfx, sfx = o.ego_names.prefix, o.ego_names.suffix
    local def = pfx and self.alchemy_ego_recipe_def.by_ego[pfx]
    if def and (not known or knowns.ego[def.define_as]) then return def end
    def = sfx and self.alchemy_ego_recipe_def.by_ego[sfx]
    if def and (not known or knowns.ego[def.define_as]) then return def end
  end
  local def = self.alchemy_obj_recipe_def.by_obj[o.define_as]
  if def and (not known or knowns.object[def.define_as]) then return def end
  return nil
end

-- cf. src/cmd7.c lines 2735-2741
local function maybe_lose_essences(n, p)
  if rng.chance(3) and rng.range(1, 52) > p:getSkillLevel('ALCHEMY') then
    return math.floor(n/rng.range(2, 3))
  end
  return n
end

-- Extract essences from the object.  'learn' can be true to force learning
-- of the recipe, false to prevent learning, or nil (the default) to learn
-- one time in three.
function _M:extractEssences(o, n, inven, item, floor, learn)
  local recipe = self:getExtractRecipeFor(o)
  if not recipe then return false end

  -- Slight Hack(TM):  Extracting essences from an ego'd item extracts all
  -- egos, getting essences from both prefix and suffix recipes if
  -- applicable.
  local r_pfx, r_sfx = nil, nil
  if recipe.ego_def and o.ego_names then
    local e_pfx, e_sfx = o.ego_names.prefix, o.ego_names.suffix
    r_pfx = e_pfx and self.alchemy_ego_recipe_def.by_ego[e_pfx]
    r_sfx = e_sfx and self.alchemy_ego_recipe_def.by_ego[e_sfx]
  end

  if not self:knowsAlchemyRecipe(recipe) and (learn or type(learn) == 'nil' and rng.chance(3)) then
    self:learnAlchemyRecipe(recipe, true)
    if r_pfx then self:learnAlchemyRecipe(r_pfx, true) end
    if r_sfx then self:learnAlchemyRecipe(r_sfx, true) end
  end
  -- Always learn what kind of thing it is.
  o:identify(true, self)

  local base_name, nbase, essences = nil, 0, {}
  if recipe.charges then
    local nn = o:extract_charges(n, self)
    if nn > 0 then
      for ess, num in pairs(recipe.essences or {}) do
	for _ = 1, nn do
	  essences[ess] = (essences[ess] or 0) + maybe_lose_essences(num, self)
	end
      end
    end
  elseif recipe.obj_def then
    base_name = recipe.base_obj
    for i = 1, n do
      local xo = self:getOneObject(inven, item, floor, {vocal_left=(i==n)})
      if game:hasEntity(xo) then game:removeEntity(xo) end
      if recipe.base_obj then nbase = nbase + 1 end
      for ess, num in pairs(recipe.essences or {}) do
	essences[ess] = (essences[ess] or 0) + maybe_lose_essences(num, self)
      end
    end
  elseif recipe.ego_def then
    base_name = o.define_as
    for i = 1, n do
      local xo = self:getOneObject(inven, item, floor, {vocal_left=(i==n)})
      if game:hasEntity(xo) then game:removeEntity(xo) end
      nbase = nbase + 1
      for ess, num in pairs(r_pfx and r_pfx.essences or {}) do
	essences[ess] = (essences[ess] or 0) + maybe_lose_essences(num, self)
      end
      for ess, num in pairs(r_sfx and r_sfx.essences or {}) do
	essences[ess] = (essences[ess] or 0) + maybe_lose_essences(num, self)
      end
    end
  end

  if base_name and nbase > 0 then
    local Object = require 'mod.class.Object'
    Object.resolve_flags.force_power = 0
    local base = game.zone:makeEntityByName(game.level, 'object', base_name)
    base:setNumber(nbase)
    self:addOrDropObject(self.INVEN_INVEN, base, nil, nil, {vocal_add=true, vocal_drop=true})
  end
  for ess, num in pairs(essences) do
    if num > 0 then
      local ess_o = game.zone:makeEntityByName(game.level, 'object', ess)
      ess_o:setNumber(num)
      self:addOrDropObject(self.INVEN_INVEN, ess_o, nil, nil, {vocal_add=true, vocal_drop=true})
    end
  end

  self:sortInven()
  util.live_update_inventory_dialogs()
  self:recalcEverything()
  self.changed = true
end

-- NOTE:  This method must be called in a coroutine context.
function _M:playerExtract(o, inven, item, floor)
  local recipe = self:getExtractRecipeFor(o)
  if not recipe then
    game.logPlayer(self, [[You don't know how to extract anything from that.]])
    return false
  end
  if recipe.charges and o.recharged then
    game.logPlayer(self, [[This item has been magically recharged; you can no longer extract its charges for essences.]])
    return false
  end

  local n, max, prompt = 1, 1, 'Number to extract:'
  local charges = o.type == 'utility' and (o.subtype == 'wand' or o.subtype == 'staff')
  if charges then
    max, prompt = o.power, 'Charges to extract:'
  else
    max = o:getNumber()
  end
  if max > 1 then
    local function n_cb(qty) self:talentDialogReturn(qty) end
    n = self:talentDialog(ui_util.get_quantity('Extract Essences', prompt, max, max, n_cb))
    if not n or n <= 0 then return false end
  end

  self:extractEssences(o, n, inven, item, floor)
  return true
end

-- Returns the number of times the essences in the provided recipe can be
-- removed from inventory.  This will place an upper bound on the number of
-- charges that can be added to a wand/staff, for instance.
function _M:numRecipeEssences(def)
  if type(def) == 'string' then def = self:getAlchemyRecipeById(def) end
  if not def or not def.essences or not next(def.essences) then return 0 end

  local max = 9999
  local inven = self:getInven('INVEN')
  for tag, num in pairs(def.essences) do
    local ess = self:findInInventoryBy(inven, 'define_as', tag)
    if not ess then return 0 end
    max = math.min(max, math.floor(ess:getNumber()/num))
  end

  return max
end

-- Remove from inventory the essences needed by this recipe.  'rpt' can be
-- specified to remove a multiple of the recipe's essences (convenient for,
-- for instance, adding multiple charges to a wand/staff).
function _M:useRecipeEssences(def, rpt)
  rpt = rpt or 1
  local inven = self:getInven('INVEN')
  for tag, num in pairs(def.essences) do
    local ess, item = self:findInInventoryBy(inven, 'define_as', tag)
    if not ess or not item then return end
    local n = math.min(num*rpt, ess:getNumber())
    local xo = self:getOneObject(inven, item, false, {num=n, vocal_left=true})
    if game:hasEntity(xo) then game:removeEntity(xo) end
  end
end

-- NOTE:  This method must be called in a coroutine context.
function _M:playerRecharge(o, inven, item)
  if not o.add_charges then
    game.logPlayer(self, [[This item can't be recharged with essences (how the heck did you get here?]])
    return false
  end

  local recipe = self:getExtractRecipeFor(o)
  if not recipe then
    game.logPlayer(self, [[You don't know how to recharge this item.]])
    return false
  end

  local n, max = 1, self:numRecipeEssences(recipe)
  if max == 0 then
    game.logPlayer(self, [[You don't have enough essences to recharge this item.]])
    return false
  end
  if max > 1 then
    local function n_cb(qty) self:talentDialogReturn(qty) end
    n = self:talentDialog(ui_util.get_quantity('Recharge', 'Number of charges to add', max, max, n_cb))
    if not n or n <= 0 then return false end
  end

  n = o:add_charges(n, self)
  if n > 0 then self:useRecipeEssences(recipe, n) end
  self:sortInven()
  util.live_update_inventory_dialogs()
  self:recalcEverything()
  self.changed = true
  return true
end

-- NOTE:  This method must be called in a coroutine context.
function _M:playerForgeItem(o, inven, item)
  local recipes = self:getForgeItemRecipes(o)
  if not recipes or #recipes == 0 then
    local that = o:getNumber() > 1 and 'those' or 'that'
    game.logPlayer(self, [[You don't know how to make anything out of %s.]], that)
    return false
  end

  local AlchemyRecipesDialog = require 'mod.dialogs.AlchemyRecipesDialog'
  local recipe = self:talentDialog(AlchemyRecipesDialog.new(self, o, recipes))
  if not recipe or not recipe.obj_def then return false end

  local max = self:numRecipeEssences(recipe)
  if max <= 0 then
    game.logPlayer(self, [[You don't have enough essences to make that.]])
    return false
  end
  max = math.min(max, o:getNumber())

  local n = 1
  if max > 1 then
    local function n_cb(qty) self:talentDialogReturn(qty) end
    n = self:talentDialog(ui_util.get_quantity('Forge Item', 'Items to make', 1, max, n_cb))
    if not n or n <= 0 then return false end
  end

  local tmpl = game.zone.object_list[recipe.obj_def]
  local value = tmpl:getPrice()
  local failprob = math.min(10 * ((tmpl.level or 0) - 2*self:getSkillLevel('ALCHEMY')), 100)
  -- Reduced fail chance for Potions of Detonations
  if recipe.obj_def == 'POT_DETONATIONS' then
    failprob = math.floor(failprob/10)
  end

  if failprob > 0 and value > 0 then
    local AlchemyGoldPopup = require 'mod.dialogs.AlchemyGoldPopup'
    local gold = self:talentDialog(AlchemyGoldPopup.new(self, 'Forge Item', failprob, value, false))
    if not gold or gold < 0 then
      game.logPlayer(self, 'Forge Item canceled.')
      return false
    end
    if gold > 0 then
      self:adjGold(-gold)
      failprob = math.max(0, failprob - math.floor((gold * 20)/value))
      game.logPlayer(self, '#LIGHT_BLUE#Adding %d gold reduces the chance of failure to %d%%.#LAST#', gold, failprob)
    end
  end

  if failprob > 0 and rng.percent(failprob) then
    -- Oopsie...
    local DamageType = require 'engine.DamageType'
    self:useRecipeEssences(recipe, n)
    local sides = math.max(1, (tmpl.level or 1) - 2*self:getSkillLevel('ALCHEMY'))
    local dummy_src = { name=self:getName().. [['s Forge Item]] }
    local t = self:getTalentFromId(self.T_FORGE_ITEM)
    spell.hit_spell(dummy_src, t, DamageType.PHYSICAL, rng.dice(3, sides), self.x, self.y, 1)
    return true
  end

  local Object = require 'mod.class.Object'
  for _ = 1, n do
    self:useRecipeEssences(recipe)
    -- Inventory has changed, so our 'item' parameter is now potentially
    -- invalid (unless we're working from the floor..
    local ii = item
    if not floor then _, ii = self:findInInventoryByObject(inven, o) end
    if not ii then break end
    local xo = self:getOneObject(inven, ii, floor, {vocal_left=true})
    if game:hasEntity(xo) then game:removeEntity(xo) end
    Object.resolve_flags.force_power = 0
    local fo = game.zone:makeEntityByName(game.level, 'object', recipe.obj_def)
    fo:setFoundHere{ type='obj_create', source='Alchemy' }
    self:addOrDropObject(self.INVEN_INVEN, fo, nil, nil, {vocal_add=true, vocal_drop=true})
  end
  self:sortInven()
  self:recalcEverything()
  self.changed = true
  return true
end

-- NOTE:  This method must be called in a coroutine context.
function _M:playerForgeEgo(o, inven, item)
  local AlchemyRecipesDialog = require 'mod.dialogs.AlchemyRecipesDialog'

  local recipes = self:getForgeEgoRecipes(o)
  if not recipes or #recipes == 0 then
    local that = o:getNumber() > 1 and 'those' or 'that'
    game.logPlayer(self, [[You don't know any egos you can add to %s.]], that)
    return false
  end

  local recipe = self:talentDialog(AlchemyRecipesDialog.new(self, o, recipes))
  if not recipe or not recipe.ego_def or not Ego:get(recipe.ego_def) then
    return false
  end

  local n, max = 1, self:numRecipeEssences(recipe)
  if max <= 0 then
    game.logPlayer(self, [[You don't have enough essences to add that ego.]])
    return false
  end
  max = math.min(max, o:getNumber())
  if max > 1 then
    local function n_cb(qty) self:talentDialogReturn(qty) end
    n = self:talentDialog(ui_util.get_quantity('Add Ego to Item', 'Number to convert', max, max, n_cb))
    if not n or n <= 0 then return false end
  end

  local ego = Ego:get(recipe.ego_def)
  local value = math.min(ego.cost, 50000)
  if o.ego_names and next(o.ego_names) then
    local side, tag = next(o.ego_names)
    local has_ego = Ego:get(tag)
    if has_ego then value = value + math.min(has_ego.cost, 50000) end
  end
  local failprob = util.bound(10*(math.floor(value/1000) + 5 - self:getSkillLevel('ALCHEMY', 100)), 0, 100)

  if failprob > 0 and value > 0 then
    local AlchemyGoldPopup = require 'mod.dialogs.AlchemyGoldPopup'
    local gold = self:talentDialog(AlchemyGoldPopup.new(self, 'Add Ego to Item', failprob, value, true))
    if not gold or gold < 0 then
      game.logPlayer(self, 'Add Ego to Item canceled.')
      return false
    end
    if gold > 0 then
      self:adjGold(-gold)
      failprob = math.max(0, failprob - math.floor((gold * 20)/value))
      game.logPlayer(self, '#LIGHT_BLUE#Adding %d gold reduces the chance of failure to %d%%.#LAST#', gold, failprob)
    end
  end

  if failprob > 0 and rng.percent(failprob) then
    -- Oopsie...
    local DamageType = require 'engine.DamageType'
    self:useRecipeEssences(recipe, n)
    local lvl = (o.level or 1) + (ego.level or 1)
    local sides = math.max(1, sides - 2*self:getSkillLevel('ALCHEMY'))
    local dummy_src = { name=self:getName().. [['s Add Ego to Item]] }
    local t = self:getTalentFromId(self.T_FORGE_ITEM)
    spell.hit_spell(dummy_src, t, DamageType.PHYSICAL, rng.dice(3, sides), self.x, self.y, 1)
    return true
  end

  local xo = self:getOneObject(inven, item, false, {num=n, vocal_left=true})
  self:useRecipeEssences(recipe, n)
  -- Sneaky Hack(TM):  I strongly suspect ego application doesn't properly
  -- work on stacks of objects, so we reduce our removed objects to one,
  -- apply the new ego, and expand back to the right number of objects.
  xo:setNumber(1)
  local RF = Object.resolve_flags
  RF.force_ego = ego.define_as
  RF.force_level = self:getSkillLevel('ALCHEMY', 100)
  table.insert(xo, self:getEgoResolver(xo))
  xo:resolve()
  xo:resolve(nil, true)
  xo:setNumber(n)
  xo:setFoundHere{ type='obj_create', source='Alchemy' }
  self:addOrDropObject(self.INVEN_INVEN, xo, nil, nil, {vocal_add=true, vocal_drop=true})
  self:sortInven()
  self:recalcEverything()
  self.changed = true
  return true
end

-- We define this here so that addons can superload it as needed for their
-- own objects.
function _M:getEgoResolver(o)
  if o.type == 'weapon' or o.type == 'ammo' or o.subtype == 'boomerang' then
    return resolvers.weapon()
  end
  if o.type == 'armor' then return resolvers.armor() end
  return resolvers.ego()
end

function _M:startBecoming(base_obj)
  local n_ess = self:getSkillLevel('ALCHEMY')
  local inv = self:getInven('INVEN')
  local ess, ess_item = self:findInInventoryBy(inv, 'define_as', 'ESS_MAGIC')
  if not ess or ess:getNumber() < n_ess then
    Dialog:simplePopup('Forge Artifact', ('You do not have the necessary %s Essences of Magic.'):format(n_ess))
    return false
  end

  local msg = [[#{bold}#WARNING:#{normal}#  Creating an artifact with this talent will #{bold}#permanently#{normal}# reduce your max life by 10.
Proceed?]]
  local function cb(ok) self:talentDialogReturn(ok) end
  local ok = self:talentDialog(ui_util.yesnoLongPopup('Forge Artifact', msg, 600, cb))
  if not ok then return false end

  self.changed = true
  self.hp_mod = self.hp_mod - 10
  self:getOneObject(inv, ess_item, false, {num=n_ess, vocal_left=true})
  base_obj.art_exp = 0
  base_obj.ego_names = nil
  base_obj.artifact_name = 'Becoming'
  self:sortInven()
  self:recalcEverything()
end
