-- $Id: PlayerFates.lua 2835 2017-04-11 00:13:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'

module(..., package.seeall, class.make)

-- This roughly duplicates the pre-resolving part of Zone:makeEntity() with
-- prob_filter=true.
local function pick_entity_target(f, type, base_list, filter, name_flags)
  local cb = function(e) return game.zone:checkFilter(e, filter, type) end
  -- Amazingly Dodgy Hack(TM):  Zone:computeRarities() expects an actual
  -- Level object; I think we can fake it with this.
  local fake_level = {
    level = f.level,
    getEntitiesList = function(x) return {} end,
  }
  local list = game.zone:computeRarities(type, base_list, fake_level, cb, filter.add_levels, filter.special_rarity)
  for _ = 1, 50 do
    local e = game.zone:pickEntity(list)
    if e then
      f.target_type = type
      f.target_def = e.define_as
      f.target_desc = e:getName(name_flags)
      return true
    end
  end
  print("[FATE] Couldn't find an "..type.." to generate")
  return false
end

_M.fate_types = {
  find_obj = {
    maybe = true,
    pick_target = function(f)
      local filter = { not_properties = { special_gene, insta_art, norm_art } }
      filter = game.zone:createWeightedObjectFilter(nil, filter)
      local name_flags = { force_qty=1, no_mods=true, no_pval=true }
      return pick_entity_target(f, 'object', game.zone.object_list, filter, name_flags)
    end,
    apply = function(f)
      local o = game.zone:makeEntityByName(game.level, "object", f.target_def)
      o:setFoundHere {
	type = 'custom_with_level',
	custom = 'You were fated to find %s',
      }
      local function ok_cb(x, y)
	return game.level.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold')
      end
      local grids = spell.grids_near(game.player.x, game.player.y, 10, 5, ok_cb, true)
      if #grids > 0 then
	local gxy = rng.table(grids)
	game.zone:addEntity(game.level, o, 'object', gxy.x, gxy.y)
      end
    end
  },
  find_mon = {
    maybe = true,
    pick_target = function(f)
      return pick_entity_target(f, 'actor', game.zone.npc_list, {}, { indef_art=true })
    end,
    apply = function(f)
      local m = game.zone:makeEntityByName(game.level, "actor", f.target_def)
      local function ok_cb(x, y)
	local g = game.level.map(x, y, Map.TERRAIN)
	return game.level.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold') and
	       -- TODO also reject altars and glyphs of warding
	       g.define_as ~= 'VOID_JUMPGATE' and
	       m:canMove(x, y)
      end
      local grids = spell.grids_near(game.player.x, game.player.y, 10, 5, ok_cb, true)
      if #grids > 0 then
	local gxy = rng.table(grids)
	game.zone:addEntity(game.level, m, 'actor', gxy.x, gxy.y)
      end
    end,
  },
  find_art = {
    maybe = true,
    pick_target = function(f)
      f.target_desc = 'an artifact'
      return true
    end,
    apply = function(f)
      local o = obj_util.create_randart()
      if o then
	o:setFoundHere {
	  type = 'custom_with_level',
	  custom = 'You were fated to find %s',
	}
	local function ok_cb(x, y)
	  return game.level.map:checkEntity(x, y, Map.TERRAIN, 'empty_bold')
	end
	local grids = spell.grids_near(game.player.x, game.player.y, 10, 5, ok_cb, true)
	if #grids > 0 then
	  local gxy = rng.table(grids)
	  game.zone:addEntity(game.level, o, 'object', gxy.x, gxy.y)
	end
      end
    end
  },
  no_die_mortal = {
    no_level = true,
    desc = function(f)
      return 'You are fated never to die by the hand of a mortal being.'
    end,
  },
  die = {
    desc = function(f)
      return 'You are fated to die on level '..f.level..'.'
    end,
  },
}

function _M:init(t)
  self.fates = {}
end

function _M:gainFate()
  -- If currently lucky, don't get a death fate.
  local i = rng.range(1, self:getLuck() > 0 and 17 or 18)
  local f = {}
  if i <= 7 then
    f.act = 'find_obj'
  elseif i <= 14 then
    f.act = 'find_mon'
  elseif i <= 16 then
    f.act = 'find_art'
  elseif i == 17 then
    f.act = 'no_die_mortal'
  elseif i == 18 then
    f.act = 'die'
  end
  local def = self.fate_types[f.act]
  if def.maybe then f.maybe = rng.percent(50) end
  if not def.no_level then
    local zone_max = game.zone.max_level
    local level = rng.range(zone_max - 20, zone_max + 20)
    f.level = util.bound(level, 1, 98)
  end
  if def.pick_target and not def.pick_target(f) then return end
  table.insert(self.fates, f)
  game.log('#VIOLET#More of your prophecy has been unearthed!#LAST#')
  game.log('#VIOLET#You should see a soothsayer quickly.#LAST#')
end

function _M:descFate(f)
  local def = self.fate_types[f.act]
  if def.desc then return def.desc(f) end

  local maybe = f.maybe and 'may' or 'are fated to'
  return 'You '..maybe..' find '..f.target_desc..' on level '..f.level..'.'
end

function _M:hasUnknownFates()
  for _, f in ipairs(self.fates) do
    if not f.known then return true end
  end
  return false
end

function _M:learnUnknownFate()
  local unknown = {}
  for _, f in ipairs(self.fates) do
    if not f.known then table.insert(unknown, f) end
  end
  if #unknown == 0 then return nil end
  local f = unknown[rng.range(1, #unknown)]
  f.known = true
  return self:descFate(f)
end

function _M:dumpFates()
  local fates = { any_unknown = false }
  if #self.fates > 0 then
    for _, f in ipairs(self.fates) do
      if f.known then
	table.insert(fates, self:descFate(f))
      else
	fates.any_unknown = true
      end
    end
  end
  return fates
end

function _M:showFates()
  local text = 'You do not currently have any fates.'
  local fates = self:dumpFates()
  if #fates > 0 then text = '• ' .. table.concat(fates, '\n• ') end
  if #fates == 0 and fates.any_unknown then text = nil end
  if fates.any_unknown then
    text = (text and (text.."\n\n") or '') .. '#YELLOW#Some of your fate is still unknown.  You should visit a soothsayer in town to learn about it.#LAST#'
  end

  Dialog:simpleLongPopup('Fates:  '..self.name, text, game.w * 0.5)
end

function _M:checkFates(zone, level)
  local lev = zone:level_adjust_level(level)
  local pending = {}
  for _, f in ipairs(self.fates) do
    if not f.no_level and f.level == lev and (not f.maybe or rng.chance(2)) then
      if def[f.act].apply then def[f.act].apply(f) end
    else
      pending[#pending+1] = f
    end
  end
  if #pending < #self.fates then
    game.log('#VIOLET#You feel that you will meet your fate here#LAST#')
  end
end
