-- $Id: AlchemyArtifactDialog.lua 3041 2017-08-30 02:40:05Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Object = require 'mod.class.Object'
local Dialog = require 'engine.ui.Dialog'
local ListColumns = require 'engine.ui.ListColumns'
local Button = require 'engine.ui.Button'
local Checkbox = require 'engine.ui.Checkbox'
local Numberbox = require 'engine.ui.Numberbox'
local Textzone = require 'engine.ui.Textzone'
local Separator = require 'engine.ui.Separator'
local StatusBox = require 'mod.dialogs.ui.StatusBox'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor, base_obj, state)
  self.actor = actor
  self.base_obj = base_obj
  local init_pval
  if state then
    self.tmpl_obj = state.tmpl_obj
    self.selected = state.selected
    self.ingr_pulled = state.ingr_pulled
    init_pval = state.pval
  else
    self.tmpl_obj = Object.new()
    table.merge(self.tmpl_obj, self.base_obj, true, {uid=true})
    self.selected = {}
    self.ingr_pulled = {}
    init_pval = self.base_obj.pval or 1
  end

  local title = self.base_obj:getName{no_mods=true, no_pval=true}
  Dialog.init(self, title, game.w*0.8, game.h*0.8)

  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function()
      self:putIngredientsBack()
      self.actor:talentDialogReturn(nil)
      game:unregisterDialog(self)
    end,
  }
  self.c_ok = Button.new {
    text = 'Review and Finish',
    fct = function()
      local state = {
	tmpl_obj = self.tmpl_obj,
	selected = self.selected,
	ingr_pulled = self.ingr_pulled,
	pval = self.c_pval.number,
      }
      self.actor:talentDialogReturn(state)
      game:unregisterDialog(self)
    end,
  }
  self.c_pval = Numberbox.new {
    title = 'Pval: ',
    number = init_pval,
    min = self.base_obj.pval or 1,
    max = 999,	-- This will be refined in :updatePval() below.
    chars = 3,
    fct = function(text) end,
    on_change = function(num) self:pvalChanged(num) end,
  }
  self.c_pval.display = function(zelf, x, y, nb_keyframes, ox, oy)
    -- Vaguely Annoying Hack(TM):  Record screen location for relative
    -- tooltip display.
    zelf.last_display_x = ox
    zelf.last_display_y = oy
    return Numberbox.display(zelf, x, y, nb_keyframes, ox, oy)
  end
  local btm_h = self.c_ok.h + self.c_pval.h

  self.c_status = StatusBox.new {
    --font = core.display.newFont('/data/font/DroidSands.ttf', 16),
    width = self.iw - 10 - self.c_ok.w - self.c_cancel.w,
    delay = 1,
  }

  self.c_tmpl_desc = Textzone.new {
    width = math.floor(self.iw/2) - 10,
    height = math.floor((self.ih - btm_h)/2) - 10,
    scrollbar = true,
    text = '',
  }

  self:generateIngredientsList()
  self.c_ingr_list = ListColumns.new {
    width = math.floor(self.iw/2) - 10,
    height = math.floor((self.ih - btm_h)/2) - 10,
    list = self.ingr_list,
    columns = { { name='Ingredients', width=100, display_prop='name' } },
    scrollbar = true,
    select = function(item, sel) self:showListTooltip(item) end,
    fct = function(item, sel) self:selectIngredient(item) end,
  }
  self.c_ingr_list.display = function(zelf, x, y, nb_keyframes, screen_x, screen_y, offset_x, offset_y, local_x, local_y)
    -- Vaguely Annoying Hack(TM), prologue:  Save this information for
    -- :updateIngredientsDesc() below.
    zelf._last_loc = {
      x = x, y = y,
      sx = screen_x, sy = screen_y,
      ox = offset_x, oy = offset_y,
      lx = local_x, ly = local_y,
    }
    ListColumns.display(zelf, x, y, nb_keyframes, screen_x, screen_y, offset_x, offset_y, local_x, local_y)
  end

  self:generateRecipeList()
  self.c_recipe_list = ListColumns.new {
    width = math.floor(self.iw/2) - 10,
    height = self.ih - btm_h - 10,
    list = self.recipe_list,
    columns = { { name='Known Flags', width=100, display_prop='name' } },
    scrollbar = true,
    select = function(item, sel) self:showListTooltip(item) end,
    fct = function(item, sel) self:selectRecipe(item) end,
  }

  local h_sep = Separator.new {
    dir = 'horizontal',
    size = self.ih - btm_h - 10,
  }
  local v_sep = Separator.new {
    dir = 'vertical',
    size = math.floor(self.iw/2) - 20,
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_recipe_list },
    { right=0, top=0, ui=self.c_tmpl_desc },
    { right=0, bottom=btm_h, ui=self.c_ingr_list },
    { hcenter=0, top=5, ui=h_sep },
    { vcenter=-math.floor(btm_h/2), right=5, ui=v_sep },
    { left=0, bottom=0, ui=self.c_ok },
    { left=self.c_ok, bottom=0, ui=self.c_status },
    { right=0, bottom=0, ui=self.c_cancel },
    { left=0, bottom=self.c_ok.h, ui=self.c_pval, hidden=true },
  }
  self:setFocus(self.c_recipe_list)
  self:setupUI()

  self:updateArtifactDesc()
  self:updateIngredientsDesc()
  self:updatePvalField()
  self:updateFinishButton()
end

function _M:putIngredientsBack()
  local inven = self.actor:getInven('INVEN')
  for _, ingrs in pairs(self.ingr_pulled) do
    for _, o in ipairs(ingrs) do self.actor:addOrDropObject(inven, o) end
  end
  self.actor:sortInven()
end

function _M:on_error()
  self:putIngredientsBack()
end

function _M:status(msg, color)
  self.c_status:setTextColor(msg, color)
end

function _M:on_focus(id, ui)
  if self.focus_ui and self.focus_ui.ui == self.c_pval then
    if self.pval_tooltip and self.c_pval.last_display_x then
      game:tooltipDisplayAtMap(self.c_pval.last_display_x + self.c_pval.w, self.c_pval.last_display_y, self.pval_tooltip)
    end
  end
end

function _M:showListTooltip(item)
  if item.last_display_x and item.tooltip then
    local x
    if self.focus_ui and self.focus_ui.ui == self.c_recipe_list then
      x = item.last_display_x + self.c_recipe_list.w
    elseif self.focus_ui and self.focus_ui.ui == self.c_ingr_list then
      x = item.last_display_x - game.tooltip.w
    end
    game:tooltipDisplayAtMap(x or item.last_display_x, item.last_display_y, item.tooltip)
  end
end

function _M:getIngredientCandidates(def)
  local ret = { n=0 }
  self.actor:inventoryApply(self.actor:getInven('INVEN'), function(_, _, o)
    if util.alchemy.ok_ingredient(o, def, self.actor) then
      ret.n = ret.n + o:getNumber()
      ret[#ret+1] = o
    end
  end)
  return ret
end

function _M:ingredientCount(def)
  local ret = self:getIngredientCandidates(def)
  return ret and ret.n or 0
end

function _M:pullIngredientsFor(def, needed)
  local cands = self:getIngredientCandidates(def)
  local inven = self.actor:getInven('INVEN')

  local function punt_on_fail(o, got)
    self:status('??? Failed to pull ' .. o:getName(), colors.ORANGE)
    for _, oo in ipairs(got) do
      self.actor:addOrDropObject(self:getInven('INVEN'), oo)
    end
    self.actor:sortInven()
    return nil
  end

  -- If we don't have enough candidates, pull what we've got and return
  -- them with a need_more flag for the remainder.
  if cands.n < needed then
    local ret = { n=cands.n, need_more=needed-cands.n }
    for _, o in ipairs(cands) do
      local pos = self.actor:itemPosition(inven, o, true)
      if not pos then return punt_on_fail(o, ret) end
      ret[#ret+1] = self.actor:removeObject(inven, pos, true)
    end
    return ret
  end
  -- If we have exactly as many candidates as we need, pull them and return
  -- them.
  if cands.n == needed then
    local ret = { n=needed }
    for _, o in ipairs(cands) do
      local pos = self.actor:itemPosition(inven, o, true)
      if not pos then return punt_on_fail(o, ret) end
      ret[#ret+1] = self.actor:removeObject(inven, pos, true)
    end
    return ret
  end
  -- Otherwise, if we have a single stack of more candidates than we need,
  -- unstack as many as we need and return them.
  if #cands == 1 then
    local ret = { n=needed }
    local pos = self.actor:itemPosition(inven, cands[1], true)
    if not pos then return punt_on_fail(o, ret) end
    ret[1] = self.actor:removeObject(inven, pos, needed)
    return ret
  end
  -- Failing that, flag these ingredients for player selection via
  -- :selectIngredient().
  return { n=0, need_sel=needed }
  --[]
end

function _M:nestedTalentDialog(d)
  local co = coroutine.create(function()
    print('[DEBUG] entering nestedTalentDialog coroutine')
    local ret = self.actor:talentDialog(d)
    print(('[DEBUG] nestedTalentDialog coroutine returning %s'):format(tostring(ret)))
    return ret
  end)
  local ok, ret = coroutine.resume(co)
  print(('[DEBUG] nestedTalentDialog returned %s, %s'):format(tostring(ok), tostring(err)))
  if not ok then
    print(debug.traceback(co))
    if d.on_error then d:on_error() end
    if ret then error(ret) end
    return nil
  end
  return ret
end

function _M:updateArtifactDesc()
  local desc = self.tmpl_obj:getDesc()
  desc = desc:gsub('%(Exp:(%-%d+)%)', '(Exp:#LIGHT_RED#%1#LAST#)')
  self.c_tmpl_desc.text = desc
  self.c_tmpl_desc:generate()
end

function _M:updateIngredientsDesc()
  self:generateIngredientsList()
  self.c_ingr_list:setList(self.ingr_list)
  if self.c_ingr_list._last_loc then
    -- Vaguely Annoying Hack(TM):  ListColumns:display() is what populates
    -- last_display_x for the list items, which we need for tooltip
    -- display; since we just installed a new list, we need to repopulate
    -- that info manually.
    local L = self.c_ingr_list._last_loc
    self.c_ingr_list:display(L.x, L.y, 0, L.sx, L.sy, L.ox, L.oy, L.lx, L.ly)
    if #self.ingr_list > 0 and not self.ingr_list[1].last_display_x then
    end
  end
end

function _M:generateIngredientsList()
  local list = {}

  for _, def in ipairs(self.actor.alchemy_flag_recipe_def) do
    local ingrs = self.ingr_pulled[def.define_as]
    if ingrs then
      for _, o in ipairs(ingrs) do
	list[#list+1] = {
	  name = o:getName(),
	  tooltip = ([[For flag '%s']]):format(def.name),
	}
      end
      if ingrs.need_sel then
	local s, name = '', def.ingredient.name
	if def.multiple and ingrs.need_sel > 1 then
	  s = 's'
	  name = ('%d %s'):format(ingrs.need_sel, def.ingredient.plural_name)
	end
	list[#list+1] = {
	  name = '#{bold}##LIGHT_BLUE#Select '..name..'#LAST##{normal}#',
	  sel_recipe = def,
	  needed = ingrs.need_sel,
	  tooltip = ([[Choose ingredient%s to use for flag '%s']]):format(s, def.name),
	}
      end
      if ingrs.need_more then
	local s, name = '', def.name
	if def.multiple and ingrs.need_more > 1 then
	  s = 's'
	  name = ('%d %s'):format(ingrs.need_more, def.plural_name)
	end
	list[#list+1] = {
	  name = '#{bold}##LIGHT_RED#Need '..name..'#LAST##{normal}#',
	  tooltip = ([[Iingredient%s needed for flag '%s']]):format(s, def.name),
	}
      end
    end
  end

  self.ingr_list = list
end

function _M:updatePvalField()
  local max, any, blockers = 999, false, {}
  local mult, spent_exp = 0, 0

  for _, def in ipairs(self.actor.alchemy_flag_recipe_def) do
    -- Check all pval-based flag recipes.
    if def.multiple then
      local n_ingr = def.ingredient and self:ingredientCount(def) or 999
      local pulled = self.ingr_pulled[def.define_as]
      if util.alchemy.has_recipe_flags(self.base_obj, def) then
	-- If we already have this flag, max pval is at most current pval,
	-- plus however much we can increase this flag with unpulled
	-- ingredients, plus however much we've already increased this
	-- flags with already pulled ingredients --- unless we don't know
	-- this flag recipe, in which case max pval is current pval and
	-- pval is locked.
	any = true
	mult = mult + (def.exp or 0)
	spent_exp = spent_exp + util.alchemy.exp_needed_for_recipe(def, (self.c_pval.number or 1))
	if self.actor:knowsAlchemyRecipe(def) then
	  max = math.min(max, (self.base_obj.pval or 0) + (pulled and pulled.n or 0) + n_ingr)
	  print(('[DEBUG] %s caps pval at cur %d + %d pulled + %d unpulled -> %d'):format(def.define_as, (self.base_obj.pval or 0), (pulled and pulled.n or 0), n_ingr, max))
	else
	  max = math.min(max, self.base_obj_pval or 0)
	  blockers[#blockers+1] = def.name
	end
      elseif self.selected[def.define_as] then
	-- If we've selected this flag, max pval is at most however much we
	-- can increase this flag with unpulled ingredients, plus however
	-- much we've already increased this flag with pulled ingredients.
	-- We count on error checking elsewhere to ensure that this is not
	-- less than current pval.
	any = true
	mult = mult + (def.exp or 0)
	spent_exp = spent_exp + util.alchemy.exp_needed_for_recipe(def, (self.c_pval.number or 1))
	max = math.min(max, n_ingr + (pulled and pulled.n or 0))
	print(('[DEBUG] %s caps pval at %d unpulled + %d pulled -> %d'):format(def.define_as, n_ingr, (pulled and pulled.n or 0), max))
      end
    end
  end

  if any and #blockers == 0 then
    -- If any pval-based flags are selected or already present, max pval is
    -- capped by how much we could increase all of them with our current
    -- experience.  Dust off your high school algebra textbooks and look up
    -- the quadratic formula, folks... ;)
    local a = mult/4
    local b = mult
    local c = -self.tmpl_obj.art_exp - spent_exp
    local root = (-b + math.sqrt(b*b - 4*a*c))/(2*a)
    print(('[DEBUG] %gx^2 + %gx = %g -> x = %g'):format(a, b, -c, root))
    max = math.min(max, math.floor(root))
  end

  if #blockers > 0 then
    self.c_pval.max = max
    self.pval_tooltip = ([[#LIGHT_RED#You cannot alter the pval of this artifact, because you have not learned the following artifact flag recipes:
  · %s#LAST#]]):format(table.concat(blockers, "\n  · "))
  else
    self.c_pval.max = 999
    if max < 999 then
      self.pval_tooltip = ([[Max pval #LIGHT_GREEN#%d#LAST#, based on currently selected flags and available ingredients and experience.]]):format(max)
    else
      self.pval_tooltip = nil
    end
  end

  -- Pval field should be visible if any selected or already present flags
  -- are pval-based or if the base object has a pval.
  self:toggleDisplay(self.c_pval, any or self.base_obj.pval)
end

function _M:updateFinishButton()
  local ok = self.tmpl_obj.art_exp >= 0
  for _, ingrs in pairs(self.ingr_pulled) do
    ok = ok and not ingrs.need_sel and not ingrs.need_more
  end
  self:toggleDisplay(self.c_ok, ok)
end

function _M:selectRecipe(item)
  if not item or not item.recipe then return end
  local def = item.recipe
  local tag = def.define_as
  local update = false

  if self.selected[tag] then
    -- Unselect a selected recipe, putting its ingredients back into the
    -- actor's inventory.
    local inven = self.actor:getInven('INVEN')
    for _, o in ipairs(self.ingr_pulled[tag] or {}) do
      self.actor:addObject(inven, o)
    end
    self.actor:sortInven()
    self.ingr_pulled[tag] = nil
    self.tmpl_obj.art_exp = self.tmpl_obj.art_exp + util.alchemy.exp_needed_for_recipe(def, self.tmpl_obj.pval)
    self.selected[tag] = nil
    update = true
  else
    -- Select a creatable unselected recipe, collecting and extracting its
    -- ingredients from the actor's inventory.
    local needed = util.alchemy.num_ingredients_needed(def, self.c_pval.number or 1, self.actor)
    local ingrs = def.ingredient and self:pullIngredientsFor(def, needed)
    if ingrs or not def.ingredient then
      self.selected[tag] = true
      if ingrs then self.ingr_pulled[tag] = ingrs end
      self.tmpl_obj.art_exp = self.tmpl_obj.art_exp - util.alchemy.exp_needed_for_recipe(def, self.c_pval.number or 1)
      update = true
    end
  end

  if update then
    self:updateRecipes()
    self:rebuildTemplate()
    self:updateArtifactDesc()
    self:updateIngredientsDesc()
    self:updatePvalField()
    self:updateFinishButton()
  end
end

function _M:selectIngredient(item)
  if item and item.sel_recipe and item.needed then
    local Ingredients = require 'mod.dialogs.AlchemyIngredientsDialog'
    local function cb(choices)
      if choices then
	local inven = self.actor:getInven('INVEN')
	for _, o in ipairs(choices) do
	  local pos = self.actor:itemPosition(inven, o, true)
	  if pos then
	    self.actor:removeObject(inven, pos, o:getNumber())
	  end
	end
	self.ingr_pulled[item.sel_recipe.define_as] = choices
	self:updateRecipes()
	self:updateIngredientsDesc()
	self:updatePvalField()
	self:updateFinishButton()
      end
    end
    local d = Ingredients.new(self.actor, item.sel_recipe, item.needed, cb)
    game:registerDialog(d)
  end
end

function _M:pvalChanged(new_pval)
  local old_pval = self.tmpl_obj.pval or 0
  local update = false
  print(('[DEBUG] pvalChanged:  %d -> %d'):format(old_pval, new_pval))

  -- Collect all the flags we're going to need to juggle ingredients for,
  -- and accumulate total experience needed/refunded for the change.
  local pval_defs, exp_needed = {}, 0
  for _, def in ipairs(self.actor.alchemy_flag_recipe_def) do
    if def.multiple and (self.selected[def.define_as] or util.alchemy.has_recipe_flags(self.base_obj, def)) then
      pval_defs[def.define_as] = def
      print(('[DEBUG]   will need to juggle recipe %s'):format(def.name))
      exp_needed = exp_needed + util.alchemy.exp_needed_for_recipe(def, new_pval, old_pval)
    end
  end

  if new_pval ~= old_pval then
    -- Screw trying to surgically add or remove ingredients as needed while
    -- jugging need_more or need_sel counts; just put back what we had and
    -- pull out what we need.  If that means the player has to reselect
    -- ingredients, at least that's somewhat more convenient now.
    local inven = self.actor:getInven('INVEN')
    for tag, def in pairs(pval_defs) do
      print(('[DEBUG]   putting back ingredients for %s'):format(def.name))
      for _, o in ipairs(self.ingr_pulled[tag] or {}) do
	print(('[DEBUG]     putting back %s'):format(o:getName()))
	self.actor:addOrDropObject(inven, o)
      end
      self.actor:sortInven()
      local dp = new_pval
      if def.multiple and util.alchemy.has_recipe_flags(self.base_obj, def) then
	dp = dp - (self.base_obj.pval or 0)
      end
      local needed = util.alchemy.num_ingredients_needed(def, dp, self.actor)
      print(('[DEBUG]   pulling %d ingredients for %s'):format(needed, def.name))
      if def.ingredient and needed > 0 then
	self.ingr_pulled[tag] = self:pullIngredientsFor(def, needed)
      else
	self.ingr_pulled[tag] = nil
      end
    end
    update = true
  end

  if update then
    self.tmpl_obj.art_exp = self.tmpl_obj.art_exp - exp_needed
    self:updateRecipes()
    self:rebuildTemplate()
    self:updateArtifactDesc()
    self:updateIngredientsDesc()
    self:updatePvalField()
    self:updateFinishButton()
  end
end

function _M:rebuildTemplate()
  local any_pval = self.base_obj.pval and true or false
  local tmpl_exp = self.tmpl_obj.art_exp
  self.tmpl_obj = Object.new()
  table.merge(self.tmpl_obj, self.base_obj, true, {uid=true})
  -- Apply all the flags we've selected.  If any of them are pval-based,
  -- we'll need to copy over the user-set pval later.
  for tag, _ in pairs(self.selected) do
    local def = self.actor.alchemy_flag_recipe_def[tag]
    any_pval = any_pval or def.multiple
    if def.flags then
      -- We set addnumbers=true here so that the 'lite' fields will
      -- merge.
      table.merge(self.tmpl_obj, def.flags, true, nil, nil, true)
    end
    if def.copy then
      table.merge(self.tmpl_obj, def.copy, true)
    end
  end
  if any_pval then self.tmpl_obj.pval = self.c_pval.number end
  self.tmpl_obj.art_exp = tmpl_exp
end

function _M:updateRecipe(item)
  local already_has = util.alchemy.has_recipe_flags(self.base_obj, item.recipe)
  local ok, def, lines = not already_has, item.recipe, {}

  local function line(hdr, text, good)
    return ('#{bold}#%s:#{normal}#  #LIGHT_%s#%s#LAST#'):format(hdr, (good and 'GREEN' or 'RED'), tostring(text))
  end

  if def.desc then lines[#lines+1] = def.desc end
  if def.skill then
    local ok_skill = def.skill <= self.actor:getSkillLevel('ALCHEMY')
    ok = ok and ok_skill
    lines[#lines+1] = line('Alchemy skill', def.skill, ok_skill)
  end
  if def.exp then
    if not def.multiple then
      local ok_exp = def.exp <= self.tmpl_obj.art_exp
      ok = ok and ok_exp
      lines[#lines+1] = line('Artifact exp', def.exp, ok_exp)
    elseif not self.selected[def.define_as] and not already_has then
      local pval = self.c_pval.number or 1
      local exp_needed = util.alchemy.exp_needed_for_recipe(def, pval)
      local ok_exp = exp_needed <= self.tmpl_obj.art_exp
      local text = ('Base %d; %+d = %d'):format(def.exp, pval, exp_needed)
      ok = ok and ok_exp
      lines[#lines+1] = ('#{bold}#Artifact exp:#{normal}#  (base %d) #LIGHT_%s#%+d=%d#LAST#'):format(def.exp, (ok_exp and 'GREEN' or 'RED'), pval, exp_needed)
    else
      local pval = self.c_pval.number or 1
      local exp_needed = util.alchemy.exp_needed_for_recipe(def, pval+1, pval)
      lines[#lines+1] = ('#{bold}#Artifact exp:#{normal}#  (base %d) #LIGHT_%s#%+d->%d=%d#LAST#'):format(def.exp, (exp_needed <= self.tmpl_obj.art_exp and 'GREEN' or 'RED'), pval, pval+1, exp_needed)
    end
  end
  if def.ingredient then
    local needed = util.alchemy.num_ingredients_needed(def, self.tmpl_obj.pval or 1, self.actor)
    local ok_ingr = needed <= self:ingredientCount(def)
    ok = ok and ok_ingr
    local desc = util.alchemy.ingredient_simple_desc(def, needed)
    lines[#lines+1] = line('Ingredient', desc, ok_ingr)
  end

  item.can_add = ok
  if #lines == 0 then
    item.tooltip = nil
  else
    item.tooltip = table.concat(lines, '\n')
  end
  if already_has then
    item.name = '#LIGHT_BLUE#' .. item.base_name .. '#LAST#'
  elseif self.selected[def.define_as] then
    item.name = '#LIGHT_GREEN#' .. item.base_name .. '#LAST#'
  elseif not ok then
    item.name = '#LIGHT_RED#' .. item.base_name .. '#LAST#'
  else
    item.name = item.base_name
  end
  if self.c_recipe_list then self.c_recipe_list:generateRow(item) end
end

function _M:updateRecipes()
  for _, item in ipairs(self.recipe_list) do
    if item.recipe then self:updateRecipe(item) end
  end
end

function _M:generateRecipeList()
  local list = {}

  for _, g in ipairs(self.actor.alchemy_flag_grp_def) do
    list[#list+1] = { name='#GOLD##{bold}#' .. g.name .. '#{normal}##LAST#' }
    for _, def in ipairs(self.actor.alchemy_flag_recipe_def.by_grp[g.tag]) do
      if self.actor:knowsAlchemyRecipe(def) then
	local item = { base_name=def.name, recipe=def }
	self:updateRecipe(item)
	list[#list+1] = item
	list[def.define_as] = item
      end
    end
  end

  self.recipe_list = list
end
