-- $Id: AlchemyReviewDialog.lua 3041 2017-08-30 02:40:05Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Button = require 'engine.ui.Button'
local Tabs = require 'engine.ui.Tabs'
local Textbox = require 'engine.ui.Textbox'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local Separator = require 'engine.ui.Separator'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor, base_obj, state)
  self.actor = actor
  self.base_obj = base_obj
  self.tmpl_obj = state.tmpl_obj
  self.orig_artifact_name = self.tmpl_obj.artifact_name
  self.selected = state.selected
  self.ingr_pulled = state.ingr_pulled

  Dialog.init(self, 'Confirm Changes', 800, game.h*0.8)

  self.c_ok = Button.new {
    text = 'Accept',
    fct = function() self:finish() end,
  }
  self.c_back = Button.new {
    text = 'Back',
    fct = function()
      self.tmpl_obj.artifact_name = self.orig_artifact_name
      self.actor:talentDialogReturn(nil)
      game:unregisterDialog(self)
    end,
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function()
      self:putIngredientsBack()
      self.actor:talentDialogReturn('cancel')
      game:unregisterDialog(self)
    end,
  }
  self.c_tabs = Tabs.new {
    width = self.iw - 5,
    tabs = {
      { title='Changes', kind='changes' },
      { title='Description', kind='desc' },
      { title='Ingredients', kind='ingr' },
    },
    on_change = function(tag)
      if self.c_desc then self.c_desc:switchItem(tag) end
    end,
  }

  local art_name = self.tmpl_obj.artifact_name
  if not art_name or art_name == 'Becoming' then
    art_name = 'of ' .. self.actor:getName()
  end
  art_name = art_name:gsub("^'(.*)'$", '%1')
  self.tmpl_obj.artifact_name = util.alchemy.cleanup_artifact_name(art_name)
  self.c_name = Textbox.new {
    title = 'Artifact name: ',
    text = art_name,
    chars = 30,
    max_len = 30,
    fct = function() end,
    on_change = function(text) self:updateArtifactName(text) end,
  }

  local vs1 = Separator.new { dir='vertical', size=790 }
  local vs2 = Separator.new { dir='vertical', size=790 }

  local desc_h = self.ih - self.c_tabs.h - self.c_ok.h - self.c_name.h - vs1.h - vs2.h - 10
  self.c_desc = TextzoneList.new {
    width = self.iw - 5,
    height = desc_h,
    scrollbar = true,
    no_color_bleed = true,
  }
  self.c_desc:switchItem('changes', self:changesText())
  self.c_desc:switchItem('desc', self.tmpl_obj:getDesc())
  self.c_desc:switchItem('ingr', self:ingredientsText())
  self.c_desc:switchItem('changes')

  self:loadUI {
    { left=0, top=0, ui=self.c_tabs },
    { left=5, top=self.c_tabs, ui=vs1 },
    { left=0, top=vs1, ui=self.c_desc },
    { left=0, bottom=0, ui=self.c_ok },
    { right=0, bottom=0, ui=self.c_cancel },
    { right=self.c_cancel, bottom=0, ui=self.c_back },
    { left=0, bottom=self.c_ok, ui=self.c_name },
    { left=5, bottom=self.c_name, ui=vs2 },
  }
  self:setupUI()
end

function _M:putIngredientsBack()
  local inven = self.actor:getInven('INVEN')
  for _, ingrs in pairs(self.ingr_pulled) do
    for _, o in ipairs(ingrs) do self.actor:addOrDropObject(inven, o) end
  end
  self.actor:sortInven()
end

function _M:on_error()
  self:putIngredientsBack()
end

function _M:updateArtifactName(text)
  self.tmpl_obj.artifact_name = util.alchemy.cleanup_artifact_name(text)
  self.c_desc:switchItem('desc', self.tmpl_obj:getDesc(), true)
  self.c_tabs:select('desc')
end

function _M:changesText()
  local lines = {}
  for _, def in ipairs(self.actor.alchemy_flag_recipe_def) do
    local line = nil
    if util.alchemy.has_recipe_flags(self.base_obj, def) and def.multiple and self.base_obj.pval < self.tmpl_obj.pval then
      line = ('#{bold}#Increase:#{normal}#  %+d %s'):format(self.tmpl_obj.pval - self.base_obj.pval, def.name)
    elseif self.selected[def.define_as] then
      if def.keep_open then
	line = '#LIGHT_BLUE#Will be kept open to continue gaining experience#LAST#'
      elseif def.multiple then
	line = ('#{bold}#Add:#{normal}#  %+d %s'):format(self.tmpl_obj.pval, def.name)
      else
	line = ('#{bold}#Add:#{normal}#  %s'):format(def.name)
      end
    end
    if line then lines[#lines+1] = line end
  end
  return table.concat(lines, '\n')
end

function _M:ingredientsText()
  local lines = {}
  for _, def in ipairs(self.actor.alchemy_flag_recipe_def) do
    for _, o in ipairs(self.ingr_pulled[def.define_as] or {}) do
      lines[#lines+1] = o:getName()
    end
  end
  if #lines == 0 then lines[1] = '(No ingredients used)' end
  return table.concat(lines, '\n')
end

function _M:isKeepOpen()
  for tag, _ in pairs(self.selected) do
    local def = self.actor.alchemy_flag_recipe_def[tag]
    if def.keep_open then return true end
  end
  return false
end

function _M:finish()
  -- Replace the Becoming artifact with the modified template we've been
  -- building.  I *think* we want to keep the UID unchanged, and
  -- :updateArtifactName() should already have installed the artifact name
  -- for us.
  local uid = self.base_obj.uid
  self.base_obj.replacedWith = false
  self.base_obj:replaceWith(self.tmpl_obj)
  self.base_obj.replacedWith = nil
  self.base_obj.uid = uid
  __uids[uid] = self.base_obj
  -- Just to be sure...
  local DamageType = require 'engine.DamageType'
  self.base_obj.ignore_dam = {
    [DamageType.ACID] = true,
    [DamageType.ELEC] = true,
    [DamageType.FIRE] = true,
    [DamageType.COLD] = true,
  }
  -- If we're not continuing the artifact, clear its exp (if we are, we
  -- already subtracted out the used exp in the template building process).
  if not self:isKeepOpen() then self.base_obj.art_exp = nil end
  -- This should update the actor accordingly.
  self.actor.changed = true
  self.actor:sortInven()

  -- And we're done.
  self.actor:talentDialogReturn(true)
  game:unregisterDialog(self)
  game.log('You create:  %s', self.base_obj:getName())
end
