-- $Id: CharacterSheet.lua 2835 2017-04-11 00:13:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Textzone = require 'engine.ui.Textzone'
local Tab = require 'engine.ui.Tab'
local Separator = require 'engine.ui.Separator'
local SurfaceZone = require 'engine.ui.SurfaceZone'
local Birther = require 'engine.Birther'
local DamageType = require 'engine.DamageType'
local Stats = require 'mod.class.interface.ActorStats'
local Tactic = require 'mod.dialogs.TacticExplorDialog'
local FontPackage = require "engine.FontPackage"
local PlayerSkills = require 'mod.class.interface.PlayerSkills'

module(..., package.seeall, class.inherit(Dialog))

local tut_txt = [[#SLATE##00FF00#Tab#LAST# or #00FF00#Shift-Tab#LAST# to cycle through tabs
#00FF00#'d'#LAST# to save character sheet to file]]

function _M:init(actor)
  self.actor = actor

  Dialog.init(self, 'Character Sheet:  ' .. self.actor.name, math.max(game.w * 0.7, 950), 500)

  self.font = core.display.newFont(FontPackage:getFont('mono_small', 'mono'))
  self.font_h = self.font:lineSkip()
  self.font_w, _ = self.font:size(' ')

  local function mk_tab(tag, title, dflt)
    return Tab.new {
      title = title,
      default = dflt,
      fct = function() end,
      on_change = function(s)
	if s then self:switchTo(tag) end
      end,
    }
  end
  self.t_general = mk_tab('general', 'General', true)
  self.t_background = mk_tab('background', 'Background', false)
  self.t_flags1 = mk_tab('flags1', 'Flags(1)', false)
  self.t_flags2 = mk_tab('flags2', 'Flags(2)', false)
  self.t_flags3 = mk_tab('flags3', 'Flags(3)', false)
  self.t_flags4 = mk_tab('flags4', 'Flags(4)', false)
  self.tab_order = {
    'general', 'background', 'flags1', 'flags2', 'flags3', 'flags4'
  }

  self.c_tut = Textzone.new {
    auto_width = true,
    auto_height = true,
    no_color_bleed = true,
    text = tut_txt,
  }

  self.vs = Separator.new{
    dir = "vertical",
    size = self.font:size(string.rep(' ', 80)),
  }

  self.c_desc = SurfaceZone.new {
    width = self.font_w * 80,
    height = self.font_h * 25,
    alpha = 0,
  }

  self:loadUI {
    { left=0, top=0, ui=self.c_tut },
    { left=0, top=self.c_tut, ui=self.t_general },
    { left=self.t_general, top=self.c_tut, ui=self.t_background },
    { left=self.t_background, top=self.c_tut, ui=self.t_flags1 },
    { left=self.t_flags1, top=self.c_tut, ui=self.t_flags2 },
    { left=self.t_flags2, top=self.c_tut, ui=self.t_flags3 },
    { left=self.t_flags3, top=self.c_tut, ui=self.t_flags4 },
    { left=0, top=self.t_general, ui=self.vs },
    { left=0, top=self.vs, ui=self.c_desc },
  }
  self:setupUI(true, true)

  self:switchTo('general')
  self:updateKeys()
end

function _M:on_register()
  game:onTickEnd(function() self.key:unicodeInput(true) end)
end

function _M:updateKeys()
  self.key:addCommands{
    _TAB = function() self:tabTabs(1) end,
    [{'_TAB', 'shift'}] = function() self:tabTabs(-1) end,
    __TEXTINPUT = function(c)
      if (c == 'd' or c == 'D') and self.actor.player then
	self:dump()
      end
    end,
  }

  self.key:addBinds{
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:tabTabs(delta)
  local idx = util.boundWrap((self.tab_pos or 1) + delta, 1, #self.tab_order)
  self:switchTo(self.tab_order[idx])
end

function _M:switchTo(tag)
  for i, ord in ipairs(self.tab_order) do
    if ord == tag then
      self.tab_pos = i
      break
    end
  end
  self:drawDialog()
  self:updateKeys()
end

function _M:drawDialog()
  local tag = self.tab_order[self.tab_pos or 1]

  self.c_desc.mouse:reset()
  self.c_desc.key:reset()

  local p = self.actor
  local s = self.c_desc.s
  s:erase(0, 0, 0, 0)
  local w, h = 0, 0

  local sets = { flags1=1, flags2=2, flags3=3, flags4=4 }

  local function put(text)
    s:drawColorStringBlended(self.font, text, w, h, 255, 255, 255, true)
    h = h + self.font_h
  end

  if tag == 'general' or tag == 'background' then
    local race = p.descriptor.race
    local d_sub = Birther:getBirthDescriptor('subrace', p.descriptor.subrace)
    if d_sub.name_mod and d_sub.name_mod == 'prefix' then
      race = p.descriptor.subrace .. ' ' .. race
    elseif d_sub.name_mod and d_sub.name_mod == 'suffix' then
      race = race .. ' ' .. p.descriptor.subrace
    end

    put(' Name  : ' .. p.name:sub(1, 22))
    -- TODO Use sex of possessed body as necessary.
    put(' Sex   : ' .. p.descriptor.sex)
    put(' Race  : ' .. race)
    put(' Class : ' .. p.descriptor.class)
    -- TODO Use race of possessed body as necessary.
    put(' Body  : Player')
    put(' God   : ' .. p.descriptor.god)

    w, h = 32*self.font_w, 0

    -- TODO Implement these
    put(('Age               %4d'):format(p.age))
    put(('Height            %4d'):format(p.height))
    put(('Weight            %4d'):format(p.weight))
    put(('Social Class      %4d'):format(p.social_class))

    w, h = 61*self.font_w, 0

    for _, s in ipairs(Stats.stats) do
      local ss = s
      local sep = p:isStatMaxed(s) and '! ' or ': '
      local val = p:getStat(s, { fmt='default' })
      local eff = p:hasEffect(p['EFF_LOSE_'..s])
      if eff or p:isStatDrained(s) then
	-- Special display of drained stats
	ss = s:lower():capitalize()
	local max_val = p:getStat(s, { base=true, no_tmp=true, fmt='default' })
	local c = eff and '#C__O#' or '#C__Y#'
	val = c .. ('%-6s'):format(val) .. '#LAST# #C_G#' .. max_val .. '#LAST#'
      else
	val = '#C_G#' .. val .. '#LAST#'
      end
      put(ss .. sep .. val)
    end

    w, h = 0, 9*self.font_h

    local show = self:computeShowVals(true)

    put((' + To Melee Hit     #C_B#%7d#LAST#'):format(show.to_h_melee))
    put((' + To Melee Damage  #C_B#%7d#LAST#'):format(show.to_d_melee))
    put((' + To Ranged Hit    #C_B#%7d#LAST#'):format(show.to_h_ranged))
    put((' + To Ranged Damage #C_B#%7d#LAST#'):format(show.to_d_ranged))
    put(('   AC               #C_B#%7s#LAST#'):format(show.ac))

    w, h = 28*self.font_w, 9*self.font_h

    local exp_c = p.exp < p.max_exp and 'C__Y' or 'C_G'

    put(('Level             #C_G#%2d#LAST#'):format(p.level))
    put(('Experience #%s#%9d#LAST#'):format(exp_c, p.exp))
    put(('Max Exp    #C_G#%9d#LAST#'):format(p.max_exp))
    put(('Exp to Adv #C_G#%9s#LAST#'):format(show.exp_adv))
    put(('Gold       #C_G#%9d#LAST#'):format(p.gold))

    w, h = 52*self.font_w, 9*self.font_h

    put('Hit Points   ' .. self:fmtResource('life'))
    put('Spell Points ' .. self:fmtResource('mana'))
    put('Sanity       ' .. self:fmtResource('sanity'))
    if p.last_piety_regen then
      put(('Piety              #C_G#%7d#LAST#'):format(p.piety))
    else
      h = h + self.font_h
    end
    put('Speed           ' .. show.speed)

    if tag == 'background' then
      w, h = 25*self.font_w, 15*self.font_h
      put('(Character Background)')

      local lines = string.splitText(p.history, 60)

      w, h = 10*self.font_w, 16*self.font_h
      for _, line in ipairs(lines) do put(line) end
    else
      w, h = 25*self.font_w, 15*self.font_h
      put('(Miscellaneous Abilities)')

      w, h = 1*self.font_w, 16*self.font_h
      put(('Fighting    : #%s#%s#LAST#'):format(self:abilRating('skill_fighting', 12)))
      put(('Bows/Throw  : #%s#%s#LAST#'):format(self:abilRating('skill_archery', 12)))
      put(('Saving Throw: #%s#%s#LAST#'):format(self:abilRating('saving_throw', 6)))
      put(('Stealth     : #%s#%s#LAST#'):format(self:abilRating('skill_stealth', 1)))
      w, h = 28*self.font_w, 16*self.font_h
      put(('Perception  : #%s#%s#LAST#'):format(self:abilRating('skill_perception', 6)))
      put(('Searching   : #%s#%s#LAST#'):format(self:abilRating('skill_searching', 6)))
      put(('Disarming   : #%s#%s#LAST#'):format(self:abilRating('skill_disarm', 8)))
      put(('Magic Device: #%s#%s#LAST#'):format(self:abilRating('skill_device', 6)))
      w, h = 55*self.font_w, 16*self.font_h
      put(('Blows/Round:  %d'):format(show.num_blows))
      put(('Shots/Round:  %d'):format(1 + p.extra_shots))
      put('Mel.dmg/Rnd:  ' .. show.dmg_desc)
      put(('Infra-Vision: %d feet'):format(p.infravision*10))
      put('Tactic:       ' .. Tactic:getCurDesc(p, 'tactic'))
      put('Explor:       ' .. Tactic:getCurDesc(p, 'explor'))
    end
  elseif sets[tag] then
    local lines, hdr_line = self:flagGrid(sets[tag], true)
    w, h = 0, 0
    put(hdr_line)
    w, h = 40*self.font_w, 0
    put(hdr_line)
    for i, line in ipairs(lines) do
      w, h = (i > 16 and 40 or 0)*self.font_w, (((i-1) % 16)+1)*self.font_h
      put(line)
    end
  end

  self.c_desc:update()
  self.changed = false
end

function _M:computeShowVals(color)
  local p = self.actor
  local ret = {
    to_h_melee = p.dis_to_h + p.to_h_melee,
    to_d_melee = p.dis_to_d + p.to_d_melee,
    to_h_ranged = p.dis_to_h + p.to_h_ranged,
    to_d_ranged = p.dis_to_d + p.to_d_ranged,
    ac = ('%d%+d'):format(p.ac, p.dis_to_a),
    exp_adv = '   ******',
  }

  local inv = p:getInven('MAIN_HAND')
  if #inv > 0 and inv[1]:isIDd() then
    ret.to_h_melee = ret.to_h_melee + (inv[1].to_h or 0)
    ret.to_d_melee = ret.to_d_melee + (inv[1].to_d or 0)
  end
  inv = p:getInven('LAUNCHER')
  if #inv > 0 and inv[1]:isIDd() then
    ret.to_h_ranged = ret.to_h_ranged + (inv[1].to_h or 0)
    ret.to_d_ranged = ret.to_d_ranged + (inv[1].to_d or 0)
  end

  if p.level < p.max_level then
    ret.exp_adv = ('%9d'):format(p:getExpChart(p.level + 1))
  end

  local spd = p.speed
  -- Questionable Hack(TM):  Don't display the Auto-search speed penalty
  -- here [cf. src/files.c line 1618]
  -- TODO Why do we do this?
  if p:isTalentActive(p.T_AUTO_SEARCH) then spd = spd + 10 end
  if spd ~= 0 then
    local c = spd > 0 and 'C_G' or 'C_U'
    spd = ('%s (%+d)'):format(spd > 0 and 'Fast' or 'Slow', spd)
    ret.speed = ('%10s'):format(spd)
    if color then ret.speed = '#'..c..'#'..ret.speed..'#LAST#' end
  else
    ret.speed = '    Normal'
  end

  -- We use number of blows for first weapon for display purposes.
  local ms = p:getMeleeStyle()
  local wpns = ms.all_weapons(p)
  ret.num_blows = ms.num_blows(p, wpns[1])
  ret.dmg_desc = ms:dmg_desc(p)

  return ret
end

function _M:fmtResource(field)
  local amt, max_amt = self.actor[field], self.actor['max_'..field]
  local red_cutoff = max_amt * config.settings.tome2.hp_warning_pct/100
  local c = amt < red_cutoff and 'C__R' or
	    amt < max_amt and 'C__Y' or 'C_G'
  return ('#%s#%6d#LAST#/#C_G#%6d#LAST#'):format(c, amt, max_amt)
end

function _M:abilRating(field, scale, no_color)
  local val = self.actor[field]
  local name, color
  if not val then
    name, color = '??????', 'C_R'
  elseif val < 0 then
    name, color = 'Very Bad', 'C_D'
  elseif val > 17 then
    val = math.floor(((val - 17)*5)/2)
    name, color = ('Legendary[%d]'):format(val), 'C_G'
  else
    local levels = {
      {  1,	'Bad',		'C__R' },
      {  2,	'Poor',		'C_R'  },
      {  4,	'Fair',		'C__O' },
      {  5,	'Good',		'C__Y' },
      {  6,	'Very Good',	'C__Y' },
      {  8,	'Excellent',	'C_G'  },
      { 13,	'Superb',	'C__G' },
      { 17,	'Heroic',	'C_G'  },
    }
    val = math.floor(val/scale)
    for _, lvl in ipairs(levels) do
      if val < lvl[1] then
	name, color = lvl[2], lvl[3]
	break
      end
    end
  end
  name = name or '??????'
  color = color or 'C_R'
  if no_color then return name end
  return color, name
end

function _M:dump()
  local p = self.actor
  local p_name = p.name:gsub("[^a-zA-Z0-9_-.]", "_")
  local now = os.date("%Y%m%d-%H%M%S")

  fs.mkdir("/character-dumps")
  local file = "/character-dumps/"..p_name.."-"..now..".txt"
  local fff = fs.open(file, "w")

  local function ln(s, ...)
    if s then fff:write(s:format(...)) end
    fff:write('\n')
  end

  local function stat_str(s)
    local ss = s
    local sep = p:isStatMaxed(s) and '! ' or ': '
    local val = p:getStat(s, { fmt='default' })
    local eff = p:hasEffect('LOSE_'..s)
    if eff or p:isStatDrained(s) then
      -- Special display of drained stats
      ss = s:lower():capitalize()
      local max_val = p:getStat(s, { base=true, no_tmp=true, fmt='default' })
      val = ('%-6s %s'):format(val, max_val)
    end
    return ss .. sep .. val
  end

  local race = p.descriptor.race
  local d_sub = Birther:getBirthDescriptor('subrace', p.descriptor.subrace)
  if d_sub.name_mod and d_sub.name_mod == 'prefix' then
    race = p.descriptor.subrace .. ' ' .. race
  elseif d_sub.name_mod and d_sub.name_mod == 'suffix' then
    race = race .. ' ' .. p.descriptor.subrace
  end

  local mod_vers = game.__mod_info.version
  ln('  [ToME Classic %d.%d.%d Character Dump]',
     mod_vers[1], mod_vers[2], mod_vers[3])
  ln()
  ln(' Name  : %-22s Age               %4d       %s',
     p.name:sub(1, 22), p.age, stat_str('STR'))
     -- TODO Use sex of possessed body as necessary.
  ln(' Sex   : %-22s Height            %4d       %s',
     p.descriptor.sex, p.height, stat_str('INT'))
  ln(' Race  : %-22s Weight            %4d       %s',
     race, p.weight, stat_str('WIS'))
  ln(' Class : %-22s Social Class      %4d       %s',
     p.descriptor.class, p.social_class, stat_str('DEX'))
  -- TODO Use race of possessed body as necessary.
  ln(' Body  : %-42s          %s', 'Player', stat_str('CON'))
  ln(' God   : %-42s          %s', p.descriptor.god, stat_str('CHA'))
  ln()

  local show = self:computeShowVals()
  local piety_line = ''
  if p.last_piety_regen then
    piety_line = ('Piety               %6d'):format(p.piety)
  end

  ln(' + To Melee Hit     %7d Level      %9d    Hit Points   %6d/%6d',
     show.to_h_melee, p.level, p.life, p.max_life)
  ln(' + To Melee Damage  %7d Experience %9d    Spell Points %6d/%6d',
     show.to_d_melee, p.exp, p.mana, p.max_mana)
  ln(' + To Ranged Hit    %7d Max Exp    %9d    Sanity       %6d/%6d',
     show.to_h_ranged, p.max_exp, p.sanity, p.max_sanity)
  ln(' + To Ranged Damage %7d Exp to Adv.%9s    %s',
     show.to_d_ranged, show.exp_adv, piety_line)
  ln('   AC               %7s Gold       %9d    Speed           %10s',
     show.ac, p.gold, show.speed)
  ln()

  ln('                         (Miscellaneous Abilities)')
  ln(' Fighting    : %-12s Perception  : %-12s Blows/Round:  %d',
     self:abilRating('skill_fighting', 12, true),
     self:abilRating('skill_perception', 6, true), show.num_blows)
  ln(' Bows/Throw  : %-12s Searching   : %-12s Shots/Round:  %d',
     self:abilRating('skill_archery', 12, true),
     self:abilRating('skill_searching', 6, true), 1)
  ln(' Saving Throw: %-12s Disarming   : %-12s Mel.dmg/Rnd:  %s',
     self:abilRating('saving_throw', 6, true),
     self:abilRating('skill_disarm', 8, true), show.dmg_desc)
  ln(' Stealth     : %-12s Magic Device: %-12s Infra-Vision: %d feet',
     self:abilRating('skill_stealth', 1, true),
     self:abilRating('skill_device', 6, true), p.infravision*10)
  ln('                                                       Tactic:       %s',
     Tactic:getCurDesc(p, 'tactic'))
  ln('                                                       Explor:       %s',
     Tactic:getCurDesc(p, 'explor'))

  local lines = string.splitText(p.history, 60)
  ln('                         (Character Background)')
  for _, line in ipairs(lines) do ln('          %s', line) end
  ln()

  ln('  [Miscellaneous information]')
  ln()
  ln(' Maximize mode:        %s',
     game.state.birth.maximize and 'ON' or 'OFF')
  ln(' Preserve mode:        %s',
     game.state.birth.preserve and 'ON' or 'OFF')
  ln(' Small Levels:         %s',
     game.state.birth.always_small_levels and 'ALWAYS' or
     config.settings.tome2.allow_small_levels and 'ON' or 'OFF')
  ln(' Arena Levels:         %s',
     config.settings.tome2.allow_empty_levels and 'ON' or 'OFF')
  ln(' Always unusual rooms: %s',
     game.state.birth.ironman_rooms and 'ON' or 'OFF')
  ln()

  local depths = {}
  for zone_id, lev in pairs(game.state.zone_recall_depth) do
    local info = game.state.zone_info[zone_id]
    if info then
      depths[#depths+1] = {
	name = info.name,
	top = info.level_range[1],
	lev = info.level_range[1] + lev - 1,
      }
    end
  end
  if #depths > 0 then
    table.sort(depths, function(a, b)
      return a.top < b.top or (a.top == b.top and a.name < b.name)
    end)
    ln()
    ln(' Recall Depth:')
    for _, z in ipairs(depths) do
      ln("        %s:  Level %d (%d')", z.name, z.lev, 50*z.lev)
    end
  end

  local are = self.dead and 'were' or 'are'

  if game:isTainted() then ln(' You have done something illegal.') end
  if p.flags.EXPERIMENTAL then ln(' You have done something experimental.') end
  -- TODO stupid monsters?

  -- TODO Use possessed monster name here.
  ln(' Your body is %s.', 'a Player')

  local eff = p:hasEffect(p.EFF_MIMIC_FORM)
  if eff and eff.mimic then
    local a = eff.mimic.name:find('^[AEIOUaeiou]') and 'an' or 'a'
    ln(' You %s disguised as %s %s.', are, a, eff.mimic.name)
  end

  if not self.dead then
    -- TODO Compute this better.
    local desc = game:getSaveDescription()
    local _, _, loc = desc.description:find('Currently (.*)$')
    if loc then ln(' You are currently %s.', loc) end
  end

  local kills = p.kills and p.kills.all or 0
  if kills == 0 then
    ln(' You have defeated no enemies yet.')
  elseif kills == 1 then
    ln(' You have defeated one enemy.')
  else
    ln(' You have defeated %d enemies.', kills)
  end

  local hook_lines = {}
  self:triggerHook{ 'CharacterSheet:dumpLine', lines=hook_lines }
  for _, line in ipairs(hook_lines) do ln(' '..line) end

  ln()
  ln(game.calendar:getTimeDate(0, 'You started your adventure the %s day of the %s year of the third age.'))
  local pfx = p.dead and 'You ended your adventure' or 'It is currently'
  ln(game.calendar:getTimeDate(0, pfx..' the %s day of the %s year of the third age.'))
  local days = math.floor(game.turn/game.calendar.DAY)
  pfx = p.dead and 'Your adventure lasted' or 'You have been adventuring for'
  ln(pfx..' %d day%s.', days, (days == 1 and '' or 's'))

  if p.dead then
    local lines = spell.self_knowledge(p)
    if #p > 0 then
      ln()
      ln('     Your Attributes:')
      for _, line in ipairs(lines) do ln(line) end
    end
  end

  local lines, hdr_line = self:flagGrid()
  ln('')
  ln('        '..hdr_line)
  for _, line in ipairs(lines) do ln('        '..line) end

  local skill_lines = {}

  local show_skill
  local function show_skill(tag)
    local def = PlayerSkills.skills_def[tag]
    local sk = p.skills[tag]
    if def.visible and not def.visible() then return false end
    if sk.mod > 0 then return true end
    for _, child in ipairs(sk.children or {}) do
      if show_skill(child) then return true end
    end
    return false
  end

  local build_skill_lines
  local function build_skill_lines(tag, indent)
    if not show_skill(tag) then return end
    local def = PlayerSkills.skills_def[tag]
    local sk = p.skills[tag]

    local name = def.name
    name = indent .. (def.children and ' - ' or ' . ') .. name
    local value = math.min(sk.value/PlayerSkills.skill_step, PlayerSkills.skill_max)
    local mod = sk.mod/PlayerSkills.skill_step

    skill_lines[#skill_lines+1] = ('%-48s %-06.3f [%05.3f]'):format(name, value, mod)
    for _, child in ipairs(def.children or {}) do
      build_skill_lines(child, indent .. '        ')
    end
  end

  for _, tag in ipairs(PlayerSkills.skills_toplevel) do
    build_skill_lines(tag, '')
  end

  ln()
  ln('Skills (points left: %d)', p.skill_points)
  for _, line in ipairs(skill_lines) do ln(line) end
  ln()

  local any_abil = false
  for _, v in pairs(p.abilities) do if v then any_abil = true end end
  if any_abil then
    ln('Abilities')
    for _, tag in ipairs(PlayerSkills.abilities_order) do
      if p.abilities[tag] then
	ln(' * %s', PlayerSkills.abilities_def[tag].name)
      end
    end
    ln()
  end

  if #p.fates > 0 then
    ln()
    ln('  [Fates]')
    ln()
    local fates = p:dumpFates()
    for _, line in ipairs(fates) do ln(line) end
    if fates.any_unknown then ln('You do not know all of your fate.') end
  end
  ln()
  ln()

  local function obj_desc(o)
    local desc = o:getDesc{chardump=true}
    if desc and desc:len() > 0 then
      for _, part in ipairs(desc:split('\n')) do
	for _, line in ipairs(string.splitText(line, 72)) do
	  ln('    ' .. line)
	end
      end
    end
  end

  local any_wield = false
  for iid = 1, #p.inven_def do
    if p.inven[iid] and p.inven_def[iid].is_worn and #p.inven[iid] > 0 then
      any_wield = true
      break
    end
  end
  if any_wield then
    ln('  [Character Equipment]')
    ln()
    local idx = 0
    for iid = 1, #p.inven_def do
      if p.inven[iid] and p.inven_def[iid].is_worn then
	for item, o in ipairs(p.inven[iid]) do
	  local ch = string.char(string.byte('a') + idx)
	  ln('%s) %s', ch, o:getName())
	  obj_desc(o)
	  idx = idx + 1
	end
      end
    end
    ln()
    ln()
  end

  ln('  [Character Inventory]')
  ln()
  for item, o in ipairs(p:getInven('INVEN')) do
    local ch = string.char(string.byte('a') + item - 1)
    ln('%s) %s', ch, o:getName())
    obj_desc(o)
  end
  ln()
  ln()

  -- Stores and the Mathom-house.
  local store_keys = {
    'town-bree-1/home-BREE_HOME',
    'town-lothlorien-1/home-LOTHLORIEN_HOME',
    'town-minas-anor-1/home-MINAS_ANOR_HOME',
    'town-gondolin-1/home-GONDOLIN_HOME',
    'town-khazad-dum-1/home-KHAZAD_DUM_HOME_HOME',
    'town-bree-1/museum-MATHOM_HOUSE',
  }
  local store_names = {
    ['town-bree-1/home-BREE_HOME'] = 'Home Inventory - Bree',
    ['town-lothlorien-1/home-LOTHLORIEN_HOME'] = 'Home Inventory - Lothlorien',
    ['town-minas-anor-1/home-MINAS_ANOR_HOME'] = 'Home Inventory - Minas Anor',
    ['town-gondolin-1/home-GONDOLIN_HOME'] = 'Home Inventory - Gondolin',
    ['town-khazad-dum-1/home-KHAZAD_DUM_HOME_HOME'] = 'Home Inventory - Khazad-dûm',
    ['town-bree-1/museum-MATHOM_HOUSE'] = 'The Mathom-house Inventory - Bree',
  }
  for _, key in ipairs(store_keys) do
    if game.state.stores and game.state.stores[key] then
      local inven = game.state.stores[key]:getInven('INVEN')
      if #inven > 0 then
	ln('  [%s]', store_names[key])
	ln()
	for item, o in ipairs(inven) do
	  -- Slight Hack(TM):  Since the Mathom-house can have more than 26
	  -- items in it, we wrap the lettering at 'z'.
	  local ch = string.char(string.byte('a') + ((item - 1) % 26))
	  ln('%s) %s', ch, o:getName())
	  obj_desc(o)
	end
	ln()
	ln()
      end
    end
  end

  fff:close()

  Dialog:simplePopup("Character dump complete", "File: "..fs.getRealPath(file))
end

local grid_rows = {
  { set=1, hdr='Add Str', grp='stat_flags', flag='STR', show_pval=true },
  { set=1, hdr='Add Int', grp='stat_flags', flag='INT', show_pval=true },
  { set=1, hdr='Add Wis', grp='stat_flags', flag='WIS', show_pval=true },
  { set=1, hdr='Add Dex', grp='stat_flags', flag='DEX', show_pval=true },
  { set=1, hdr='Add Con', grp='stat_flags', flag='CON', show_pval=true },
  { set=1, hdr='Add Cha', grp='stat_flags', flag='CHA', show_pval=true },
  { set=1, hdr='Mul Mana', grp='pval_flags', flag='mul_mana' },
  { set=1, hdr='Mul SPower', grp='pval_flags', flag='spellpower' },
  { set=1, hdr='Add Stea.', grp='pval_flags', flag='skill_stealth', show_pval=true },
  { set=1, hdr='Add Sear.', grp='pval_flags', flag='skill_searching', show_pval=true },
  { set=1, hdr='Add Infra', grp='pval_flags', flag='infravision', show_pval=true },
  { set=1, hdr='Add Tun..', grp='pval_flags', flag='adj_digging', show_pval=true },
  { set=1, hdr='Add Speed', grp='pval_flags', flag='speed', show_pval=true },
  { set=1, hdr='Add Blows', grp='pval_flags', flag='extra_blows', show_pval=true },
  { set=1, hdr='Chaotic', grp='flags', flag='CHAOTIC' },
  { set=1, hdr='Vampiric', grp='flags', flag='VAMPIRIC' },
  { set=1, hdr='Slay Anim', grp='slay_flags', flag='ANIMAL' },
  { set=1, hdr='Slay Evil', grp='slay_flags', flag='EVIL' },
  { set=1, hdr='Slay Und.', grp='slay_flags', flag='UNDEAD' },
  { set=1, hdr='Slay Demon', grp='slay_flags', flag='DEMON' },
  { set=1, hdr='Slay Orc', grp='slay_flags', flag='ORC' },
  { set=1, hdr='Slay Troll', grp='slay_flags', flag='TROLL' },
  { set=1, hdr='Slay Giant', grp='slay_flags', flag='GIANT' },
  { set=1, hdr='Slay Drag.', grp='slay_flags', flag='DRAGON' },
  { set=1, hdr='Kill Drag.', grp='slay_flags', flag='X_DRAGON' },
  { set=1, hdr='Sharpness', grp='flags', flag='VORPAL' },
  { set=1, hdr='Impact', grp='convey_flags', flag='IMPACT' },
  { set=1, hdr='Poison Brd', grp='brand_dam', flag=DamageType.POISON },
  { set=1, hdr='Acid Brand', grp='brand_dam', flag=DamageType.ACID },
  { set=1, hdr='Elec Brand', grp='brand_dam', flag=DamageType.ELEC },
  { set=1, hdr='Fire Brand', grp='brand_dam', flag=DamageType.FIRE },
  { set=1, hdr='Cold Brand', grp='brand_dam', flag=DamageType.COLD },
  { set=2, hdr='Sust Str', grp='convey_flags', flag='SUST_STR', always_in_dump=true },
  { set=2, hdr='Sust Int', grp='convey_flags', flag='SUST_INT', always_in_dump=true },
  { set=2, hdr='Sust Wis', grp='convey_flags', flag='SUST_WIS', always_in_dump=true },
  { set=2, hdr='Sust Dex', grp='convey_flags', flag='SUST_DEX', always_in_dump=true },
  { set=2, hdr='Sust Con', grp='convey_flags', flag='SUST_CON', always_in_dump=true },
  { set=2, hdr='Sust Cha', grp='convey_flags', flag='SUST_CHA', always_in_dump=true },
  { set=2, hdr='Invisible', grp='pval_flags', flag='invisibility', always_in_dump=true },
  { set=2, hdr='Mul life', grp='pval_flags', flag='hp_mult', always_in_dump=true },
  { set=2, hdr='Imm Acid', grp='convey_immune_dam', flag=DamageType.ACID, always_in_dump=true },
  { set=2, hdr='Imm Elec', grp='convey_immune_dam', flag=DamageType.ELEC, always_in_dump=true },
  { set=2, hdr='Imm Fire', grp='convey_immune_dam', flag=DamageType.FIRE, always_in_dump=true },
  { set=2, hdr='Imm Cold', grp='convey_immune_dam', flag=DamageType.COLD, always_in_dump=true },
  { set=2, hdr='Sens Fire', grp='convey_vuln_dam', flag=DamageType.FIRE, always_in_dump=true },
  { set=2, hdr='Reflect', grp='convey_flags', flag='REFLECT', always_in_dump=true },
  { set=2, hdr='Free Act', grp='convey_flags', flag='FREE_ACT', always_in_dump=true },
  { set=2, hdr='Hold Life', grp='convey_flags', flag='HOLD_LIFE', always_in_dump=true },
  { set=2, hdr='Res Acid', grp='convey_resist_dam', flag=DamageType.ACID, always_in_dump=true },
  { set=2, hdr='Res Elec', grp='convey_resist_dam', flag=DamageType.ELEC, always_in_dump=true },
  { set=2, hdr='Res Fire', grp='convey_resist_dam', flag=DamageType.FIRE, always_in_dump=true },
  { set=2, hdr='Res Cold', grp='convey_resist_dam', flag=DamageType.COLD, always_in_dump=true },
  { set=2, hdr='Res Pois', grp='convey_resist_dam', flag=DamageType.POISON, always_in_dump=true },
  { set=2, hdr='Res Fear', grp='convey_resist_dam', flag=DamageType.FEAR, always_in_dump=true },
  { set=2, hdr='Res Light', grp='convey_resist_dam', flag=DamageType.LITE, always_in_dump=true },
  { set=2, hdr='Res Dark', grp='convey_resist_dam', flag=DamageType.DARK, always_in_dump=true },
  { set=2, hdr='Res Blind', grp='convey_resist_dam', flag=DamageType.BLIND, always_in_dump=true },
  { set=2, hdr='Res Conf', grp='convey_resist_dam', flag=DamageType.CONFUSION, always_in_dump=true },
  { set=2, hdr='Res Sound', grp='convey_resist_dam', flag=DamageType.SOUND, always_in_dump=true },
  { set=2, hdr='Res Shard', grp='convey_resist_dam', flag=DamageType.SHARDS, always_in_dump=true },
  { set=2, hdr='Res Neth', grp='convey_resist_dam', flag=DamageType.NETHER, always_in_dump=true },
  { set=2, hdr='Res Nexus', grp='convey_resist_dam', flag=DamageType.NEXUS, always_in_dump=true },
  { set=2, hdr='Res Chaos', grp='convey_resist_dam', flag=DamageType.CHAOS, always_in_dump=true },
  { set=2, hdr='Res Disen', grp='convey_resist_dam', flag=DamageType.DISENCHANT, always_in_dump=true },
  { set=3, hdr='Aura Fire', grp='convey_flags', flag='SH_FIRE' },
  { set=3, hdr='Aura Elec', grp='convey_flags', flag='SH_ELEC' },
  { set=3, hdr='Auto Curse', grp='convey_flags', flag='AUTO_CURSE' },
  { set=3, skip_line=true },
  { set=3, hdr='NoTeleport', grp='convey_flags', flag='NO_TELE' },
  { set=3, hdr='AntiMagic', grp='convey_flags', flag='NO_MAGIC' },
  { set=3, hdr='WraithForm', grp='convey_flags', flag='WRAITH' },
  { set=3, hdr='EvilCurse', grp='flags', flag='TY_CURSE' },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, hdr='Levitate', grp='convey_flags', flag='FEATHER' },
  { set=3, hdr='Lite', grp='flags', flag='LITE1' },
  { set=3, hdr='See Inv', grp='convey_flags', flag='SEE_INVIS' },
  { set=3, skip_line=true },
  { set=3, hdr='Digestion', grp='convey_flags', flag='SLOW_DIGEST' },
  { set=3, hdr='Regen', grp='convey_flags', flag='REGENERATE' },
  { set=3, hdr='Xtra Might', grp='pval_flags', flag='extra_might' },
  { set=3, hdr='Xtra Shots', grp='cum_flags', flag='extra_shots' },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, skip_line=true },
  { set=3, hdr='Activate', grp='flags', flag='ACTIVATE' },
  { set=3, hdr='Drain Exp', grp='convey_flags', flag='DRAIN_EXP' },
  { set=3, hdr='Teleport', grp='convey_flags', flag='TELEPORT' },
  { set=3, hdr='Aggravate', grp='convey_flags', flag='AGGRAVATE' },
  { set=3, hdr='Blessed', grp='convey_flags', flag='BLESSED' },
  { set=3, hdr='Cursed', grp='flags', flag='CURSED' },
  { set=3, hdr='Hvy Curse', grp='flags', flag='HEAVY_CURSE' },
  { set=3, hdr='Prm Curse', grp='flags', flag='PERMA_CURSE' },
  { set=4, hdr='No blows', grp='flags', flag='NEVER_BLOW' },
  { set=4, hdr='Precogn.', grp='convey_flags', flag='PRECOGNITION' },
  { set=4, hdr='B.Breath', grp='convey_flags', flag='BLACK_BREATH' },
  { set=4, hdr='Recharge', grp='flags', flag='RECHARGE' },
  { set=4, hdr='Fly', grp='convey_flags', flag='FLY' },
  { set=4, hdr='Mrg.Curse', grp='flags', flag='DG_CURSE' },
  { set=4, skip_line=true },
  { set=4, skip_line=true },
  { set=4, hdr='Sentient', grp='flags', flag='LEVELS' },
  { set=4, hdr='Clone', grp='flags', flag='CLONE' },
  { set=4, skip_line=true },
  { set=4, hdr='Climb', grp='convey_flags', flag='CLIMB' },
  { set=4, skip_line=true },
  { set=4, skip_line=true },
  { set=4, skip_line=true },
  { set=4, hdr='Imm Neth', grp='convey_immune_dam', flag=DamageType.NETHER },
  { set=4, hdr='Orc.ESP', grp='esp_flags', flag='ORC' },
  { set=4, hdr='Troll.ESP', grp='esp_flags', flag='TROLL' },
  { set=4, hdr='Dragon.ESP', grp='esp_flags', flag='DRAGON' },
  { set=4, hdr='Giant.ESP', grp='esp_flags', flag='GIANT' },
  { set=4, hdr='Demon.ESP', grp='esp_flags', flag='DEMON' },
  { set=4, hdr='Undead.ESP', grp='esp_flags', flag='UNDEAD' },
  { set=4, hdr='Evil.ESP', grp='esp_flags', flag='EVIL' },
  { set=4, hdr='Animal.ESP', grp='esp_flags', flag='ANIMAL' },
  { set=4, hdr='TLord.ESP', grp='esp_flags', flag='THUNDERLORD' },
  { set=4, hdr='Good.ESP', grp='esp_flags', flag='GOOD' },
  { set=4, hdr='Nliv.ESP', grp='esp_flags', flag='NONLIVING' },
  { set=4, hdr='Unique.ESP', grp='esp_flags', flag='UNIQUE' },
  { set=4, hdr='Spider.ESP', grp='esp_flags', flag='SPIDER' },
  { set=4, skip_line=true },
  { set=4, skip_line=true },
  { set=4, hdr='Full ESP', grp='esp_flags', flag='ALL' },
}

-- cf. display_player_ben_one() in src/files.c
function _M:flagGrid(set, color)
  local ret = {}
  local p = self.actor

  local eq, head, idx = {}, {}, 0
  for iid = 1, #p.inven_def do
    if p.inven[iid] and p.inven_def[iid].is_worn then
      for _, o in ipairs(p.inven[iid]) do
	eq[#eq+1] = o
	head[#head+1] = string.char(string.byte('a') + idx)
	idx = idx + 1
      end
    end
  end
  local hdr_line = '            ' .. table.concat(head) .. '@'

  for _, row in ipairs(grid_rows) do
    if not set or row.set == set then
      if row.skip_line then
	if set then ret[#ret+1] = '' end
      else
	local line, any = {}, false
	for _, o in ipairs(eq) do
	  local c = '.'
	  if o:knownFieldValue(row.grp, row.flag) then
	    if row.pval and o.pval then
	      c = math.abs(o.pval) > 9 and '*' or (''..math.abs(o.pval))
	      if color and o.pval < 0 then c = '#C__R#'..c..'#LAST#' end
	    else
	      c = '+'
	    end
	    any = true
	  end
	  line[#line+1] = c
	end
	local plyr_has = self:playerHasFlag(row.grp, row.flag) 
	any = any or plyr_has
	line[#line+1] = plyr_has and '+' or '.'
	if any or set or row.always_in_dump then
	  local name = row.hdr
	  if color then
	    for i, c in ipairs(line) do
	      -- Alternating green and slate every two columns, unless
	      -- we've already set red above for negative pvals.
	      if c:len() == 1 then
		line[i] = (((i-1)%4) >= 2 and '#C__G#' or '#C__S#')..c..'#LAST#'
	      end
	    end
	  end
	  ret[#ret+1] = ('%-10s: %s'):format(row.hdr, table.concat(line))
	  if any and color then
	    -- Slight Hack(TM):  We can't do this before formatting, or
	    -- alignment won't work.
	    ret[#ret] = ret[#ret]:gsub('^'..row.hdr, '#C_G#'..row.hdr..'#LAST#')
	  end
	elseif set then
	  ret[#ret+1] = ''
	end
      end
    end
  end

  return ret, hdr_line
end

-- cf. player_flags() in src/files.c
function _M:playerHasFlag(group, field)
  local p = self.actor

  if p:queryLeveledFlagBundles(p.gain_at_level, p.level, group, field) then
    return true
  end

  for _, def in pairs(PlayerSkills.skills_def) do
    local sk = p:getSkillLevel(def.short_name)
    if p:queryLeveledFlagBundles(def.convey_at_level, sk, group, field) then
      return true
    end
  end

  -- TODO Flags from gods
  -- TODO Possessed monster body flags
  -- TODO WRAITH flags if we're in Mandos
  if field == 'BLACK_BREATH' then
    return p:hasEffect(p.EFF_BLACK_BREATH) and true or false
  end
  -- TODO hp_mult if we have HP mods from Maeglin quest, Melkor sacrifice, etc.

  return false
end
