-- $Id: obj-util.lua 3002 2017-07-26 01:36:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Assorted object-related utility functions.

obj_util = {
  -- cf. wpn_confuse() et al. below
  once_per_wpn_hit = {},
  once_per_melee_attack = {},
}

-- Utility method for computing object bonuses.
-- cf. src/object2.c lines 2075-2114
function obj_util.m_bonus(max, lev)
  lev = math.min(lev, 127)
  local bonus = math.floor((max*lev)/128)
  bonus = bonus + (rng.range(0, 127) < ((max*lev) % 128) and 1 or 0)
  local std = math.floor(max/4)
  std = std + (rng.range(0, 3) < (max%4) and 1 or 0)
  return util.bound(rng.normal(bonus, std), 0, max)
end

-- Adds a random resist to a provided object, for use in resolvers.
-- cf. random_resistance() in src/spells2.c
function obj_util.random_resistance(o, idx)
  local DamageType = require 'engine.DamageType'
  local random_resist_flags = {
    { 'convey_immune_dam', DamageType.ACID },
    { 'convey_immune_dam', DamageType.ELEC },
    { 'convey_immune_dam', DamageType.COLD },
    { 'convey_immune_dam', DamageType.FIRE },
    { 'convey_resist_dam', DamageType.ACID },
    { 'convey_resist_dam', DamageType.ACID },
    { 'convey_resist_dam', DamageType.ELEC },
    { 'convey_resist_dam', DamageType.ELEC },
    { 'convey_resist_dam', DamageType.FIRE },
    { 'convey_resist_dam', DamageType.FIRE },
    { 'convey_resist_dam', DamageType.COLD },
    { 'convey_resist_dam', DamageType.COLD },
    { 'convey_resist_dam', DamageType.ACID },
    { 'convey_resist_dam', DamageType.ELEC },
    { 'convey_resist_dam', DamageType.FIRE },
    { 'convey_resist_dam', DamageType.COLD },
    { 'convey_resist_dam', DamageType.POISON },
    { 'convey_resist_dam', DamageType.POISON },
    { 'convey_resist_dam', DamageType.FEAR },
    { 'convey_resist_dam', DamageType.FEAR },
    { 'convey_resist_dam', DamageType.LITE },
    { 'convey_resist_dam', DamageType.DARK },
    { 'convey_resist_dam', DamageType.BLIND },
    { 'convey_resist_dam', DamageType.BLIND },
    { 'convey_resist_dam', DamageType.CONFUSION },
    { 'convey_resist_dam', DamageType.CONFUSION },
    { 'convey_resist_dam', DamageType.SOUND },
    { 'convey_resist_dam', DamageType.SOUND },
    { 'convey_resist_dam', DamageType.SHARDS },
    { 'convey_resist_dam', DamageType.SHARDS },
    { 'convey_resist_dam', DamageType.NETHER },
    { 'convey_resist_dam', DamageType.NETHER },
    { 'convey_resist_dam', DamageType.NEXUS },
    { 'convey_resist_dam', DamageType.NEXUS },
    { 'convey_resist_dam', DamageType.CHAOS },
    { 'convey_resist_dam', DamageType.CHAOS },
    { 'convey_resist_dam', DamageType.DISENCHANT },
    { 'convey_resist_dam', DamageType.DISENCHANT },
  }
  local ff = random_resist_flags[idx]
  local group, damtype = ff[1], ff[2]
  o[group] = o[group] or {}
  o[group][damtype] = true
end

-- Adds a random ability to a provided object, for use in resolvers.
-- cf. src/object1.c lines 1194-1226
function obj_util.random_ability(o)
  local abilities = {
    { 'convey_flags', 'FEATHER' },
    { 'lite_hack' },
    { 'convey_flags', 'SEE_INVIS' },
    { 'esp_flags', 'ALL' },
    { 'convey_flags', 'SLOW_DIGEST' },
    { 'convey_flags', 'REGENERATE' },
    { 'convey_flags', 'FREE_ACT' },
    { 'convey_flags', 'HOLD_LIFE' },
  }
  local ff = rng.table(abilities)
  if ff[1] == 'lite_hack' then
    o.lite = (o.lite or 0) + 1
  else
    local group, flag = ff[1], ff[2]
    o[group] = o[group] or {}
    o[group][flag] = true
  end
end

-- Select a random coin type loosely weighted by level.
function obj_util.choose_coin_type(lev)
  local Object = require 'mod.class.Object'
  if #Object.coin_types == 0 then
    for _, o in ipairs(game.zone.object_list) do
      if o.type == 'money' and o.subtype == 'money' and o.define_as ~= 'MONEY_BASE' then
	Object.coin_types[#Object.coin_types+1] = o.define_as
      end
    end
  end

  local i = math.floor(rng.range(3, lev+4)/2)
  if rng.chance(20) then i = i + rng.range(1, lev+1) end
  i = math.min(i, #Object.coin_types)
  return Object.coin_types[i]
end

-- Marks the object as cursed and negates the specified list of fields, if
-- resolve_flags.power is negative.  Intended for use by resolvers.
function obj_util.check_cursed(o, to_negate)
  local Object = require 'mod.class.Object'
  if (Object.resolve_flags.power or 0) < 0 then
    o.cursed = true
    for _, field in ipairs(to_negate or {}) do
      o[field] = o[field] and -math.abs(o[field])
    end
  end
end

-- Attempts to enchant an object, a la Scrolls of Enchant Foo (which
-- actually use this method internally).  Artifacts have a chance to resist
-- enchantment, and enchanting an item to or above +0 can break curses.
-- 'which' can contain fields 'to_h', 'to_d', 'to_a' and 'pval'.
-- [cf. enchant() in src/spells2.c]
function obj_util.enchant(o, n, which)
  if not which or not next(which) then
    -- Didn't ask to enchant anything, apparently.
    return false
  end

  -- Large stacks are harder to enchant, except for ammo.
  local prob = o:getNumber() * 100
  if o.type == 'ammo' then prob = math.floor(prob/20) end
  local probs = {
    10, 50, 100, 200, 300, 400, 500, 650, 800, 950, 987, 993, 995, 998, 1000
  }

  local is_art = o:isArtifact()
  local any = false

  -- Try n times.
  for _ = 1, n do
    if rng.range(0, prob-1) < 100 then
      for f, v in pairs(which) do
	if v then
	  local max, mult = 15, 1
	  if f == 'pval' then max, mult = 6, 2 end
	  local chance = o[f] <= 0 and 0 or o[f] > max and 1000 or probs[o[f]*mult]
	  if rng.range(1, 1000) > chance and (not is_art or rng.chance(2)) then
	    o:forAllStack(function(oo) oo[f] = oo[f] + 1 end)
	    any = true
	    if o.cursed and not o:hasFlag('flags', 'PERMA_CURSE') and o[f] >= 0 and rng.chance(4) then
	      game.log('The curse is broken!')
	      o:setCursed(false, true)
	    end
	  end
	end
      end
    end
  end

  -- I think we can assume the player is doing this. ;)
  if any then game.player:recalcEverything() end
  return any
end

function obj_util.consume_desc_suffix(o, simple_name)
  local repl_name = nil
  if o.alchemy_replace_on_use and game.player:getSkillLevel('ALCHEMY') > 0 then
    local xo = game.zone.object_list[o.alchemy_replace_on_use]
    if xo then repl_name = xo:getName{force_id=true} end
  end
  if repl_name then return ', leaving behind ' .. repl_name end
  return ', consuming the ' .. simple_name
end

-- cf. kind_is_good() in src/object2.c
function obj_util.good_kind(o)
  -- Any armor that doesn't start with an AC penalty.
  if o.type == 'armor' then return o.to_a >= 0 end
  -- Any weapon or bow/crossbow/sling that doesn't start with a to-hit/dam
  -- penalty.
  if o.type == 'weapon' or o.type == 'shooter' then
    return o.to_h >= 0 and o.to_d >= 0
  end
  -- Any digger (the T2 code checks for to-hit/dam >= 0, which must be
  -- legacy code, since those aren't used for diggers anymore).
  if o.type == 'tool' and o.subtype == 'digger' then return true end
  -- Any boomerang that doesn't start with a to-hit/dam penalty.
  if o.type == 'throwing' or o.type == 'boomerang' then
    return o.to_h >= 0 and o.to_d >= 0
  end
  -- Any trapkit that doesn't start with a to-hit/dam penalty.
  if o.type == 'misc' and o.type == 'trapkit' then
    return o.to_h >= 0 and o.to_d >= 0
  end
  -- All ammo.
  if o.type == 'ammo' then return true end
  -- Silver rods or higher
  if o.type == 'utility' and o.subtype == 'rod' then
    return o.max_power >= 100
  end
  -- Any rod tip more expensive than Probing.
  if o.type == 'utility' and o.subtype == 'rodtip' then
    return o.cost >= 4500
  end
  -- Any Tome.  [The Mage quest reward Tome is special_gene and thus
  -- already filtered out above.]
  if o.type == 'magic' and o.subtype == 'spellbook' then
    return o.define_as:find('TOME_', 1, true) and true or false
  end
  -- Certain rings and amulets.
  if o.type == 'jewelry' then
    local good = {
      RING_SPEED		= true,
      AMULET_THE_MAGI	= true,
      AMULET_DEVOTION	= true,
      AMULET_WEAPONMASTERY	= true,
      AMULET_TRICKERY	= true,
      AMULET_RESISTANCE	= true,
      AMULET_REFLECTION	= true,
      AMULET_TELEPATHY	= true,
    }
    return good[o.define_as] and true or false
  end
  -- Nothing else.
  return false
end

-- Returns true if a randart can be made out of this object.
-- [cf. kind_is_artifactable() in src/object2.c]
function obj_util.randartable_kind(o)
  local Entity = require 'engine.Entity'
  local powers = Entity:loadList('/data/general/objects/randart-powers.lua')

  -- If any randart power can be applied to this object, we're good.
  for _, pow in ipairs(powers) do
    if obj_util.ok_tsval(o, pow.tsval_filter) then return true end
  end
  return false
end

-- Returns 0 if the object has none of the flags specified in ff, 2 if it
-- has all of them, and 1 if it has some of them.
function obj_util.check_flags(ff, o)
  local has_any, missing_any = false, false
  for field, flags in pairs(ff) do
    if not o[field] then
      missing_any = true
    elseif field == 'lite' then
      if o.lite then has_any = true else missing_any = true end
    else
      for f, _ in pairs(flags) do
	if o[field][f] then has_any = true else missing_any = true end
      end
    end
  end
  return not has_any and 0 or not missing_any and 2 or 1
end

-- Returns true if the provided object matches the provided tval/sval
-- filter.  The filter must contain a number 'tval' field, which must match
-- the object's tval; the 'sval' field can be absent, matching any sval, a
-- number, matching only that sval, or a two-element list of the minimum
-- and maximum matching svals.  As a convenience, a list of tval/sval
-- filters may also be passed in; we return true if any of them match.
function obj_util.ok_tsval(o, tsv)
  if not o.tval or not o.sval then return end
  if not tsv then return false end
  if #tsv > 0 then
    for _, tsv_ in ipairs(tsv) do
      if obj_util.ok_tsval(o, tsv_) then return true end
    end
    return false
  end

  if o.tval ~= tsv.tval then return false end
  if not tsv.sval then return true end
  if type(tsv.sval) == 'number' then return tsv.sval == o.sval end
  if type(tsv.sval) == 'table' then
    return o.sval >= tsv.sval[1] and o.sval <= tsv.sval[2]
  end
  error("ok_tsval():  what kind of sval is '"..type(tsv.sval).."'?")
  return false
end

-- cf. create_artifact() in src/randart.c
function obj_util.convert_to_randart(o, force_no_cursed)
  if not force_no_cursed and rng.chance(13) then
    o.cursed = true
  end

  local Entity = require 'engine.Entity'
  local powers = Entity:loadList('/data/general/objects/randart-powers.lua')
  local n_powers = powers.RA_POWER_COUNT_DUMMY.n_powers()
  if o.cursed then
    n_powers = math.floor(n_powers/2)
  elseif rng.chance(30) then
    n_powers = n_powers * 2
  end

  -- Only keep the ones that match our tval/sval.
  local ok_powers = {}
  for _, pow in ipairs(powers) do
    local ok = obj_util.ok_tsval(o, pow.tsval_filter)
    if (pow.max_pval or 0) > 0 then
      ok = ok and (o.pval or 0) < pow.max_pval
    end
    -- Also filter out "bad" powers with 'power' < 0.  I think at some
    -- point the code must have allowed for generation of "bad" randarts
    -- with only "bad" powers, but that doesn't seem to be used now.
    ok = ok and pow.power > 0
    if ok then ok_powers[#ok_powers+1] = pow end
  end

  local function adj(val)
    if not val or val == 0 then return 0 end
    if val < 0 then return -rng.range(1, -val) end
    return rng.range(1, val)
  end

  local used = {}
  local max_pval = 0
  while n_powers > 0 do
    if #ok_powers == 0 then break end
    local pow, idx = rng.table(ok_powers)
    -- If we have a forbidden flag, take it out of the table.
    local ok = not pow.tsval_filter.forbid or
	       obj_util.check_flags(pow.tsval_filter.forbid, o) == 0
    if not ok then table.remove(ok_powers, idx) end
    -- Some powers can only be used a specified number of times.
    ok = ok and (not pow.max_uses or (used[pow.define_as] or 0) < pow.max_uses)
    -- Check for player level
    if pow.min_player_level > game.player.level then
      ok = ok and rng.chance(pow.min_player_level - game.player.level)
    end
    -- Check rarity
    ok = ok and rng.range(1, pow.rarity[2]) <= pow.rarity[1]

    if ok then
      -- Success!
      print('[RANDART] applying power '..pow.define_as)
      table.mergeAddAppendArray(o, pow.flags, true)
      if o:hasFlag('flags', 'CURSED') then o.cursed = true end
      o.to_h = o.to_h and o.to_h + adj(pow.vals.max_to_h)
      o.to_d = o.to_d and o.to_d + adj(pow.vals.max_to_d)
      o.to_a = o.to_a and o.to_a + adj(pow.vals.max_to_a)
      if max_pval == 0 or (pow.vals.max_pval and pow.vals.max_pval > 0 and pow.vals.max_pval > max_pval) then
	max_pval = pow.vals.max_pval or max_pval
      end
      used[pow.define_as] = (used[pow.define_as] or 0) + 1
      n_powers = n_powers - 1
    end
  end
  o.pval = adj(max_pval)

  -- Just to be sure.
  local DamageType = require 'engine.DamageType'
  table.mergeAddAppendArray(o, {
    ignore_dam = {
      [DamageType.ACID] = true,
      [DamageType.ELEC] = true,
      [DamageType.FIRE] = true,
      [DamageType.COLD] = true,
    }
  }, true);

  if o.cursed then
    -- Curse the artifact.
    if o.pval then o.pval = -math.abs(o.pval) - rng.range(1, 4) end
    if o.to_a then o.to_a = -math.abs(o.to_a) - rng.range(1, 4) end
    if o.to_h then o.to_h = -math.abs(o.to_h) - rng.range(1, 4) end
    if o.to_d then o.to_d = -math.abs(o.to_d) - rng.range(1, 4) end
    local tele = rng.chance(2) and true or nil
    local bad_stuff = {
      flags = { CURSED=true, HEAVY_CURSE=true },
      convey_flags = {
	TY_CURSE = rng.chance(3) and true or nil,
	AGGRAVATE = rng.chance(2) and true or nil,
	DRAIN_EXP = rng.chance(3) and true or nil,
	BLACK_BREATH = rng.chance(3) and true or nil,
	TELEPORT = tele,
	NO_TELE = not tele and rng.chance(3) and true or nil,
      },
    }
    table.mergeAddAppendArray(o, bad_stuff, true)
  end

  local NameGenerator2 = require "engine.NameGenerator2"
  local ngd = NameGenerator2.new('/data/names/randarts.txt')
  local name = ngd:generate()
  o.artifact_name = rng.chance(3) and "'"..name.."'" or 'of '..name
end

-- Create a randart.
function obj_util.create_randart(def)
  local o
  if def then
    o = game.zone:makeEntityByName(game.level, 'object', def)
  else
    local Object = require 'mod.class.Object'
    Object.resolve_flags.force_randart_base = true
    local filter = game.zone:createWeightedObjectFilter({
      treasure=101, combat=101, magic=101, tools=101, junk=0
    })
    o = game.zone:makeEntity(game.level, 'object', filter, lev, true)
  end
  if o then obj_util.convert_to_randart(o) end
  return o
end

-- Create a Nazgul ring.  Used as the drop_special() method of all the
-- Nazgul.
function obj_util.create_nazgul_ring(who)
  local o = obj_util.create_randart('RING_RING_OF_POWER')
  if o then
    o.artifact_name = 'of ' .. who:getName()
  end
  return o
end

-- Used by Object:resolve() to try to turn an object into an artifact.
-- [cf. make_artifact() in src/object2.c]
function obj_util.try_promote_to_artifact(o)
  for _, a in ipairs(game.zone.artifacts_list) do
    -- *sigh* My kingdom for a 'continue' keyword...
    local ok = a.keep_base == o.define_as and not a.insta_art and not a.special_gene

    -- Already-created check
    local key = 'mod.class.Object/'..a.unique
    ok = ok and not game.uniques[key]

    local Object = require 'mod.class.Object'
    local lev = Object.resolve_flags.force_level or game.zone:level_adjust_level(game.level, 'object')
    if ok and a.level > lev then ok = rng.chance((a.level - lev)*2) end

    -- Artifact rarity check
    local adj_rarity = math.floor(a.rarity/2)
    adj_rarity = game.player:getLuckScale(-adj_rarity, adj_rarity)
    ok = ok and rng.chance(a.rarity - adj_rarity)

    if ok then
      -- Success!
      -- Bizarre Hack(TM):  I *think* it'll work to simply copy the
      -- artifact entity on top of the object entity.  Keep your fingers
      -- crossed...
      table.mergeAppendArray(o, a, true, {uid=true})
      print('[OBJUTIL] Promoted to '..o:getName{force_id=true})
      return true
    end
  end
  return false
end

-- cf. acquirement() in src/object2.c
function obj_util.acquirement(x, y, n, great, known, found)
  local Object = require 'mod.class.Object'
  local lev = game.zone:level_adjust_level(game.level)
  for _ = 1, n do
    local RF = Object.resolve_flags
    RF[great and 'force_great' or 'force_good'] = true
    local o = game.zone:makeEntity(game.level, 'object', nil, lev)
    if o then
      if known then o:identify(true) end
      if found then
	if found.getEntityKind and found:getEntityKind() == 'object' then
	  o:setFoundHere({
	    type = 'obj_create',
	    source = o:getName{no_mods=true, no_pval=true, no_inscr=true, force_id=true}
	  })
	else
	  o:setFoundHere(found)
	end
      end
    end
    local dx, dy = util.drop_loc_near(x, y, 10)
    if dx and dy then
      -- Slight Hack(TM):  Remember the object if we can see it.
      if game.player and game.player:hasLOS(dx, dy, nil, 18) then
	o:markForMap(true)
      end
      game.level.map:addObject(dx, dy, o)
    end
  end
end

-- The guts of Actor:sortInven(), moved here so that it can be reused in
-- stores.
function obj_util.sort_inven(inven)
  -- Stack objects first, from bottom
  for i = #inven, 1, -1 do
    -- If it is stackable, look for objects before it that it could stack into
    if inven[i]:stackable() then
      for j = i - 1, 1, -1 do
	if inven[j]:stack(inven[i]) then
	  table.remove(inven, i)
	  break
	end
      end
    end
  end

  -- Sort them
  table.sort(inven, function(a, b)
    -- Sort by decreasing tval first.
    if a.tval ~= b.tval then return b.tval < a.tval end

    -- For flavored types, sort unknown flavors to the end.
    if a:isFlavored() and b:isFlavored() then
      if a:isFlavorKnown() ~= b:isFlavorKnown() then
	return a:isFlavorKnown()
      end
    end

    -- Next sort by increasing sval.
    if a.sval ~= b.sval then return a.sval < b.sval end

    -- For same sval, sort un-ID'd objects to the end.
    if a:isIDd() ~= b:isIDd() then return a:isIDd() end

    -- Some objects have custom sorting.
    if a.sort_order then
      local ret = a:sort_order(b)
      if ret ~= nil then return ret end
    end

    -- TODO T2 sorts by decreasing sale value here.

    -- TODO Should probably sort by various types of enchantment.

    return a:getName() < b:getName()
  end)
end

--
-- Assorted methods related to recharging wands/staffs/rods.
--

-- Attempt to recharge the provided object.  As a convenience, if a nil
-- object is provided, the user will be prompted for an object to recharge
-- via a ShowEquipInvenFloor dialog (in which case the method must be
-- called in a coroutine context).
--
-- Objects that can be recharged should define a recharge{} subtable with
-- two methods:
--   - try(self, who, power):  Attempts to recharge the object with the
--     provided power.  Returns true/false for success resp. failure.
--   - on_fail(self, who, fail):  Called if the recharge fails, to process
--     possible object destruction or loss of previous charges.  This will
--     be skipped for artifacts, which only get their power drained.
--
-- cf. recharge() in src/spells2.c
function obj_util.recharge(who, power, o, inven, item, floor)
  if not o then
    local function filter_cb(o)
      return o.recharge and not o:hasFlag('flags', 'NO_RECHARGE')
    end
    local function sel_cb(o, inven, floor, item, btn, event)
      who:talentDialogReturn(true, {o=o, inven=inven, item=item, floor=floor})
      return true
    end
    local ShowEquipInvenFloor = require 'mod.dialogs.ShowEquipInvenFloor'
    local ret, p = who:talentDialog(ShowEquipInvenFloor.new('Select item to recharge', who, filter_cb, sel_cb))
    if not ret then return nil end
    return obj_util.recharge(who, power, p.o, p.inven, p.item, p.floor)
  end

  -- Caller should already have caught this, but just in case...
  if not o.recharge then return nil end
  -- Try to recharge according to the object.
  local success = o.recharge.try(o, power)
  -- Mark as recharged (this is apparently for Alchemists).
  o.recharged = true
  if success then return true end
  if o.unique then
    -- Artifacts are never destroyed.
    game.log('The recharging backfires - your %s is completely drained!', o:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true})
    o.power = 0
  else
    o.recharge.on_fail(self, who, inven, item, floor)
  end
  return true
end

--
-- Some methods used in actor callbacks for weapon flags.
--

-- Add an actor callback to 'src', accounting for the possibility of such a
-- callback already existing.
function obj_util.add_callback(src, grp, event, cb)
  src[grp] = src[grp] or {}
  if not src[grp][event] then
    src[grp][event] = cb
  elseif _G.type(src[grp][event]) == 'function' then
    src[grp][event] = { src[grp][event], cb }
  elseif _G.type(src[grp][event]) == 'table' then
    table.insert(src[grp][event], cb)
  else
    error(('obj_util.add_callback(%s, %s):  cannot add to existing %s'):format(grp, event, _G.type(src[grp][event])))
  end
end

-- Attempts to confuse the target.
-- cf. src/cmd1.c lines 2987-3006
function obj_util.wpn_confuse(who, target)
  if obj_util.once_per_wpn_hit.confuse then return end
  obj_util.once_per_wpn_hit.confuse = true

  local DamageType = require 'engine.DamageType'
  local t_name = target:getName{def_art=true}
  if target:resistDam(DamageType.CONFUSION, {or_immune=true, learn=true}) then
    game.logSeen(target, '%s is unaffected.', t_name:capitalize())
  elseif rmg.percent(target.level) then
    game.logSeen(target, '%s is unaffected.', t_name:capitalize())
  else
    game.logSeen(target, '%s appears confused.', t_name:capitalize())
    target:setEffect(target.EFF_COFUSED, 10+math.floor(rng.range(1, src:getSkillLevel('COMBAT'))/5), {merge='combine'})
  end
end

-- Heals the attacker based on the amount of damage done to the target.
-- [cf. src/cmd1.c lines 2938-2975]
function obj_util.wpn_vampric_drain(who, target, dam)
  if obj_util.once_per_wpn_hit.vampiric_drain then return end
  obj_util.once_per_wpn_hit.vampiric_drain = true
  -- No effect if we didn't do any damage.
  if dam <= 0 then return end
  -- No effect on undead or non-living.
  if target:has('UNDEAD') or target:has('NONLIVING') then return end
  -- Apparently we can only drain a certain amount per melee attack.
  -- [cf. MAX_VAMPIRIC_DRAIN in src/cmd1.c]
  local OUOPMA = obj_util.once_per_melee_attack
  OUOPMA.vampiric_drain = OUOPMA.vampiric_drain or 100

  local heal = rng.dice(4, math.floor(dam/6))
  heal = math.min(heal, OUOPMA.vampiric_drain)
  OUOPMA.vampiric_drain = OUOPMA.vampiric_drain - heal
  if heal > 0 then
    if not OUOPMA.done_drain_msg then
      local t_name = target:getName{def_art=true}
      game.logPlayer(who, 'Your weapon drains life from %s!', t_name)
      OUOPMA.done_drain_msg = true
    end
    who:heal(heal, target)
  end
end

-- cf. src/dungeon.c lines 3007-3030
function obj_util.obj_teleport(o, who)
  if game.zone.worldmap then return end
  if o.cursed and not who:has('NO_TELE') then
    if who.restStop then
      who:restStop()
      who:run_stop()
    end
    spell.teleport(who, 50)
  elseif who.player then
    -- TODO I think inscribing the object with '.' is supposed to suppress
    -- this; maybe have some way of configuring this?
    who:restStop()
    who:run_stop()
    local function cb(ok) if ok then spell.teleport(who, 50) end end
    local Dialog = require 'engine.ui.Dialog'
    Dialog.yesnoPopup('Auto-Teleport', 'Teleport?', cb)
  end
end

-- cf. activate_ty_curse() in src/spells2.c
function obj_util.apply_ty_curse(o, who)
  local ActorStats = require 'mod.class.interface.ActorStats'
  -- Vaguely Convoluted Hack(TM) to emulate C switch "fall-through"
  -- behavior in the original T2 code.
  local malus = {
    'A', 'A', 'A', 'B', 'B', 'B', 'C', 'C', 'C', 'D', 'D', 'D', 'E', 'E',
    'E', 'A', 'A', 'C', 'E', 'E', 'F', 'F', 'F', 'G', 'H', 'I', 'I'
  }
  local stop_ty = false
  repeat
    local mal = rng.table(malus)
    if mal == 'A' then
      spell.aggravate_near(who)
      if rng.chance(6) then break else mal = 'B' end
    end
    if mal == 'B' then
      -- TODO activate_hi_summon()
      if rng.chance(6) then break else mal = 'C' end
    end
    if mal == 'C' then
      -- TODO some sort of summon
      if rng.chance(6) then break else mal = 'D' end
    end
    if mal == 'D' then
      game.logPlayer(who, 'You feel your life draining away...')
      who:gainExp(-math.floor(who.exp/16))
      if rng.chance(6) then break else mal = 'E' end
    end
    if mal == 'E' then
      local free = who:has('FREE_ACT')
      if not free or not rng.percent(who.saving_throw) then
	game.logPlayer(who, 'You feel like a statue!')
	who:setEffect(who.EFF_PARALYZED, rng.range(1, (free and 3 or 13)))
	stop_ty = true
      end
      if rng.chance(6) then break else mal = 'F' end
    end
    if mal == 'F' then
      who:bumpStatDown(rng.table(ActorStats.stats))
      if rng.chance(6) then break else mal = 'G' end
    end
    if mal == 'G' then
      game.logPlayer(who, 'Huh?  Who am I?  What am I doing?')
      -- TODO lose_all_info()
      break;
    end
    if mal == 'H' then
      local lev = game.zone:level_adjust_level(game.level, 'actor')
      if lev > 65 and not stop_ty then
	-- TODO summon_cyber()
	stop_ty = true
	break;
      end
      mal = 'I'
    end
    if mal == 'I' then
      for _, s in ipairs(ActorStats.stats) do
	if rng.chance(2) then who:bumpStatDown(s) end
      end
    end
  until stop_ty or not rng.chance(3)
end

-- cf. activate_dg_curse() in src/spells2.c
function obj_util.apply_dg_curse(o, who)
  local ActorStats = require 'mod.class.interface.ActorStats'
  -- Vaguely Convoluted Hack(TM) to emulate C switch "fall-through"
  -- behavior in the original T2 code.
  local malus = {
    'A', 'A', 'A', 'B', 'B', 'B', 'C', 'C', 'C', 'D', 'D', 'D', 'E', 'E',
    'E', 'A', 'A', 'C', 'F', 'F', 'G', 'G', 'G', 'H', 'J', 'K', 'I', 'I',
    'I', 'K'
  }
  local stop_dg = false
  repeat
    local mal = rng.table(malus)
    if mal == 'A' then
      spell.aggravate_near(who)
      if rng.chance(8) then break else mal = 'B' end
    end
    if mal == 'B' then
      game.logPlayer(who, 'Oh!  You feel that the curse is replicating itself!')
      -- TODO replicate
      if rng.chance(8) then break else mal = 'C' end
    end
    if mal == 'C' then
      -- TODO curse equipment
      if rng.chance(8) then break else mal = 'D' end
    end
    if mal == 'D' then
      game.logPlayer(who, 'You feel your life draining away...')
      who:gainExp(-math.floor(who.exp/12))
      if rng.chance(2) and not who:hasEffect(who.EFF_BLACK_BREATH) then
	who:setEffect(who.EFF_BLACK_BREATH, 1, {dg_hack=true})
      end
      if rng.chance(8) then break else mal = 'E' end
    end
    if mal == 'E' then
      local free = who:has('FREE_ACT')
      if not free or not rng.percent(who.saving_throw) then
	game.logPlayer(who, 'You feel like a statue!')
	who:setEffect(who.EFF_PARALYZED, rng.range(1, (free and 3 or 13)))
	stop_ty = true
      end
      if rng.chance(7) then break else mal = 'F' end
    end
    if mal == 'F' then
      game.logPlayer(who, 'Whoa!  You see 10 little Morgoths dancing before you!')
      who:setEffect(who.EFF_CONFUSED, rng.range(1, 26), {merge='combine'})
      if rng.chance(2) then stop_dg = true end
      if rng.chance(7) then break else mal = 'G' end
    end
    if mal == 'G' then
      who:bumpStatDown(rng.table(ActorStats.stats))
      if rng.chance(7) then break else mal = 'H' end
    end
    if mal == 'H' then
      game.logPlayer(who, 'Huh?  Who am I?  What am I doing?')
      -- TODO lose_all_info()
      break;
    end
    if mal == 'I' then
      local inven = who:getInven('MAIN_HAND')
      local wpn = inven and inven[1]
      if wpn then
	game.logPlayer(who, 'Your weapon now seems useless...')
	wpn:setFlag('flags', 'NEVER_BLOW')
	if who.recalcEverything then who:recalcEverything() end
      end
      break;
    end
    if mal == 'J' then
      local lev = game.zone:level_adjust_level(game.level, 'actor')
      if lev > 25 and not stop_dg then
	-- TODO summon Thunderlords
	if rng.chance(2) then stop_dg = true end
	break;
      end
      mal = 'K'
    end
    if mal == 'K' then
      for _, s in ipairs(ActorStats.stats) do
	if rng.chance(2) then who:bumpStatDown(s) end
      end
    end
  until stop_dg or rng.chance(4)
end

function obj_util.obj_clone(target)
  -- TODO
end

--
-- Assorted utility methods used by potions.
--

function obj_util.simple_drink(actor)
  game.logSeen(actor, 'You feel less thirsty.')
  return { used=true, destroy=true, id=true }
end

function obj_util.heal_and_learn(actor, amt, src)
  local learn = actor == game.player and actor.life < actor.max_life
  actor:heal(amt, src)
  return learn and actor == game.player
end

function obj_util.set_effect_and_learn(actor, eff_id, dur, p, silent)
  local learn = actor == game.player and not actor:hasEffect(eff_id)
  actor:setEffect(eff_id, dur, p, silent)
  return learn and actor == game.player
end

function obj_util.reduce_effect_and_learn(actor, eff_id, dur, p, silent)
  local learn = actor == game.player and actor:hasEffect(eff_id)
  actor:setEffect(eff_id, dur, p, silent)
  return learn
end

function obj_util.remove_effects_and_learn(actor, ...)
  local learn = false
  for _, eff_id in ipairs({...}) do
    learn = learn or actor:hasEffect(eff_id)
    actor:removeEffect(eff_id)
  end
  return learn and actor == game.player
end

function obj_util.raise_stat_and_learn(actor, s)
  local boost, restore = actor:bumpStatUp(s)
  local learn = (boost or restore) and actor == game.player 
  return { used=true, destroy=true, id=learn }
end

function obj_util.reduce_stat_and_learn(actor, s)
  local drain, sust = actor:bumpStatDown(s)
  local learn = (drain or sust) and actor == game.player 
  return { used=true, destroy=true, id=learn }
end
