-- $Id: util.lua 3037 2017-08-26 23:51:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Assorted utility functions.  We reuse the util{} table created by the
-- engine's utils.lua.

-- cf. PY_FOOD_* in src/defines.h
util.FOOD = {
  max		= 15000,	-- Above this is "gorged".
  full		= 10000,	-- Between this and 'hungry' is normal.
  hungry	= 2000,
  weak		= 1000,
  fainting	= 500,
  starve	= 100,
}

-- Returns true if the two provided tables have the same keys and values.
-- Also returns true if both provided tables are null, or false if only one
-- is.  Subtable fields are compared recursively if 'deep' is true.
function table.equal(t1, t2, deep)
  if not t1 and not t2 then return true end
  if not t1 or not t2 then return false end

  for k, v in pairs(t1) do
    if not t2[k] or (not deep and t2[k] ~= v) then return false end
    if deep and t2[k] ~= v and (type(v) ~= 'table' or type(t2[k]) ~= 'table' or not table.equal(v, t2[k], deep)) then return false end
  end

  for k, v in pairs(t2) do
    if not t1[k] then return false end
    -- Elegant Hack(TM):  If t1[k] and t2[k] are both non-nil, then we've
    -- already compared them in the previous loop.
  end

  return true
end

-- Returns a new table containing the k, v pairs of the original table for
-- which f(k, v) return true.
function table.grep(t, f)
  local ret = {}
  for k, v in pairs(t) do if f(k, v) then ret[k] = v end end
  return ret
end

-- Reverse of T-Engine's string.capitalize().
function string.uncapitalize(s)
  if #s > 1 then
    return s:sub(1, 1):lower() .. s:sub(2)
  elseif #s == 1 then
    return s:lower()
  else
    return s
  end
end

-- Split the provided string on whitespace into lines no longer than the
-- specified number of characters.
function string.splitText(text, width)
  local lines = {}
  while text:len() > width do
    -- Find last space character within our width.
    local idx
    for i = width+1, 1, -1 do
      if text:sub(i, i) == ' ' then
	idx = i
	break
      end
    end
    if not idx then break end
    -- Find non-space characters before and after our found space.
    local e_idx, s_idx = idx, idx
    for i = idx-1, 1, -1 do
      if text:sub(i, i) ~= ' ' then
	e_idx = i
	break
      end
    end
    for i = idx+1, text:len() do
      if text:sub(i, i) ~= ' ' then
	s_idx = i
	break
      end
    end
    if not e_idx or not s_idx then break end
    -- Put the first part of the text into our lines list...
    lines[#lines+1] = text:sub(1, e_idx)
    -- ...and continue with the rest
    text = text:sub(s_idx)
  end
  lines[#lines+1] = text
  return lines
end

-- Equivalent to C's (v << n), since Lua doesn't have a left-shift operator.
function util.lshift(v, n)
  for _ = 1, n do v = v * 2 end
  return v
end

-- Formats a number with the specified format, coloring it green or red
-- according to its sign.
function util.colorize_num(n, fmt)
  local s = (fmt or '%+d'):format(n) .. '#LAST#'
  return (n >= 0 and '#LIGHT_GREEN#' or '#RED#') .. s
end

-- Returns the name of the specified DamageType with its defined color, as
-- a string or TString.
function util.colorize_damtype(typ, ts)
  local dt = require('engine.DamageType'):get(typ)
  if ts then
    local ret = tstring{}
    if dt.color then ret:add({'color', dt.color.r, dt.color.g, dt.color.b}) end
    ret:add(dt.name)
    if dt.color then ret:add({'color', 'LAST'}) end
    return ret
  else
    return (dt.text_color or '#AAAAAA#') .. dt.name .. '#LAST#'
  end
end

-- Returns an object suitable as a 'source' for handling damage from a
-- timed effect with its own source.  The object's getName() will reflect
-- the timed effect and its source (f.i. "Poison from Poison Needle Trap"),
-- and its rsolveSource() will resolve to the effect's source.
function util.effect_source(eff, src, no_src)
  local TimedEffectSource = require 'mod.class.TimedEffectSource'
  return TimedEffectSource.new(eff, src, no_src)
end

-- Returns true if two otherwise identical weapons should be allowed to
-- stack.
function util.weapons_can_stack(w1, w2)
  -- May seem unnecssary, but two Daggers of Slaying that got different
  -- damage dice boosts shouldn't stack.
  if w1.combat[1] ~= w2.combat[1] or w1.combat[2] ~= w2.combat[2] then
    return false
  end
  -- To-hit/to-dam is checked in Object:canStack().
  -- We don't check config.settings.tome2.stack_wpn_armor here, because
  -- we're also called for ammo.
  return true
end

-- Picks a random item out of the target's inventory that matches the
-- provided match_cb() and applies the provided apply_cb() to it.  This
-- mimics code that appears multiple times in src/melee1.c and src/traps.c
-- to do something bad to the player's inventory.
-- apply_cb() may return a value, which will be passed back to the caller.
-- It may also return a second true value, signaling for the outer loop to
-- continue; in that case, the first return value from prior calls to
-- apply_cb() is passed as an additional parameter to subsequent calls.
function util.pick_inven(target, match_cb, apply_cb)
  local ret = nil
  local inven = target:getInven('INVEN')
  if not inven or #inven == 0 then return end
  for _ = 1, 10 do
    -- The T2 code makes ten attempts to randomly pick a suitable item out
    -- of inventory picking from the player's entire 23-slot inventory and
    -- counting an empty slot as a "miss".  We replicate this as a kindness
    -- to the target.
    local item = rng.range(1, 23)
    local o = inven[item]
    if o and match_cb(target, inven, item, o) then
      local keep_going
      ret, keep_going = apply_cb(target, inven, item, o, ret)
      if not keep_going then break end
    end
  end
  return ret
end

-- Used in the change_for_dmgtype() callback of DOOR and LOCKED_DOOR to
-- convert the door into a JAMMED_DOOR.
function util.jam_door(grid, x, y)
  game.logSeenXY(x, y, 'The door seems stuck.')
  local jam = grid.door_lock_power or grid.door_jam_power or 0
  if jam < 7 then
    local function post_cb(_, _, _) spell.door_jam_power = nil end
    spell.door_jam_power = jam+1
    return game.zone.grid_list['JAMMED_DOOR']
  end
  return nil
end

-- Returned as the post_cb return value of the change_for_dmgtype() method
-- of terrain types that have gold embedded in them, to place said gold
-- when dug out.
function util.place_gold_in_grid(grid, x, y)
  local gen = game.zone:getGenerator('object', game.level, gene.level.data.generator.object)
  gen:generateGold { x=x, y=y }
  game.logSeenXY(x, y, 'You have found something!')
end

-- Returned as the post_cb return value of the change_for_dmgtype() method
-- of RUBBLE, to place an object under the rubble 10% of the time.
function util.place_obj_under_rubble(grid, x, y)
  if rng.percent(10) then
    local gen = game.zone:getGenerator('object', game.level, gene.level.data.generator.object)
    gen:generateOne { x=x, y=y, found={type='rubble'} }
    game.logSeenXY(x, y, 'You have found something!')
  end
end

-- Returns a list of {x=...,y=...} map coordinates matching the provided
-- ok_cb() callback, "flood-filled" from the specified sx,sy starting
-- coordinates.  ok_cb() receives (x,y) parameters, and should return true
-- if the grid should be included in the fill, false if not, and nil if the
-- grid should be included but the flood fill should not be continued to
-- its neighbors.
function util.flood_grids(sx, sy, ok_cb)
  local map = game.level.map
  local q = { {x=sx,y=sy} }
  local touched = {}
  local ret = {}

  while #q > 0 do
    local xy = table.remove(q, 1)
    local t_idx = xy.x + xy.y*map.w
    if not touched[t_idx] then
      local ok = ok_cb(xy.x, xy.y)
      if ok or ok == nil then
	touched[t_idx] = true
	ret[#ret+1] = xy
	if ok then
	  q[#q+1] = { x=xy.x-1, y=xy.y }
	  q[#q+1] = { x=xy.x, y=xy.y+1 }
	  q[#q+1] = { x=xy.x+1, y=xy.y }
	  q[#q+1] = { x=xy.x, y=xy.y-1 }
	  q[#q+1] = { x=xy.x+1, y=xy.y-1 }
	  q[#q+1] = { x=xy.x+1, y=xy.y+1 }
	  q[#q+1] = { x=xy.x-1, y=xy.y-1 }
	  q[#q+1] = { x=xy.x-1, y=xy.y+1 }
	end
      end
    end
  end

  return ret
end

-- Returns x,y coordinates of a grid within the specified radius of the
-- specified location that would be suitable for dropping objects onto.
-- Loosely based on drop_near() in src/object2.c
function util.drop_loc_near(cx, cy, rad, try_hard)
  local Map = require 'mod.class.Map'
  if game.level.map:isBound(cx, cy) and game.level.map:checkEntity(cx, cy, Map.TERRAIN, 'floor_bold') and not game.level.map(cx, cy, Map.TRAP) then
    -- Well, that was easy...
    return cx, cy
  end

  local Actor = require 'mod.class.Actor'
  local cands = {}
  -- Find a grid nearby.
  for dx = cx - rad, cx + rad do
    for dy = cy - rad, cy + rad do
      local d = core.fov.distance(cx, cy, dx, dy)
      -- Check in bounds, in radius, floor, no traps, LOS.
      if game.level.map:isBound(dx, dy) and d <= rad and
	 game.level.map:checkEntity(dx, dy, Map.TERRAIN, 'floor_bold') and
	 not game.level.map(dx, dy, Map.TRAP) and
	 Actor.hasLOS(nil, dx, dy, nil, rad, cx, cy)
      then
	cands[d] = cands[d] or {}
	table.insert(cands[d], {dx, dy})
      end
    end
  end
  -- Pick the closest spot we can.
  for d = 0, rad do
    if cands[d] and #cands[d] > 0 then
      local cand = rng.table(cands[d])
      return cand[1], cand[2]
    end
  end

  if try_hard then
    -- Find any suitable map location, and report back that we had to.
    local map = game.level.map
    for _ = 1, 1000 do
      local dx, dy = rng.range(1, map.w - 1), rng.range(1, map.h - 1)
      if map:checkEntity(dx, dy, 'floor_bold') and not map(dx, dy, Map.TRAP) then
	return dx, dy, true
      end
    end
  end

  -- Nothing suitable.
  return nil, nil
end

-- Searches the provided inventory for an object that matches the provided
-- callback.  Returns the first found matching object and its inventory
-- index, or nil if no match is found.  The callback will be called with
-- three parameters:  the object, the index, and the inventory.
function util.findInInventoryBy(inven, f)
  for item, o in ipairs(inven) do
    if f(o, item, inven) then return o, item end
  end
end

-- Searches all of the provided Actor's inventories for an object that
-- matches the provided callback.  Returns the first found matching object,
-- its inventory index, and the ID of the inventory it was found in, or nil
-- if no match is found.  The callback will be called with three
-- parameters:  the object, the index, and the inventory.
function util.findInAllInventoriesBy(who, f)
  for inv_id, inven in pairs(who.inven) do
    local o, item = util.findInInventoryBy(inven, f)
    if o and item then return o, item, inv_id end
  end
end

-- Randomly alter dungeon size as allowed/required by the zone's map flags
-- [cf. src/generate.c lines 8734-8788], and handle one-time-only special
-- levels.  This is intended to be used as the alter_level_data() method of
-- a zone definition.
function util.adjust_level(zone, lev, res)
  if res.generator.map.one_time_level and game.state:doneSpecialLevel(res.generator.map.one_time_level) then
    -- We've already done this special level; copy back the zone-default
    -- generator data.
    res.generator = table.clone(zone.generator, true, zone._no_save_fields)
    res.enter_desc = nil
    res.level_name = nil
  end

  if game.state.birth.ironman_rooms then
    -- Dodgy Hack(TM):  Turn off connectivity checking when ironman rooms
    -- is active, as it's never gonna pass.
    if res.generator.map.class == 'mod.class.generator.map.T2DungeonLevel' then
      res.no_level_connectivity = true
    end
  end

  -- Don't resize wilderness zones, towns, the world map or special levels.
  local ok_resize = not zone.town and not zone.wilderness and
		    not zone.worldmap and not res.generator.map.one_time_level
  if ok_resize then
    local ff = res.generator.map.flags or {}
    if not ff.BIG then
      if ff.SMALLEST then
	if game.player:has('PRECOGNITION') then
	  game.log('[precognition] small level')
	end
	print('[ZONE] force 1/3x1/3 level on SMALLEST flag')
	res.width = math.floor(res.width/3)
	res.height = math.floor(res.height/3)
	-- Small level chance defined as SMALL_LEVEL in src/generate.c
      elseif game.state.birth.always_small_levels or ff.SMALL or (config.settings.tome2.allow_small_levels and rng.chance(6)) then
	if game.player:has('PRECOGNITION') then
	  game.log('[precognition] small level')
	end
	local wf = rng.range(1, 2)
	local hf = rng.range(1, 3)
	print(('[ZONE] generating %d/3x%d/3 level'):format(wf, hf))
	res.width = math.floor((res.width*wf)/3)
	res.height = math.floor((res.height*hf)/3)
      end
    end
  end
end

-- Returns a filter suitable for passing to Zone:makeEntity() that matches
-- any eitity up to 'lev' levels out of depth.  This is intended for use
-- in defineTile() definitions in the map files of various special levels.
function util.ood_filter(lev, base_filter)
  return {
    -- This will be filled in by Static:generate().
    random_filter = base_filter or {},
    add_levels = lev or 0,
  }
end

-- Convenience method to check whether a zone (by table or by name) is the
-- worldmap zone or a wilderness/town/dungeon zone.  Used in various
-- methods called from Game:changeLevel(), which may be called before the
-- destination zone has been resolved from its name.
function util.zone_is(zone, which)
  local zs = type(zone) == 'string' and zone or nil
  local zt = type(zone) == 'table' and zone or nil
  if not zs and not zt then return false end

  if which == 'worldmap' then
    return (zs and zs == 'worldmap') or (zt and zt.worldmap)
  elseif which == 'wilderness' then
    return (zs and zs == 'wilderness') or (zt and zt.wilderness)
  elseif which == 'town' then
    return (zs and zs:find('^town-')) or (zt and zt.town)
  elseif which == 'dungeon' then
    return not util.zone_is(zone, 'worldmap') and not util.zone_is(zone, 'wilderness') and not util.zone_is(zone, 'town')
  else
    return (zs and zs == which) or (zt and (zt[which] or zt.short_name == which))
  end
end

-- Common definition of the on_leave() method for town and wilderness zones.
function util.on_leave_town_wild(from_zone, to_lev, from_lev, to_zone, params)
  params = params or {}
  if util.zone_is(to_zone, 'worldmap') then
    -- Remember where we were in the zone, so that if we come back down
    -- immediately, we'll be returned to the same place.  This will be
    -- cleared when we move on the world map.
    game.player.to_wild_xy = { x=game.player.x, y=game.player.y }
  end
  if util.zone_is(from_zone, 'town') then
    -- Remember this as our recall town.
    game.state:setRecallDest()
  end
  if util.zone_is(from_zone, 'wilderness') and util.zone_is(to_zone, 'dungeon') then
    -- Slight Hack(TM):  If we're entering a dungeon zone from the top via
    -- its wilderness entrance, set town recall target to this wilderness
    -- zone.
    game.state:setRecallDest(to_zone)
  end
end

-- Common definition of the entry_point() method for town and wilderness
-- zones.
function util.town_wild_entry_point(to_zone, to_lev, from_lev, from_zone, params)
  params = params or {}
  -- If we've been explicitly given an entry point, use it.
  if params.entry_point then return params.entry_point end
  -- If we're not coming from another zone, let Game:changeLevel() decide.
  if not from_zone then return nil end
  -- If we're recalling back to town, use what we remembered as our last
  -- location in town before recalling.
  if util.zone_is(to_zone, 'town') and params.recall then
    local dest = game.state:getRecallDest(false)
    if dest and dest.loc then return dest.loc end
  end
  -- If we're coming back from a zone with an entrance in town, use the
  -- location of said entrance.
  if util.zone_is(to_zone, 'town') and to_zone.entry_points_from_zone then
    local spot = to_zone.entry_points_from_zone[from_zone.short_name]
    if spot then return game.level:pickSpot(spot) end
  end
  -- If we went up to the worldmap and immediately back down, return back
  -- in the same place.
  if util.zone_is(from_zone, 'worldmap') and game.player.to_wild_xy then
    local wxy = game.player.to_wild_xy
    game.player.to_wild_xy = nil
    return wxy
  end
  -- If we're coming into the wilderness from a dungeon zone, either via
  -- recall or via going up stairs from the top level, start on the zone
  -- entrance.
  if util.zone_is(to_zone, 'wilderness') and util.zone_is(from_zone, 'dungeon') then
    return { use='default_down' }
  end
  --- (shrug) I got nothin'.
  return nil
end

-- Set quest entrances to the appropriate tile based on the corresponding
-- quest status.  Suitable for use in (or as) the on_enter() method of a
-- zone definition.
function util.juggle_quest_entrances(to_zone, to_lev, from_lev, from_zone, params)
  local Map = require 'mod.class.Map'
  for q_id, loc in pairs(to_zone.quest_entrances or {}) do
    local q = game.player:hasQuest(q_id)
    if q and q.status and loc.replace[q.status] then
      local spot = game.level:pickSpot(loc)
      local repl = loc.replace[q.status]
      if type(repl) == 'string' then
	game.level.map(spot.x, spot.y, Map.TERRAIN, game.zone.grid_list[repl])
      elseif type(repl) == 'table' then
	if repl.grid then
	  game.level.map(spot.x, spot.y, Map.TERRAIN, game.zone.grid_list[repl.grid])
	end
	if repl.trap then
	  local e = game.zone:makeEntityByName(game.level, 'trap', repl.trap)
	  game.zone:addEntity(game.level, e, 'trap', spot.x, spot.y, Map.TRAP)
	end
      end
    end
  end
end

-- Returns a formatted description of a KeyBind-style key mapping
-- definition.  Modified from engine.KeyBind:formatKeyString() for slightly
-- better readability (f.i. Ctrl-Q instead of CQ).
function util.formatKeyString(ks)
  if not ks then return '<<?? unknown>>' end

  local function mods(sfx, ctrl, shift, alt, meta)
    return (meta == 'true' and 'Meta-' or '') ..
	   (alt == 'true' and 'Alt-' or '') ..
	   (shift == 'true' and 'Shift-' or '') ..
	   (ctrl == 'true' and 'Ctrl-' or '') .. sfx
  end

  if ks:find("^uni:") then
    return ks:sub(5)
  elseif ks:find("^sym:") then
    local i, j, sym, ctrl, shift, alt, meta = ks:find("^sym:([^:]+):([a-z]+):([a-z]+):([a-z]+):([a-z]+)$")
    if not i then return '<<?? unknown>>' end
    if sym:sub(1, 1) == "=" then
      sym = sym:sub(2)
    else
      local KeyBind = require('engine.KeyBind')
      if tonumber(sym) then sym = tonumber(sym) else sym = KeyBind[sym] end
      sym = core.key.symName(sym)
      -- Slight Hack(TM):  symName('_z') returns 'Z'.  We want 'z' with
      -- no modifiers, 'Shift-z' with just shift, and eg. 'Ctrl-Z' with
      -- anything else.
      if string.len(sym) == 1 and ctrl ~= 'true' and alt ~= 'true' and meta ~= 'true' then
	sym = string.lower(sym)
      end
    end
    return mods(sym, ctrl, shift, alt, meta)
  elseif ks:find("^mouse:") then
    local i, j, sym, ctrl, shift, alt, meta = ks:find("^mouse:([a-zA-Z0-9]+):([a-z]+):([a-z]+):([a-z]+):([a-z]+)$")
    if not i then return '<<?? unknown>>' end
    return mods(sym, ctr, shift, alt, meta)
  elseif ks:find("^gest:") then
    local i, j, sym = ks:find("^gest:([a-zA-Z0-9]+)$")
    if not i then return '<<?? unknown>>' end
    return sym
  else
    return '<<?? unknown>>'
  end
end

-- Returns a formmatted description of the key bound to a KeyBind virtual
-- action.  If 'sep' is non-null, up to two key descriptions concatenated
-- with the provided separator.
function util.getFormattedKeyBinds(virtual, sep)
  local KeyBind = require('engine.KeyBind')
  if not KeyBind.binds_def[virtual] then return '<<?? unknown>>' end
  local t = KeyBind:getBindTable(KeyBind.binds_def[virtual])
  local ret = {}
  for i = 1, 3 do
    if t[i] and #ret < 2 then ret[#ret+1] = util.formatKeyString(t[i]) end
  end
  if #ret == 0 then return '<<?? unknown>>' end
  return sep and table.concat(ret, sep) or ret[1]
end

-- Returns a list of NPC powers that would become available to the player
-- upon possessing the body of or entering symbiosis with the provided
-- Actor, suitable for use as a worn_talents{} object field.  If a skill
-- name is provided, a skill requirements table will be included; 'minor'
-- and 'major' are the associated skill levels to be required for minor or
-- major powers.  Returns nil if no suitable talents are found.
function util.body_powers(a, sk, minor, major)
  local talents = {}
  for id, t in pairs(a.talents_def or {}) do
    if t.player and a:knowTalent(id) and not t.hide then
      local lev = t.player.major and major or minor
      table.insert(talents, id)
      if sk then
	talents[id] = { skill = { tag=sk, value=lev } }
      end
    end
  end
  if #talents > 0 then return talents end
  return nil
end

-- Utility method to juggle visibility of monsters when the player
-- gains/loses the BLIND or SEE_INVIS flag.
function util.juggle_can_see(who)
  if game.level then
    who:resetCanSeeCache()
    if who.player then
      for _, e in pairs(game.level.entities) do
	if e.x then game.level.map:updateMap(e.x, e.y) end
      end
    end
    game.level.map.changed = true
  end
end

-- Utility method for use by (or as) the on_merge() method of a temporary
-- effect definition.  The effect duration will be altered as specified by
-- the 'merge' field of the effect parameters table passed to :setEffect(),
-- as follows:
--   - If merge='replace', cur.dur will be set to new.dur.  This is also
--     the default if no merge method is specified.
--   - If merge='no_replace', cur.dur will be unchanged.
--   - If merge='combine', new.dur will be added to cur.dur.  In addition,
--     if new{} includes a 'combine_msg' field, it will be logged via
--     game.logSeen(), with #Target#/#target# substitutions.
--   - If merge is a function, it will be called with cur.dur and new.dur
--     as arguments and cur.dur will be set to the result.
-- In addition, the effect definition can specify certain fields to be
-- optionally copied from new{} to cur{} by this method:
--   - Fields listed in def.copy_on_merge{} will be copied unconditionally
--     from new{} to cur{}.
--   - Fields listed in def.copy_if_merge{} will be copied from new{} to
--     cur{} if a merge actually happens (ie. if merge~='no_replace').
--   - Fields listed in def.combine_if_merge{} will be set to the sum of
--     new[field] and cur[field] if merge='combine'; otherwise they are
--     treated as in copy_if_merge{}
-- Returns the in-place-modified cur{} table, as required by the on_merge()
-- signature, unless the resulting merged duration is zero or less, in
-- which case it invokes :removeEffect() and returns nil.
function util.merge_effect_duration(who, cur, new, def)
  for _, f in ipairs(def.copy_on_merge or {}) do cur[f] = new[f] or cur[f] end
  if not new.merge or new.merge == 'replace' then
    cur.dur = new.dur
  elseif new.merge == 'no_replace' then
    return cur
  elseif new.merge == 'combine' then
    cur.dur = cur.dur + new.dur
    if new.dur > 0 and new.combine_msg then
      local name = who:getName{def_art=true, check_seen=true}
      local s = new.combine_msg:gsub('#Target#', name:capitalize()):gsub('#target#', name)
      game.logSeen(who, s)
    end
  elseif type(new.merge) == 'function' then
    cur.dur = new.merge(cur.dur, new.dur)
  else
    print('[EFFECT] unrecognized merge rule '..new.merge)
  end
  if cur.dur <= 0 then
    who:removeEffect(cur.effect_id)
    return nil
  end
  for _, f in ipairs(def.copy_if_merge or {}) do cur[f] = new[f] or cur[f] end
  for _, f in ipairs(def.combine_if_merge or {}) do
    local old_f = new.merge == 'combine' and cur[f] or 0
    cur[f] = (new[f] and new[f] + old_f) or cur[f]
  end
  new.merge = nil
  return cur
end

-- on_melee_hit() callback for NPCs that can give Black Breath on melee hit.
-- [cf. make_attack_normal() and black_breath_attack() in src/melee1.c]
function util.black_breath_melee_hit(src, target, atk)
  -- Let's not try giving Black Breath to other NPCs...
  if not target.player then return end
  -- The code also blocks this if the player has the "protection from
  -- undead" timed effect, but there appears to be no way to confer said
  -- effect.

  local chance
  if src:has('NAZGUL') then
    chance = 4
  elseif src:has('UNDEAD') and src:has('UNIQUE') and src.level >= 35 then
    chance = 300 - src.level
  elseif src:has('UNDEAD') and not src:has('UNIQUE') and src.level >= 40 then
    chance = 450 - src.level
  end

  if rng.chance(chance) then
    game.logPlayer(target, 'Your foe calls upon your soul!')
    target:setEffect(target.EFF_BLACK_BREATH, 1, {})
  end
end

-- on_die_post() callback for the Nazgul and Sauron.
function util.nazgul_die(who, src)
  local Dialog = require 'engine.ui.Dialog'
  local q = game.player:hasQuest('one-ring')
  if not q or not q:isEnded() then
    -- One ring isn't destroyed; we'll be back.  Clear the actor's
    -- generated-unique state.
    who:removed()
    local fmt = [[As %s falls, you sense %s spirit escape %s body and flee.  You suspect you have not seen the last of %s.]]
    local name = who:getName()
    local its = who:pronoun('possessive')
    local it = who:pronoun('object')
    if who.define_as == 'SAURON_THE_SORCERER' then
      fmt = [[As %s falls, you sense %s spirit escape %s body and flee.  It seems Galadriel was right; %s cannot be permanently defeated while the One Ring endures.]]
      it = who:pronoun('subject')
    end
    Dialog:simpleLongPopup(name, fmt:format(name, its, its, it), 600)
  elseif who.define_as == 'SAURON_THE_SORCERER' then
    -- Permanently killed Sauron!  Juggle quests.
    game.player:setQuestStatus('sauron', engine.Quest.DONE)
    local function cb()
      game.player:grantQuest('morgoth')
    end
    local text = [[Sauron falls, and you finally feel his spirit dissipate and vanish.]]
    if game.zone.short_name == 'angband' then
      text = text .. [[  In the distance, you can feel the magic blocking the stairs also fading.  The way is now clear for you to finally face Morgoth.]]
    else
      text = text .. [[  The way should now be clear for you to finally face Morgoth.]]
    end
    Dialog:simpleLongPopup(who:getName(), text, 600, cb)
  else
    -- Permanently killed a Nazgul.  Record toward our achievement.
    world:gainAchievement('NAZGUL_KILL', game.player, who)
  end
end

-- ok_change_level() callback for down stairs on Ang 99 and Ang 100.
function util.angband_down_special(grid, who)
  local data = {
    DOWN_QUEST_SAURON = {
      quest = 'sauron',
      text = [[The stairs seem to be blocked by Sauron's magic.  No doubt you will have to slay him permanently to pass them.]],
    },
    DOWN_QUEST_MORGOTH = {
      quest = 'sauron',
      text = [[The stairs seem to be blocked by Morgoth's magic.  No doubt you will have to slay him to pass them.]],
    }
  }

  if not data[grid.define_as] then return true end
  if not game.player:isQuestStatus(data[grid.define_as].quest, engine.Quest.DONE) then
    local Dialog = require 'engine.ui.Dialog'
    Dialog:simpleLongPopup('Stairs Blocked', data[grid.define_as].text, 600)
    return false
  end
  return true
end

-- Update participating inventory dialogs to reflect a change in inventory
-- caused by a delayed callback or on a separate coroutine.
-- [cf. Player:doWear()]
function util.live_update_inventory_dialogs()
  for _, d in ipairs(game.dialogs) do
    if d.inventoryLiveUpdate then d:inventoryLiveUpdate() end
  end
end

-- Checks a changed value against a list of thresholds.  Arguments are the
-- before and after values, the list of thresholds (in increasing order),
-- and a callback that will be called if the value has crossed one or more
-- thresholds.  The callback will be called with the before and after
-- values, the last threshold crossed (highest if after > before, lowest if
-- after < before), and a list of all thresholds crossed in increasing
-- order.
function util.check_threshold(before, after, levels, callback)
  local threshold, thresholds = nil, {}
  if after > before then
    for _, lev in ipairs(levels) do
      if after >= lev and before < lev then
	threshold = lev
	table.insert(thresholds, lev)
      end
    end
  elseif after < before then
    for _, lev in ipairs(levels) do
      if after < lev and before >= lev then
	threshold = threshold or lev
	table.insert(thresholds, lev)
      end
    end
  end
  if threshold then callback(before, after, threshold, thresholds) end
end

-- Drain 'base' plus 2/'frac' times current exp from provided player,
-- checking for hold life.
function util.plyr_lose_exp(plyr, base, frac, p)
  if not plyr.player then return end
  p = p or {}
  p.hold_life_pct = p.hold_life_pct or 75
  local msg, lose = (p.msg or 'You feel your life draining away!'), 0
  if plyr:has('HOLD_LIFE') and p.hold_life_pct > 0 then
    if rng.percent(p.hold_life_pct) then
      game.logPlayer(plyr, 'You keep hold of your life force.')
    else
      game.logPlayer(plyr, 'You feel your life slipping away!')
      lose = math.floor(base/10) + 2*math.floor(plyr.exp/(frac*10))
    end
  else
    -- Slight Hack(TM):  May want a different message for special cases.
    game.logPlayer(plyr, msg)
    lose = base + 2*math.floor(plyr.exp/frac)
  end
  if lose > 0 then plyr:gainExp(-lose) end
end

-- Selects and returns the x,y coordinates of a grid adjacent to the
-- provided player that matches the provided match(x,y) callback.  If only
-- one adjacent grid matches, it will be returned immediately (unless the
-- 'no_auto' parameter is true); if more than one matches, the player will
-- be prompted to select via targeting.  Returns nil if there were no
-- matching adjacent grids, or if the player cancels targeting, or if the
-- grid selected by the player doesn't match.  'no_match_msg' and
-- 'bad_choice_msg' are optional messages to be logged if, respectively,
-- there is no matching grid adacent to the player or the grid selected by
-- the player does not match the matching callback.
-- NOTE:  This method *must* be called in a coroutine context, as it uses
-- targeting.
function util.select_adjacent(plyr, match, no_match_msg, bad_choice_msg, no_auto)
  local count = 0
  local tx, ty
  for x = plyr.x - 1, plyr.x + 1 do
    for y = plyr.y - 1, plyr.y + 1 do
      if x ~= plyr.x or y ~= plyr.y then
	if game.level.map:isBound(x, y) and match(x, y) then
	  count = count + 1
	  tx, ty = x, y
	end
      end
    end
  end
  if count == 0 then
    game.log(no_match_msg or 'There is nothing suitable here.')
    return nil, nil
  elseif count > 1 or no_auto then
    game.log('Which direction?')
    tx, ty = plyr:getTarget{ range=1, immediate_keys=true, default_target=plyr }
    if not tx or not ty then return nil, nil end
    if not match(tx, ty) then
      game.log(no_match_msg or 'There is nothing suitable there.')
      return nil, nil
    end
  end
  return tx, ty
end

--
-- Alchemy-related functions.
--

util.alchemy = {}

-- Returns a simple description of the specified number (or one, if the
-- flag is not pval-based) of the ingredient required by the provided
-- artifact flag recipe.
function util.alchemy.ingredient_simple_desc(def, needed)
  if not def.ingredient then return nil end
  if not (def.multiple or def.keep_open) or needed == 1 then
    return def.ingredient.name
  end
  return ('%d %s'):format(needed, def.ingredient.plural_name)
end

-- Returns the number of ingredients required for the provided artifact
-- flag recipe, potentially depending on the prevailing pval or the
-- player's Alchemy skill level.
function util.alchemy.num_ingredients_needed(def, pval, actor)
  if def.keep_open then return actor:getSkillLevel('ALCHEMY') end
  if def.multiple then return pval end
  return 1
end

-- Returns true if the specified object is suitable for use as an
-- ingredient for the provided artifact flag recipe.
function util.alchemy.ok_ingredient(o, def, actor)
  -- Must match the recipe's ingredient filter.
  if not game.zone:checkFilter(o, def.ingredient.match, 'object') then
    return false
  end
  -- We must be able to drop it.
  if o:check('on_preremoveobject', actor, actor:getInven('INVEN'), true) or o:on_drop(actor, true) then
    return false
  end
  -- Can't be cursed.
  if o.cursed then return false end
  -- Can't be a Becoming artifact.
  if o.art_exp then return false end
  -- Looks like we're good to go, then.
  return true
end

-- Returns true if the provided object has all of the flags added by the
-- specified artifact flag recipe.
function util.alchemy.has_recipe_flags(o, def)
  if not o then return false end
  if def.flag_group then
    return def.flags and util.alchemy.has_recipe_flags(o, def.flags)
  end
  -- Past this point, 'def' is the flags{} table from the original recipe,
  -- or some subtable thereof.
  for f, v in pairs(def) do
    if f == 'lite' then
      -- Backing the LITE1/2/3 flags out of the 'lite' field is probably
      -- impossible; just assume we can always add them.
      return false
    else
      local vv = o[f]
      if _G.type(v) == 'table' then
	vv = util.alchemy.has_recipe_flags(o[f], v)
      end
      if not vv then return false end
    end
  end
  return true
end

-- Returns a cleaned up version of a candidate artifact name.  The provided
-- name is modified as follows:
--   - Leading and trailing whitespace is removed.
--   - All sequences of consecutive whitespace are collapsed to a single
--     space.
--   - If the result after the preceding is empty, replace the name with
--     "of <name>", where <name> is the :getName() of the provided actor.
--     'who' defaults to the player.
--   - If the resulting name does not begin with "of ", it is surrounded
--     with single quotes.
-- Mainly used by Alchemy artifact creation, but we share it with Scrolls
-- of Artifact Creation.
function util.alchemy.cleanup_artifact_name(name, who)
  name = name:gsub("^%s*(.-)%s*$", "%1")
  name = name:gsub("%s+", ' ')
  if name:len() == 0 then name = 'of ' .. (who or game.player):getName() end
  if not name:match("^of ") then name = "'" .. name .. "'" end
  return name
end

-- Returns the experience needed by the provided recipe for the specified
-- pval (relative to the previous pval, if specified).  Experience needed
-- increases quadratically with pval, specifically x^2/4 + x.
-- [cf. get_flags_exp() in src/cmd7.c]
function util.alchemy.exp_needed_for_recipe(def, pval, prev_pval)
  if not def.multiple then return def.exp or 0 end
  pval = pval or 0
  prev_pval = prev_pval or 0
  local factor = pval*pval/4 + pval
  local prev_factor = prev_pval*prev_pval/4 + prev_pval
  return math.floor((def.exp or 0)*(factor - prev_factor))
end
